(* CPPL *)
Require Export Common.
Open Scope list_scope .

Require Constants.
Module C := Constants.

(* Types *)
  Inductive letval : Set :=
  | New_nonce : letval
  | New_symkey : letval
  | New_pubkey : letval
  | Accept : letval
  | Connect : var -> letval
  | LVMsg : msg -> letval .

Hint Constructors letval.

  Lemma letval_eq_dec : forall x y : letval, {x = y} + {x <> y}.
	
  Proof.
   decide equality. apply var_eq_dec. apply msg_eq_dec.
  Qed.

  Definition letval_vars lv :=
  match lv with
  | New_nonce => empty
  | New_symkey => empty
  | New_pubkey => empty
  | Accept => empty
  | Connect v => (singleton (varid v))
  | LVMsg m => message_vars m
  end .

  Inductive cont : Set :=
  | Fail : cont
  | Return : list formula -> list var -> cont
  | Let : var -> letval -> cont -> cont
	
  | Derive : list formula -> cont -> cont -> cont
  | Send : list formula -> var -> msg -> cont -> cont -> cont
  | Recv : var -> msg -> list formula -> cont -> cont -> cont
  | Call : list formula -> const -> list var -> list var -> list formula -> cont -> cont -> cont
  | Match : var -> msg -> list formula -> cont -> cont -> cont .

Hint Constructors cont.

  Fixpoint cont_returns c :=
	
  match c with
	
  | Fail => nil
  | Return _ _ => c :: nil
  | Let _ _ c => cont_returns c
  | Derive _ sk fk => (cont_returns sk) ++ (cont_returns fk)
  | Send _ _ _ sk fk => (cont_returns sk) ++ (cont_returns fk)
  | Recv _ _ _ sk fk => (cont_returns sk) ++ (cont_returns fk)
  | Call _ _ _ _ _ sk fk => (cont_returns sk) ++ (cont_returns fk)
  | Match _ _ _ sk fk => (cont_returns sk) ++ (cont_returns fk)
  end.

  Lemma cont_eq_dec: forall x y : cont, {x = y} + {x <> y}.
  Proof.
   decide equality;
   try (apply (list_eq_dec formula_eq_dec));
   try (apply msg_eq_dec);
   try (apply letval_eq_dec);
   try (apply var_eq_dec);
   try (apply const_eq_dec);
   try (apply (list_eq_dec var_eq_dec)).
  Qed.

  Definition return_path c :=
   List.length (cont_returns c) > 0.

Hint Unfold return_path.

  Lemma return_path_dec:
   forall (c:cont), {return_path c} + {~ return_path c}.
  Proof.
   intros c.
   unfold return_path, gt.
   case (length (cont_returns c)).
   right. intro LT.
   apply Arith.Lt.lt_asym with 0 0.
   assumption. assumption.
   intros n.
   left. apply Arith.Lt.lt_O_Sn.
  Qed.

  Inductive proc : Set :=
  | Proc : const -> list var -> list formula -> cont -> proc.

Hint Constructors proc.

  Lemma proc_eq_dec: forall x y : proc, {x = y} + {x <> y}.
  Proof.
   decide equality.
   apply cont_eq_dec.
   apply (list_eq_dec formula_eq_dec).
   apply (list_eq_dec var_eq_dec).
   apply const_eq_dec.
  Qed.

    Definition return_path_proc p :=
      match p with
	| Proc _ _ _ c => return_path c
      end.

(* Type Check *)
    Definition type_letval vs v lv :=
      match lv with
	| New_nonce =>
	  (type_varT v Nonce)
	| New_symkey =>
	  (type_varT v Symkey)
	| New_pubkey =>
	  (type_varT v Pubkey)
	| Accept =>
	  (type_varT v Channel)
	| Connect nv =>
	  (type_varT v Channel)
	  /\ (exist_var vs nv)
	  /\ (type_varT nv Name)
	| LVMsg m =>
	  (type_varT v Msg)
	  /\ (type_msg_send vs m)
      end .

    Lemma type_letval_dec:
     forall (vs:IdSet.t) (v:var) (lv:letval),
     {type_letval vs v lv} + {~ type_letval vs v lv}.
    Proof.
     intros vs v lv.
     destruct lv as [| | | | nv | m ];
     simpl;
     try (apply type_varT_dec).

     unfold exist_var.
     destruct (type_varT_dec v Channel) as [Tc | Fc];
     destruct (In_dec (varid nv) vs) as [TInv | FInv];
     destruct (type_varT_dec nv Name) as [Tnv | Fnv];
     intuition.

     destruct (type_varT_dec v Msg) as [Tv | Fv];
     destruct (type_msg_send_dec vs m) as [Tm | Fm];
     intuition.
    Qed.     

    Lemma type_letval_vars :
      forall (vs:IdSet.t) (v:var) (lv:letval),
	type_letval vs v lv ->
	Subset (letval_vars lv) vs .
    Proof.
      induction lv; simpl; intuition.
      unfold exist_var in H.
      unfold Subset in |- *.
      intros.
      assert (E.eq (varid v0) a).
      apply singleton_1.
      exact H1.
      rewrite <- H3 in |- *.
      exact H.
    Qed.

Hint Resolve type_letval_vars.

  Fixpoint type_cont vs (c:cont) {struct c} : Prop :=
  match c with
  | Fail => True
  | Return f evs =>
    (exist_vars (union vs (formulas_vars_bound f))
     (varlist_set evs))
    /\ (exist_vars vs (formulas_vars_prove f))
  | Let v lv c =>
    (type_letval vs v lv)
    /\ (type_cont (add (varid v) vs) c)
  | Derive f sk fk =>
    let fvs := formulas_vars_bound f in
    (exist_vars vs (formulas_vars_prove f))
    /\ (type_cont (union vs fvs) sk)
    /\ (type_cont vs fk)
  | Send f v m sk fk =>
    let fvs := formulas_vars_bound f in
    (exist_vars vs (formulas_vars_prove f))
    /\ (exist_var (union vs fvs) v)
    /\ (type_varT v Channel)
    /\ (type_msg_send (union vs fvs) m)
    /\ (type_cont (union vs fvs) sk)
    /\ (type_cont vs fk)
  | Recv v m f sk fk =>
    let fvs := formulas_vars_bound f in
	
    let mvs := message_vars_bound m in
    (exist_vars (union vs mvs) (formulas_vars_prove f))
    /\ (exist_var vs v)
    /\ (type_varT v Channel)
    /\ (type_msg_recv vs m)
    /\ (type_cont (union (union vs mvs) fvs) sk)
    /\ (type_cont vs fk)
  | Call bf p svs rvs ef sk fk =>
    let bfvs := formulas_vars_bound bf in
    let efvs := formulas_vars_bound ef in
	
    (exist_vars vs (formulas_vars_prove bf))
    /\ (exist_vars (union vs bfvs) (varlist_set svs))
    /\ (exist_vars (union (union vs bfvs) (varlist_set rvs))
        (formulas_vars_prove ef))
    /\ (type_cont (union (union vs bfvs) 
		  (union (varlist_set rvs) efvs)) sk)
    /\ (type_cont vs fk)
  | Match v m f sk fk =>
    let fvs := formulas_vars_bound f in
	
    let mvs := message_vars_bound m in
    (exist_vars (union vs mvs) (formulas_vars_prove f))
    /\ (exist_var vs v)
    /\ (type_varT v Msg)
    /\ (type_msg_recv vs m)
    /\ (type_cont (union (union vs mvs) fvs) sk)
    /\ (type_cont vs fk)
  end.

  Lemma type_cont_dec:
   forall (vs:t) (c:cont),
    {type_cont vs c} + {~ type_cont vs c}.
  Proof.
   intros vs c. revert c vs.
   simple induction c; clear c; simpl;
   [ idtac
   | intros f evs
   | intros v lv c Hc
   | intros f sk Hsk fk Hfk
   | intros f v m sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk
   | intros bf p svs rvs ef sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk ];
   intros vs;
   unfold exist_var, exist_vars;
   try (intuition; fail);
   try (destruct (Hsk (union vs (formulas_vars_bound f)))
    as [Tsk | Fsk]);
   try (destruct (Hfk vs) as [Tfk | Ffk ]);
   try (intuition; fail);
   try (destruct (In_dec (varid v) vs) as [Inv | NInv]);
   try (destruct (In_dec (varid v)
     (union vs (formulas_vars_bound f))) as [Invb | NInvb]);
   try (destruct (type_msg_send_dec 
        (union vs (formulas_vars_bound f)) m) as [Tsm | Fsm]);
   try (destruct (type_msg_recv_dec vs m) as [Trm | Frm]);
   try (destruct (Hsk (union (union vs (message_vars_bound m))
                       (formulas_vars_bound f)))
                 as [Tsk' | Fsk']);
   try (intuition; fail);
   try (destruct (type_letval_dec vs v lv) as [Tlv | Flv]);
   try (destruct (Hc (add (varid v) vs)) as [Tc | Fc]);
   try (destruct (type_varT_dec v Channel) as [Tvc | Fvc]);
   try (destruct (Subset_dec (varlist_set svs)
         (union vs (formulas_vars_bound bf))) as [Ssvs | NSsvs]);
   try (destruct (Hsk (union (union vs (formulas_vars_bound bf))
    (union (varlist_set rvs) (formulas_vars_bound ef))))
     as [Tsk | Fsk]);
   try (destruct (type_varT_dec v Msg) as [Tvm | Fvm]);
   try (destruct (Subset_dec (varlist_set evs)
        (union vs (formulas_vars_bound f)))
   as [Sevs | NSevs]);
   try (destruct (Subset_dec (formulas_vars_prove f) vs)
    as [Sfvp | NSfvp]);
   try (destruct (Subset_dec (formulas_vars_prove f)
         (union vs (message_vars_bound m))) as [Sfvpb | NSfvpb]);
   try (intuition; fail).

   set (vs' := (union (union vs (formulas_vars_bound bf))
                (varlist_set rvs))).
   destruct (Subset_dec (formulas_vars_prove bf) vs) 
   as [Sbfvp | NSbfvp];
   destruct (Subset_dec (formulas_vars_prove ef) vs') 
   as [Sefvp | NSefvp];
   intuition.
  Qed.

  Definition returns_same c :=
  match cont_returns c with
  | nil => True
  | r1 :: rs =>
    count_occ cont_eq_dec rs r1 = length rs
  end.

  Lemma returns_same_dec:
   forall (c:cont),
    {returns_same c} + {~ returns_same c}.
  Proof.
   intro c.
   unfold returns_same.
   set (rs := cont_returns c).
   destruct rs as [|r1 rs].
   left. trivial.
   set (l := length rs).
   set (occ := count_occ cont_eq_dec rs r1).
   destruct (eq_nat_decide occ l) as [ EQ | NEQ ].
   left. rewrite <- eq_nat_is_eq. assumption.
   right. rewrite <- eq_nat_is_eq. assumption.
  Qed.

    Definition type_proc (p:proc) :=
      match p with
	| Proc _ vs f c =>
	  let fvs := formulas_vars_bound f in
            exist_vars (varlist_set vs) (formulas_vars_prove f)
	    /\ type_cont (union (varlist_set vs) fvs) c
            /\ return_path c
            /\ returns_same c
      end .

  Lemma type_proc_dec:
   forall (p:proc), {type_proc p} + {~ type_proc p}.
  Proof.
   destruct p as [p vs f c].
   simpl. unfold exist_vars.
   destruct (Subset_dec (formulas_vars_prove f) (varlist_set vs))
   as [Sfvp | NSfvp];
   destruct (type_cont_dec 
    (union (varlist_set vs) (formulas_vars_bound f)) c) 
    as [Tc | Fc];
   destruct (return_path_dec c) as [Tr | Fr];
   destruct (returns_same_dec c) as [Ts | Fs];
   intuition.
  Qed.

  Add Morphism type_letval : type_letval_m.
  Proof.
    destruct x0; simpl; unfold exist_var; try (rewrite H);
    try (rewrite <- H); intuition.
  Qed.

  Add Morphism type_cont : type_cont_m.
  Proof.
   intros x1 x2 EQ c. revert c x1 x2 EQ.
   simple induction c; clear c; simpl;
   [ idtac
   | intros f evs
   | intros v lv c Hc
   | intros f sk Hsk fk Hfk
   | intros f v m sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk
   | intros bf p svs rvs ef sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk ];
   intros x1 x2 EQ;
   unfold exist_var, exist_vars.

   split; trivial.

   rewrite EQ. split; trivial.

   rewrite (Hc (add (varid v) x1) (add (varid v) x2)).
   rewrite EQ. reflexivity.
   rewrite EQ. split; trivial.

   rewrite (Hfk x1 x2). assumption.
   rewrite (Hsk (union x1 (formulas_vars_bound f))
                (union x2 (formulas_vars_bound f))).
   rewrite EQ. reflexivity.
   rewrite EQ.
   split; trivial.

   rewrite (Hfk x1 x2). assumption.
   rewrite (Hsk (union x1 (formulas_vars_bound f))
                (union x2 (formulas_vars_bound f))).
   rewrite EQ. reflexivity.
   rewrite EQ. split; trivial.

   rewrite (Hfk x1 x2). assumption.
   rewrite (Hsk (union (union x1 (message_vars_bound m))
                 (formulas_vars_bound f))
                (union (union x2 (message_vars_bound m))
                 (formulas_vars_bound f))).
   rewrite EQ. reflexivity.
   rewrite EQ. split; trivial.

   rewrite (Hfk x1 x2). assumption.
   rewrite (Hsk (union (union x1 (formulas_vars_bound bf))
                 (union (varlist_set rvs) (formulas_vars_bound ef)))
                (union (union x2 (formulas_vars_bound bf))
                 (union (varlist_set rvs) (formulas_vars_bound ef)))).
   rewrite EQ. reflexivity.
   rewrite EQ. split; trivial.

   rewrite (Hfk x1 x2). assumption.
   rewrite (Hsk (union (union x1 (message_vars_bound m))
                 (formulas_vars_bound f))
                (union (union x2 (message_vars_bound m))
                 (formulas_vars_bound f))).
   rewrite EQ. reflexivity.
   rewrite EQ. split; trivial.
  Qed.

  Lemma type_letval_subset:
   forall (vs:t) (v:var) (lv:letval),
	
    type_letval vs v lv ->
	
    forall (vs':t),
     Subset vs vs' ->
     type_letval vs' v lv.
  Proof.
   intros vs v lv.
   destruct lv; simpl; intuition.
   unfold exist_var in *. unfold Subset in *.
   intuition.
   apply type_msg_send_subset with vs;
   assumption.
  Qed.

Hint Resolve type_letval_subset.

  Lemma type_letval_ext:
    forall (vs:IdSet.t) (v:var) (lv:letval),
      type_letval vs v lv ->
      forall (vs':IdSet.t),
	type_letval (union vs vs') v lv.
  Proof.
   intros vs v lv TLV vs'.
   apply type_letval_subset with vs.
   assumption.
   intros a. set_iff. intuition.
  Qed.

Hint Resolve type_letval_ext.

  Lemma type_cont_subset:
    forall (vs:t) (c:cont),
      type_cont vs c ->
      forall (vs':t),
	
        Subset vs vs' ->
	type_cont vs' c.
  Proof.
   intros vs c. revert vs.
   induction c; simpl; unfold exist_var, exist_vars in *;
   intuition.

   apply subset_trans with (union vs (formulas_vars_bound l)).
   assumption. intro. set_iff. intuition.
   apply subset_trans with vs; assumption.

   apply type_letval_subset with vs; assumption.

   apply IHc with (add (varid v) vs). assumption.
   unfold Subset. intro a. set_iff. intuition.

   apply subset_trans with vs; assumption.

   apply IHc1 with (union vs (formulas_vars_bound l)). assumption.
   unfold Subset. intro a. set_iff. intuition.
   apply IHc2 with vs; assumption.

   apply subset_trans with vs; assumption.
   revert H. set_iff. intuition.

   apply type_msg_send_subset
   with (union vs (formulas_vars_bound l)).
   assumption. intro. set_iff. intuition.

   apply IHc1 with (union vs (formulas_vars_bound l)). assumption.
   unfold Subset. intro a. set_iff. intuition.
   apply IHc2 with vs; assumption.

   apply subset_trans with (union vs (message_vars_bound m)).
   assumption. intro. set_iff. intuition.

   apply type_msg_recv_subset with vs; assumption.

   apply IHc1 with (union (union vs (message_vars_bound m))
                    (formulas_vars_bound l)).
   assumption. unfold Subset. intro a. set_iff. intuition.
   apply IHc2 with vs; assumption.

   apply subset_trans with vs; assumption.
   apply subset_trans
    with (union vs (formulas_vars_bound l)).
   assumption.
   unfold Subset. intro a. set_iff. intuition.

   apply subset_trans
   with (union (union vs (formulas_vars_bound l))
          (varlist_set l1)).
   assumption.
   unfold Subset. intro a. set_iff. intuition.

   apply IHc1 with (union (union vs (formulas_vars_bound l)) 
                    (union (varlist_set l1) (formulas_vars_bound l2))).
   assumption. unfold Subset. intro a. set_iff. intuition.
   apply IHc2 with vs; assumption.

   apply subset_trans with (union vs (message_vars_bound m)).
   assumption. intro. set_iff. intuition.

   apply type_msg_recv_subset with vs; assumption.

   apply IHc1 with (union (union vs (message_vars_bound m)) 
                    (formulas_vars_bound l)).
   assumption. unfold Subset. intro a. set_iff. intuition.
   apply IHc2 with vs; assumption.
  Qed.

Hint Resolve type_cont_subset.

  Lemma type_cont_ext:
    forall (vs:IdSet.t) (c:cont),
      type_cont vs c ->
      forall (vs':IdSet.t),
	type_cont (union vs vs') c.
  Proof.
   intros vs c Tc vs'.
   apply type_cont_subset with vs.
   assumption.
   unfold Subset. intro a. set_iff. intuition.
 Qed.

Hint Resolve type_cont_ext.

(* Semantics *)
    Inductive eval_cont : cont -> strand -> t -> Prop :=
      | FAIL :
	forall c, eval_cont c (faildesc :: nil) empty
      | RETURN :
	forall (phi:list formula) (xs:list var),
	  eval_cont
	  (Return phi xs)
	  (((Outgoing, (Join (MConst C.RETURN) 
                         (Join (MVar ai) (vars_msg xs)))), phi)
	    :: nil)
	  empty
      (*| TAIL_CALL :
	forall (sig:sigT) (gamma:gammaT) 
	  (phi psi:list formula) (n:const) (xs ys:list var) (fk:cont),
	  eval_cont sig gamma
	  (Call phi n xs ys psi (Return psi ys) fk)
	  (((Outgoing, (call_msg n xs)), phi) :: nil)
	  empty*)
      | LET_NEW_NONCE :
	forall (s:strand) (v:t) (x:var) (c:cont),
	    (eval_cont c s v) ->
            let s' := (Outgoing, (MConst C.NEW_NONCE), 
                       (Formula C.NEW_NONCE ((LogicV x) :: nil)) :: nil)
              :: s in
	      (eval_cont (Let x New_nonce c) s' (add (varid x) v))
      | LET_NEW_SYMKEY :
	forall (s:strand) (v:t) (x:var) (c:cont),
	    (eval_cont c s v) ->
            let s' := (Outgoing, (MConst C.NEW_SYMKEY),
                       (Formula C.NEW_SYMKEY ((LogicV x) :: nil)) :: nil)
              :: s in
	      (eval_cont (Let x New_symkey c) s' (add (varid x) v))
      | LET_NEW_PUBKEY :
	forall (s:strand) (v:t) (x:var) (c:cont),
	    (eval_cont c s v) ->
            let s' := (Outgoing, (MConst C.NEW_PUBKEY), 
                       (Formula C.NEW_PUBKEY ((LogicV x) :: nil)) :: nil)
              :: s in
	      (eval_cont (Let x New_pubkey c) s' (add (varid x) v))
      | LET_ACCEPT :
	forall (s:strand) (v:t) (x:var) (c:cont),
	    (eval_cont c s v) ->
            let s' := (Outgoing, (MConst C.ACCEPT),
                       (Formula C.ACCEPT ((LogicV x) :: nil)) :: nil)
              :: s in
	      (eval_cont (Let x Accept c) s' v)
      | LET_CONNECT :
	forall (s:strand) (v:t) (x:var) (p:var) (c:cont),
	    (eval_cont c s v) ->
            let s' := (Outgoing, (MConst C.CONNECT),
                       (Formula C.CONNECT ((DataV p) :: (LogicV x) :: nil)) :: nil)
              :: s in
	      (eval_cont (Let x (Connect p) c) s' v)
      | LET_LVMSG :
	forall (s:strand) (v:t) (x:var) (m:msg) (c:cont),
            (eval_cont c s v) ->
            let s' := (Outgoing, (MConst C.BIND),
                       (Formula C.BIND ((LogicV x) :: nil)) :: nil)
              :: strand_replace x m s in                  
              (eval_cont (Let x (LVMsg m) c) s' v)
      | ASSERT :
	forall (s:strand) (v:t) (phi:list formula) (sk fk:cont),
	    eval_cont sk s v ->
	    let s' := ((Outgoing, MConst C.ASSERT), phi) :: s in
	      eval_cont (Derive phi sk fk) s' v
      | ASSERT_ALTERNATIVE:
	forall (s:strand) (v:t) (phi:list formula) (sk fk:cont),
         eval_cont fk s v ->
         eval_cont (Derive phi sk fk) s v
      | SEND_WITH_GUARANTEE :
	forall (s:strand) (v:t) (phi:list formula) (x:var) (m:msg) (sk fk:cont),
	  (eval_cont sk s v) ->
	    let s' := 
	      ((Outgoing,
		(Join (MConst C.MSG)
		  (Join (MVar x) m))), phi) :: s
	      in
	      (eval_cont (Send phi x m sk fk) s' v)
      | SEND_ALTERNATIVE :
	forall (s:strand) (v:t) (phi:list formula) (x:var) (m:msg) (sk fk:cont),
	  (eval_cont fk s v) ->
	  (eval_cont (Send phi x m sk fk) s v)
      | RECEIVE_AND_RELY :
	forall (s:strand) (v:t) (x:var) (m:msg) (psi:list formula) (sk fk:cont),
	  (eval_cont sk s v) ->
	      let s' :=
		((Incoming,
		  (Join (MConst C.MSG)
		    (Join (MVar x) m))), psi) :: s
		in
		(eval_cont (Recv x m psi sk fk) s' v) 
      | RECEIVE_ALTERNATIVE :
	forall (s:strand) (v:t) (x:var) (m:msg) (psi:list formula) (sk fk:cont),
	  (eval_cont fk s v) ->
	  (eval_cont (Recv x m psi sk fk) s v)
      | MATCH_AND_RELY :
	forall (s:strand) (v:t) (x:var) (m:msg) (psi:list formula) (sk fk:cont),
	  (eval_cont sk s v) ->
	      let s' := 
		((Outgoing, (Join (MConst C.MATCH) (MVar x))), nil)
		:: ((Incoming, (Join (MConst C.MATCH) m)), psi)
		:: s
		in
		(eval_cont (Match x m psi sk fk) s' v)
      (* XXX There are many cases where this logically should not fail *)
      (*      For example, x has already been matched, or is Bind x m *)
      | MATCH_ALTERNATIVE :
	forall  (s:strand) (v:t) (x:var) (m:msg) (psi:list formula) (sk fk:cont),
	  (eval_cont fk s v) ->
	  (eval_cont (Match x m psi sk fk) s v)
      | CALL_AND_RELY :
	forall (s:strand) (v:t) (phi:list formula) (n:const) (xs:list var) (ys:list var)
          (psi:list formula) (sk fk:cont),
	  (eval_cont sk s v) ->	
          (* XXX Should be a new ai *)
	  let v' := add (varid ai) v in
	    let s' := 
	      ((Outgoing, (call_msg n xs)), phi) ::
	      ((Incoming,
		(Join (MConst C.RETURN)
		  (Join (MVar ai)
		    (vars_msg ys)))), psi) ::
	      s
	      in		      
	      (eval_cont (Call phi n xs ys psi sk fk) s' v')
      | CALLEE_FAILS :
	forall (s:strand) (v:t)
	  (phi:list formula) (n:const) (xs:list var) (ys:list var) 
          (psi:list formula) (sk fk:cont),
	  (eval_cont fk s v) ->
	    let v' := add (varid ai) v in
	      let s' := 
		((Outgoing, (call_msg n xs)), phi) ::
		((Incoming, (Join (MConst C.FAIL) (MVar ai))), nil) ::
		s
		in		      
		(eval_cont (Call phi n xs ys psi sk fk) s' v')
      | CALL_ALTERNATIVE :
	forall (s:strand) (v:t)
	  (phi:list formula) (n:const) (xs:list var) (ys:list var)
          (psi:list formula) (sk fk:cont),
	  (eval_cont fk s v) ->
	  (eval_cont (Call phi n xs ys psi sk fk) s v) .

Hint Constructors eval_cont.

    Inductive eval_proc : proc -> strand -> t -> Prop :=
    | PROC :
      forall (s:strand) (v:t)
	     (n:const) (psi:list formula) (xs:list var) (c:cont),
      let msg := call_msg n xs in
       (eval_cont c s v) ->
       let s' := ((Incoming, msg), psi) :: s in
	(eval_proc (Proc n xs psi c) s' v) .

Hint Constructors eval_proc.

  Lemma eval_cont_not_nil:
   forall (c:cont) (s:strand) (v:t),
    eval_cont c s v ->
    s <> nil.
  Proof.
   simple induction c; clear c; simpl;
   [ idtac
   | intros f evs
   | intros v lv c Hc
   | intros f sk Hsk fk Hfk
   | intros f v m sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk
   | intros bf p svs rvs ef sk Hsk fk Hfk
   | intros v m f sk Hsk fk Hfk ];
   intros s nv ECc;
   inversion ECc; subst;
   try (discriminate);
   try (apply Hfk with nv; assumption).
  Qed.

Hint Resolve eval_cont_not_nil.

Lemma eval_cont_total : 
  forall (c:cont),
    exists s:strand, exists v:t, 
      eval_cont c s v.
Proof.
  intros.
  exists (faildesc :: nil).
  exists empty.
  apply FAIL.
Qed.

Hint Resolve eval_cont_total.

Lemma eval_cont_prog :
  forall (c:cont),
    return_path c ->
    exists s:strand, exists v:t,
      (eval_cont c s v)
      /\ ((List.last s faildesc) <> faildesc).
Proof.
  unfold return_path.
  induction c; simpl; intros Rc; eauto.

  assert (False). apply Arith.Gt.gt_asym with 0 0.
  assumption. assumption.
  contradiction.

  exists 
    ((((Outgoing, (Join (MConst C.RETURN) (Join (MVar ai)
      (vars_msg l0)))), l) :: nil)).
  exists empty. 
  split. apply RETURN.
  simpl. discriminate.

  destruct (IHc Rc) as [s [nu [EC LEQ]]]. clear IHc Rc.
  destruct l.

  exists ((Outgoing, (MConst C.NEW_NONCE), 
           (Formula C.NEW_NONCE ((LogicV v) :: nil)) :: nil) :: s).
  exists (add (varid v) nu).
  simpl. intuition.
  apply LEQ. destruct s; intuition.

  exists ((Outgoing, (MConst C.NEW_SYMKEY),
    (Formula C.NEW_SYMKEY ((LogicV v) :: nil))::nil) :: s).
  exists (add (varid v) nu).
  simpl. intuition.
  apply LEQ. destruct s; intuition.

  exists ((Outgoing, (MConst C.NEW_PUBKEY),
    (Formula C.NEW_PUBKEY ((LogicV v) :: nil))::nil) :: s).
  exists (add (varid v) nu).
  simpl. intuition.
  apply LEQ. destruct s; intuition.

  exists ((Outgoing, (MConst C.ACCEPT),
    (Formula C.ACCEPT ((LogicV v) :: nil))::nil) :: s).
  exists nu.
  simpl. intuition.
  apply LEQ. destruct s; intuition.

  exists ((Outgoing, (MConst C.CONNECT), 
    (Formula C.CONNECT ((DataV v0) :: (LogicV v) :: nil))
    ::nil) :: s).
  exists nu.
  simpl. intuition.
  apply LEQ. destruct s; intuition.

  exists ((Outgoing, (MConst C.BIND),
    (Formula C.BIND ((LogicV v) :: nil))::nil) 
  :: (strand_replace v m s)).
  exists nu.
  simpl. intuition.
  generalize (strand_replace_faildesc s LEQ v m).
  destruct (strand_replace v m s); intuition.

  rewrite app_length in Rc.
  set (cr1 := length (cont_returns c1)).
  set (cr2 := length (cont_returns c2)).
  destruct (plus_gt_or cr1 cr2 Rc) as [Rc1 | Rc2].
  destruct (IHc1 Rc1) as [s [nu [EC LEQ]]].
  clear IHc1 Rc1.
  exists (((Outgoing, MConst C.ASSERT), l) :: s).
  exists nu. simpl.
  intuition.
  apply LEQ. destruct s; intuition.

  destruct (IHc2 Rc2) as [s [nu [EC LEQ]]].
  clear IHc2 Rc2.
  exists s.
  exists nu.
  intuition.

  rewrite app_length in Rc.
  set (cr1 := length (cont_returns c1)).
  set (cr2 := length (cont_returns c2)).
  destruct (plus_gt_or cr1 cr2 Rc) as [Rc1 | Rc2].
  destruct (IHc1 Rc1) as [s [nu [EC LEQ]]].
  clear IHc1 Rc1.
  exists (((Outgoing,
    (Join (MConst C.MSG)
      (Join (MVar v) m))), l) :: s).
  exists nu. simpl.
  intuition.
  apply LEQ. destruct s; intuition.

  destruct (IHc2 Rc2) as [s [nu [EC LEQ]]].
  clear IHc2 Rc2.
  exists s.
  exists nu.
  intuition.

  rewrite app_length in Rc.
  set (cr1 := length (cont_returns c1)).
  set (cr2 := length (cont_returns c2)).
  destruct (plus_gt_or cr1 cr2 Rc) as [Rc1 | Rc2].
  destruct (IHc1 Rc1) as [s [nu [EC LEQ]]].
  clear IHc1 Rc1.
  exists (((Incoming,
    (Join (MConst C.MSG)
      (Join (MVar v) m))), l) :: s).
  exists nu. simpl.
  intuition.
  apply LEQ. destruct s; intuition.

  destruct (IHc2 Rc2) as [s [nu [EC LEQ]]].
  clear IHc2 Rc2.
  exists s.
  exists nu.
  intuition.

  rename c1 into c0. rename c2 into c1. rename c3 into c2.
  rewrite app_length in Rc.
  set (cr1 := length (cont_returns c1)).
  set (cr2 := length (cont_returns c2)).
  destruct (plus_gt_or cr1 cr2 Rc) as [Rc1 | Rc2].
  destruct (IHc1 Rc1) as [s [nu [EC LEQ]]].
  clear IHc1 Rc1.
  exists (((Outgoing, (call_msg c0 l0)), l) ::
    ((Incoming, 
      (Join (MConst C.RETURN)
        (Join (MVar ai) (vars_msg l1)))), l2) ::
    s).
  exists (add (varid ai) nu).
  simpl. intuition.
  apply LEQ. destruct s; intuition.

  destruct (IHc2 Rc2) as [s [nu [EC LEQ]]].
  clear IHc2 Rc2.
  exists s.
  exists nu.
  intuition.

  rewrite app_length in Rc.
  set (cr1 := length (cont_returns c1)).
  set (cr2 := length (cont_returns c2)).
  destruct (plus_gt_or cr1 cr2 Rc) as [Rc1 | Rc2].
  destruct (IHc1 Rc1) as [s [nu [EC LEQ]]].
  clear IHc1 Rc1.
  exists (((Outgoing,
    (Join (MConst C.MATCH)
      (MVar v))), nil) ::
  (Incoming,
    (Join (MConst C.MATCH)
      m), l) :: s).
  exists nu.
  simpl. intuition.
  apply LEQ. destruct s; intuition.

  destruct (IHc2 Rc2) as [s [nu [EC LEQ]]].
  clear IHc2 Rc2.
  exists s.
  exists nu.
  intuition.
Qed.

Hint Resolve eval_cont_prog.

Lemma eval_cont_type :
  forall (sig:t) (c:cont) (s:strand) (v:t),
    In (varid pr) sig ->
    In (varid ai) sig ->
    type_cont sig c ->
    eval_cont c s v ->
    type_strand sig s .
Proof.
  intros sig c s v. revert c sig s v.
  simple induction c; clear c; simpl;
    [ idtac
      | intros phi xs
      | intros vr lv c IHc
      | intros phi sk IHsk fk IHfk
      | intros phi vr m sk IHsk fk IHfk
      | intros vr m psi sk IHsk fk IHfk
      | intros phi n svs rvs psi sk IHsk fk IHfk
      | intros vr m psi sk IHsk fk IHfk];
    intros sig s v InPR InAI TC EC;
      inversion EC; subst; simpl;
        unfold exist_var, exist_vars in *; set_iff;
          intuition;
            try (apply IHfk with v; assumption).
  
  rewrite type_msg_send_vars_msg. assumption.

  apply IHc with v0.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset with (add (varid vr) sig).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  assumption.

  apply IHc with v0.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset with (add (varid vr) sig).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  assumption.

  apply IHc with v0.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset with (add (varid vr) sig).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  assumption.

  apply IHc with v.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset with (add (varid vr) sig).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  unfold E.eq, IOT.eq in *. subst a.
  simpl in H. unfold exist_var in H. intuition.

  apply type_strand_subset with (add (varid vr) sig).
  unfold Subset. intro a. set_iff. intuition.
  apply IHc with v.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset with (add (varid vr) sig).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  assumption.

  simpl in H. decompose [and] H.
  apply strand_replace_type.
  apply type_msg_send_recv.
  apply type_msg_send_subset with sig. assumption.     
  unfold Subset. intro a. set_iff. intuition.
  apply type_msg_send_subset with sig. assumption.     
  unfold Subset. intro a. set_iff. intuition.
  apply IHc with v.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset with (add (varid vr) sig).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  assumption. assumption.

  apply IHsk with v.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset 
    with (union sig (formulas_vars_bound phi)).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  assumption.

  revert H2. set_iff. intuition.

  apply IHsk with v.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset
    with (union sig (formulas_vars_bound phi)).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  assumption.

  apply subset_trans with
    (union sig (message_vars_bound m)).
  assumption. intro. set_iff. intuition.

  apply IHsk with v.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset
    with (union (union sig (message_vars_bound m)) 
      (formulas_vars_bound psi)).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  assumption.

  rewrite type_msg_send_vars_msg; assumption.

  apply subset_trans with
    (union (union sig (formulas_vars_bound phi))
      (varlist_set rvs)).
  assumption. rewrite vars_msg_bound.
  intro. set_iff. intuition.

  apply IHsk with v0.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset
    with (union (union sig (formulas_vars_bound phi))
      (union (varlist_set rvs) (formulas_vars_bound psi))).
  assumption.
  rewrite vars_msg_bound.
  unfold Subset. intro a. set_iff. intuition.
  assumption.

  rewrite type_msg_send_vars_msg; assumption.

  apply type_strand_subset with sig.
  unfold Subset. intro a. set_iff. intuition.
  apply IHfk with v0.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset with sig.
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  assumption.

  apply subset_trans with (union sig (message_vars_bound m)).
  assumption. intro a. set_iff. intuition.

  apply IHsk with v.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset
    with (union (union sig (message_vars_bound m)) 
      (formulas_vars_bound psi)).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  assumption.
Qed.

Hint Resolve eval_cont_type.

Lemma eval_proc_total : 
  forall (p:proc),
    exists s:strand, exists v:t, 
      eval_proc p s v .
Proof.
  intros.
  elim p.
  intros.
  destruct (eval_cont_total c0) as [s [v EC]].
  exists ((Incoming, call_msg c l, l0) :: s).
  exists v.
  intuition.
Qed.

Hint Resolve eval_proc_total.

Lemma eval_proc_prog :
  forall (p:proc),
    return_path_proc p ->
    exists s:strand, exists v:t,
      (eval_proc p s v)
      /\ ((List.last s faildesc) <> faildesc).
Proof.
  destruct p as [r xs phi c].
  simpl. intros Rc.
  destruct (eval_cont_prog c Rc) as [s [v [Ec Leq]]].
  exists ((Incoming, call_msg r xs, phi) :: s).
  exists v.
  intuition.
  apply Leq. destruct s; intuition.
Qed.

Hint Resolve eval_proc_prog.

Lemma eval_proc_type :
  forall (p:proc),
    type_proc p ->
    forall (s:strand) (v:t),
      eval_proc p s v ->
      forall sig,
	type_strand sig s.
Proof.
  intros p TP s v EP sig.
  destruct p as [n vs fs c].
  simpl in TP. unfold exist_vars in *.
  destruct TP as [SUB [TC [RP RS]]].
  inversion EP.
  subst s' v0 c0.
  subst fs s vs.
  simpl. intuition.
     
  rewrite vars_msg_bound. 
  apply subset_trans with (varlist_set xs).
  assumption. intro. set_iff. intuition.

  apply eval_cont_type with c v.
  set_iff. intuition.
  set_iff. intuition.
  apply type_cont_subset with
      (union (varlist_set xs) (formulas_vars_bound psi)).
  assumption.
  intros a. rewrite vars_msg_bound. set_iff. intuition.
  assumption.
Qed.

Hint Resolve eval_proc_type.
