(* Common to CPPL and WPPL *)
Require Export IdSet.
Open Scope list_scope .
Require Constants.
Module C := Constants.

(* Types *)
Definition const := C.t.
Lemma const_eq_dec : forall x y : const, {x = y} + {x <> y}.
Proof.
  apply C.eq_dec.
Qed.

Definition id := E.t.
Lemma id_eq_dec : forall x y : id, {x = y} + {x <> y}.
Proof.
  apply ME.eq_dec.
Qed.

Inductive type : Set :=
| Text : type
| Msg : type
| Nonce : type
| Channel : type
| Name : type
| Symkey : type
| Pubkey : type.

Hint Constructors type.

Lemma type_eq_dec : forall x y : type, {x = y} + {x <> y} .
Proof.
  decide equality.
Qed.

Definition type_lt t1 t2 :=
  if type_eq_dec t1 t2 then
    False
    else
      match (t1,t2) with
        | (_,Text) => True
        | (Text,_) => False
        | (_,Msg) => True
        | (Msg,_) => False
        | (_,Nonce) => True
        | (Nonce,_) => False
        | (_,Channel) => True
        | (Channel,_) => False
        | (_,Name) => True
        | (Name,_) => False
        | (_,Symkey) => True
        | (Symkey,_) => False
        | (Pubkey,Pubkey) => False
      end.

Lemma type_lt_trans :
  forall x y z : type, 
    type_lt x y -> type_lt y z -> type_lt x z.
Proof.
  intros x y z.
  unfold type_lt.
  case (type_eq_dec x y); intros Cxy;
    case (type_eq_dec y z); intros Cyz;
      case (type_eq_dec x z); intros Cxz;
        try (trivial);
          try (intros; contradiction);
            subst.

  destruct y; destruct z; intuition.
  destruct x; destruct y; destruct z; intuition.
Qed.

Hint Resolve type_lt_trans.

Lemma type_lt_antirefl:
  forall x : type, ~ type_lt x x.
Proof.
  intros x.
  unfold type_lt.
  case (type_eq_dec x x).
  intro EQ. intro F. contradiction.
  intro NEQ. absurd_hyp NEQ.
  intro NEQ'. apply NEQ. reflexivity.
  assumption.
Qed.

Hint Resolve type_lt_antirefl.

Lemma type_lt_dec:
  forall x y : type, {type_lt x y} + {~ type_lt x y}.
Proof.
  intros x y.
  unfold type_lt.
  case (type_eq_dec x y); intro Cxy.
  right. intro. contradiction.
  destruct x; destruct y; intuition.
Qed.

Lemma type_lt_eq_dec:
  forall x y : type, x <> y -> {type_lt x y} + {type_lt y x}.
Proof.
  intros x y NEQ.
  unfold type_lt.
  case (type_eq_dec x y); intro Cxy.
  contradiction.
  case (type_eq_dec y x); intros Cyx.
  subst y.
  absurd_hyp Cxy.
  intro F. apply F. reflexivity.
  assumption.
  destruct x; destruct y; intuition.
Qed.

Inductive var : Set :=
| Var : id -> type -> var .

Hint Constructors var.

Lemma var_eq_dec : forall x y : var, {x = y} + {x <> y}.
Proof. 
  decide equality; [apply type_eq_dec | apply id_eq_dec].
Qed.

Definition varid v := match v with Var i t => i end .

Hint Unfold varid.

Definition VarListSet vs s :=
  forall a:id, In a s <-> exists t:type, List.In (Var a t) vs.

Hint Unfold VarListSet.

Definition SmallestSetWithList P s vs :=
  (SmallestSet P s) /\ (VarListSet vs s).

Hint Unfold SmallestSetWithList.

Definition SmallestSetWithListEx P :=
  { snvs:(t * list var) |
    let (s,vs) := snvs in
      SmallestSetWithList P s vs }.

Hint Unfold SmallestSetWithListEx.

Inductive fvar : Set :=
| DataV : var -> fvar
| LogicV : var -> fvar.

Hint Constructors fvar.

Lemma fvar_eq_dec :
  forall x y : fvar, {x = y} + {x <> y}.
Proof.
  decide equality; apply var_eq_dec.
Qed.

Inductive formula : Set :=
| Formula : const -> list fvar -> formula .

Hint Constructors formula.

Lemma formula_eq_dec : forall x y : formula, {x = y} + {x <> y}.
Proof. 
  decide equality. apply (list_eq_dec fvar_eq_dec). apply const_eq_dec.
Qed.

Inductive msg : Set :=
| MNil : msg
| MVar : var -> msg
| MConst : const -> msg
| Join : msg -> msg -> msg 
| Hash : msg -> msg
| SymEnc : var -> msg -> msg
| SymSign : var -> msg -> msg
| PubEnc : var -> msg -> msg
| PubSign : var -> msg -> msg
| Bind : var -> msg -> msg .

Hint Constructors msg.

Lemma msg_eq_dec : forall x y : msg, {x = y} + {x <> y}.
Proof.
  decide equality; try(apply var_eq_dec); try(apply const_eq_dec).
Qed.

(* var list *)
Fixpoint varlist_set l :=
  match l with
    | nil => empty
    | v :: l =>
      add (varid v) (varlist_set l)
  end.

Lemma varlist_set_unionapp :
  forall (l1 l2:list var),
    Equal (union (varlist_set l1) (varlist_set l2))
    (varlist_set (l1 ++ l2)).
Proof.
  induction l1.
  
  induction l2; simpl.
  unfold Equal; intro; set_iff; intuition.
  simpl in IHl2. apply empty_union_1.
  apply empty_is_empty_2. apply equal_refl.

  simpl. intro l2.
  rewrite union_add.
  rewrite IHl1. apply equal_refl.
Qed.

Hint Resolve varlist_set_unionapp.

  Lemma varlist_set_in:
   forall (vs:list var),
    VarListSet vs (varlist_set vs).
  Proof.
   intros vs vi.
   induction vs as [|v vs]; simpl.
   set_iff. intuition. destruct H as [x F]. contradiction.
   destruct v as [vi' vt']. simpl. set_iff.
   rewrite IHvs. clear IHvs.
   unfold E.eq, IOT.eq.
   split. 
   intros [EQ | [vt IN]]. subst vi.
   exists vt'. left. reflexivity.
   exists vt. right. assumption.
   intros [vt [EQ | IN]].
   inversion EQ. left. reflexivity.
   right. exists vt. assumption.
  Qed.

Hint Resolve varlist_set_in.

  Lemma varlistset_eq:
   forall (vs:list var) (s:t),
    VarListSet vs s <-> Equal s (varlist_set vs).
  Proof.
   intros vs s. unfold VarListSet. split.
   intros VLS a. rewrite VLS. clear VLS s.
   induction vs as [|[vi vt] vs]; simpl.
   split. intros [vt' F]. contradiction.
   set_iff. intros. contradiction.
   set_iff. rewrite <- IHvs.
   unfold E.eq, IOT.eq.
   split. intros [vt' [EQ | IN]].
   inversion EQ. left. reflexivity.
   right. exists vt'. assumption.
   intros [EQ | [vt' IN]].
   subst a. exists vt. left. reflexivity.
   exists vt'. right. assumption.

   intros EQ a. rewrite EQ. clear EQ.
   induction vs as [|[vi vt] vs]; simpl.
   set_iff. split. intros. contradiction.
   intros [vt F]. contradiction.
   set_iff. rewrite IHvs. clear IHvs.
   unfold E.eq, IOT.eq.
   intuition. subst a. exists vt. left. reflexivity.
   destruct H0 as [vt' IN]. exists vt'. right. assumption.
   destruct H as [vt' [EQ | IN]].
   inversion EQ. left. reflexivity.
   right. exists vt'. assumption.
  Qed.

Hint Rewrite -> varlistset_eq : ppl.

(* Patterns *)
Definition formula_dvars f :=
  match f with
    | Formula _ fvs =>
      List.filter
       (fun fv => match fv with DataV v => true | _ => false end)
       fvs
  end.

Definition formula_lvars f :=
  match f with
    | Formula _ fvs =>
      List.filter
       (fun fv => match fv with LogicV v => true | _ => false end)
       fvs
  end.

Definition fvars fvs :=
  List.map
   (fun fv => match fv with DataV v => v | LogicV v => v end)
   fvs.

Definition formula_vars_prove f :=
  match f with
    | Formula _ fvs =>
      varlist_set (fvars (formula_dvars f))
  end.

Definition formula_vars_bound f :=
  match f with
    | Formula _ fvs =>
      varlist_set (fvars (formula_lvars f))
  end.

Definition formula_vars_proveL f :=
  match f with
    | Formula _ fvs => (fvars (formula_dvars f))
  end.

Lemma formula_vars_to_proveL:
  forall (f:formula) (vi:id),
    In vi (formula_vars_prove f) ->
    exists vt:type,
      List.In (Var vi vt) (formula_vars_proveL f).
Proof.
  intros f vi. destruct f as [n fvs].
  simpl.
  induction fvs as [|fv fvs]; simpl.
  set_iff. intros. contradiction.
  destruct fv; simpl.
  set_iff. intros [EQ | InR].
  unfold E.eq, IOT.eq in EQ.
  subst vi. destruct v as [vi vt].
  exists vt. left. reflexivity.
  destruct (IHfvs InR) as [vt IN].
  exists vt. right. assumption.
  apply IHfvs.
Qed.

Hint Resolve formula_vars_to_proveL.

Definition formula_vars_boundL f :=
  match f with
    | Formula _ fvs => (fvars (formula_lvars f))
  end.

Definition formulas_vars_prove fs :=
  List.fold_right (fun e a => union (formula_vars_prove e) a)
  empty fs.

Definition formulas_vars_proveL fs :=
  List.fold_right (fun e a => (formula_vars_proveL e) ++ a) nil fs.


Lemma formulas_vars_to_proveL:
  forall (fs:list formula) (vi:id),
    In vi (formulas_vars_prove fs) ->
    exists vt:type,
      List.In (Var vi vt) (formulas_vars_proveL fs).
Proof.
  intros fs vi.
  induction fs as [|f fs]; simpl.
  set_iff. intros. contradiction.
  set_iff. intros [InF | InR].
  destruct (formula_vars_to_proveL f vi InF) as [vt IN].
  exists vt. apply in_or_app. left. assumption.
  destruct (IHfs InR) as [vt IN].
  exists vt. apply in_or_app. right. assumption.
Qed.

Hint Resolve formulas_vars_to_proveL.

Lemma formula_vars_proveL_equiv:
  forall (f:formula) (vi:id),
    In vi (formula_vars_prove f) <->
    exists vt:type,
      List.In (Var vi vt) (formula_vars_proveL f).
Proof.
  intros f vi. split.
  apply formula_vars_to_proveL.
  intros [vt InFVP]. revert InFVP.
  destruct f as [n vs]. simpl.
  induction vs as [|v vs]; simpl.
  intros. contradiction.
  destruct v as [v | v]; simpl.
  set_iff. unfold E.eq, IOT.eq.
  intros [EQ | IN].
  left. destruct v as [vi' vt'].
  inversion EQ. simpl. reflexivity.
  right. apply IHvs. assumption.
  apply IHvs.
Qed.

Hint Rewrite <- formula_vars_proveL_equiv : ppl.

Lemma formulas_vars_proveL_equiv:
  forall (fs:list formula) (vi:id),
    In vi (formulas_vars_prove fs) <->
    exists vt:type,
      List.In (Var vi vt) (formulas_vars_proveL fs).
Proof.
  intros fs vi.
  split. apply formulas_vars_to_proveL.
  intros [vt InFVP]. revert InFVP.
  induction fs as [|f fs]; simpl.
  intros. contradiction.
  set_iff. rewrite <- in_app_iff.
  intros [InF | InR].
  left. rewrite formula_vars_proveL_equiv.
  exists vt. assumption.
  right. apply IHfs. assumption.
Qed.

Hint Rewrite <- formulas_vars_proveL_equiv : ppl.

Definition formulas_vars_bound fs :=
  List.fold_right (fun e a => union (formula_vars_bound e) a)
  empty fs.

Definition formulas_vars_boundL fs :=
  List.fold_right (fun e a => (formula_vars_boundL e) ++ a) nil fs.

(* The variables needed to construct m *)
Fixpoint message_vars m :=
  match m with
    | MNil => empty
    | MVar v => singleton (varid v)
    | MConst _ => empty
    | Join m1 m2 =>
      union (message_vars m1) (message_vars m2)
    | Hash m => message_vars m
    | SymEnc v m =>
      add (varid v) (message_vars m)
    | SymSign v m =>
      add (varid v) (message_vars m)
    | PubEnc v m =>
      add (varid v) (message_vars m)
    | PubSign v m =>
      add (varid v) (message_vars m)
    | Bind v m =>	    	  	    
      message_vars m
  end .
Fixpoint message_varsL m :=
  match m with
    | MNil => nil
    | MVar v => v :: nil
    | MConst _ => nil
    | Join m1 m2 =>
      (message_varsL m1) ++ (message_varsL m2)
    | Hash m => message_varsL m
    | SymEnc v m =>
      v :: (message_varsL m)
    | SymSign v m =>
      v :: (message_varsL m)
    | PubEnc v m =>
      v :: (message_varsL m)
    | PubSign v m =>
      v :: (message_varsL m)
    | Bind v m =>	    	  	    
      message_varsL m
  end.

Lemma message_to_varsL:
  forall (m:msg) (vi:id),
    In vi (message_vars m) ->
    exists vt:type,
      List.In (Var vi vt) (message_varsL m).
Proof.
  induction m; simpl;
    intros vi; set_iff;
      unfold E.eq, IOT.eq.

  intro F. contradiction.

  destruct v as [ vi' vt ].
  simpl. intro EQ. subst vi'.
  exists vt. left. reflexivity.

  intro F. contradiction.

  intros [ InM1 | InM2 ].
  destruct (IHm1 vi InM1) as [ vt INM1 ].
  exists vt.
  apply in_or_app. left. assumption.
  destruct (IHm2 vi InM2) as [ vt INM2 ].
  exists vt.
  apply in_or_app. right. assumption.

  intro IN. apply IHm. assumption.
  
  intros [ EQv | INm ];
    [ destruct v as [ vi' vt ];
      simpl in EQv; subst vi';
        exists vt; left; reflexivity
      | destruct (IHm vi INm) as [ vt INmL ];
        exists vt; right; assumption ].

  intros [ EQv | INm ];
    [ destruct v as [ vi' vt ];
      simpl in EQv; subst vi';
        exists vt; left; reflexivity
      | destruct (IHm vi INm) as [ vt INmL ];
        exists vt; right; assumption ].

  intros [ EQv | INm ];
    [ destruct v as [ vi' vt ];
      simpl in EQv; subst vi';
        exists vt; left; reflexivity
      | destruct (IHm vi INm) as [ vt INmL ];
        exists vt; right; assumption ].

  intros [ EQv | INm ];
    [ destruct v as [ vi' vt ];
      simpl in EQv; subst vi';
        exists vt; left; reflexivity
      | destruct (IHm vi INm) as [ vt INmL ];
        exists vt; right; assumption ].

  apply IHm.
Qed.  

Hint Resolve message_to_varsL.

(* The variables bound by m if match'd *)
Fixpoint message_vars_bound m :=
  match m with
    | MNil => empty
    | MVar v => singleton (varid v)
    | MConst _ => empty
    | Join m1 m2 =>
      union (message_vars_bound m1) (message_vars_bound m2)
    | Hash m =>
      empty
    | SymEnc v m =>
      message_vars_bound m
    | SymSign v m =>
      message_vars_bound m
    | PubEnc v m =>
      message_vars_bound m
    | PubSign v m =>
      message_vars_bound m	    
    | Bind v m =>	    
      add (varid v) (message_vars_bound m)
  end .  

(* The variables need to match m *)
Fixpoint message_vars_match m :=
  match m with
    | MNil => empty
    | MVar v => empty
    | MConst _ => empty
    | Join m1 m2 =>
      union (message_vars_match m1) (message_vars_match m2)
    | Hash m => message_vars m
    | SymEnc v m =>
      add (varid v) (message_vars_match m)
    | SymSign v m =>
      add (varid v) (message_vars_match m)
    | PubEnc v m =>
      add (varid v) (message_vars_match m)
    | PubSign v m =>
      add (varid v) (message_vars_match m)
    | Bind v m =>
      message_vars_match m
  end .
Fixpoint message_vars_matchL m :=
  match m with
    | MNil => nil
    | MVar _ => nil
    | MConst _ => nil
    | Join m1 m2 =>
      (message_vars_matchL m1) ++ (message_vars_matchL m2)
    | Hash m => message_varsL m
    | SymEnc v m =>
      v :: (message_vars_matchL m)
    | SymSign v m =>
      v :: (message_vars_matchL m)
    | PubEnc v m =>
      v :: (message_vars_matchL m)
    | PubSign v m =>
      v :: (message_vars_matchL m)
    | Bind v m =>
      message_vars_matchL m
  end .

Lemma message_vars_match_subset :  
  forall (m:msg),	  
    Subset (message_vars_match m) (message_vars m).
Proof.
  unfold Subset. induction m; simpl; intro; set_iff; intuition.
Qed.

Hint Resolve message_vars_match_subset.

Lemma message_to_match_varsL:
  forall (m:msg) (vi:id),
    In vi (message_vars_match m) ->
    exists vt:type,
      List.In (Var vi vt) (message_vars_matchL m).
Proof.
  induction m; simpl;
    intros vi; set_iff;
      unfold E.eq, IOT.eq.

  intro F. contradiction.
  intro F. contradiction.
  intro F. contradiction.

  intros [ InM1 | InM2 ].
  destruct (IHm1 vi InM1) as [ vt INM1 ].
  exists vt.
  apply in_or_app. left. assumption.
  destruct (IHm2 vi InM2) as [ vt INM2 ].
  exists vt.
  apply in_or_app. right. assumption.

  apply message_to_varsL.
  
  intros [ EQv | INm ];
    [ destruct v as [ vi' vt ];
      simpl in EQv; subst vi';
        exists vt; left; reflexivity
      | destruct (IHm vi INm) as [ vt INmL ];
        exists vt; right; assumption ].

  intros [ EQv | INm ];
    [ destruct v as [ vi' vt ];
      simpl in EQv; subst vi';
        exists vt; left; reflexivity
      | destruct (IHm vi INm) as [ vt INmL ];
        exists vt; right; assumption ].

  intros [ EQv | INm ];
    [ destruct v as [ vi' vt ];
      simpl in EQv; subst vi';
        exists vt; left; reflexivity
      | destruct (IHm vi INm) as [ vt INmL ];
        exists vt; right; assumption ].

  intros [ EQv | INm ];
    [ destruct v as [ vi' vt ];
      simpl in EQv; subst vi';
        exists vt; left; reflexivity
      | destruct (IHm vi INm) as [ vt INmL ];
        exists vt; right; assumption ].

  apply IHm.
Qed.  

Hint Resolve message_to_match_varsL.

Lemma message_match_in_varsL:
  forall (m:msg) (v:var),
    List.In v (message_vars_matchL m) ->
    List.In v (message_varsL m).
Proof.
  simple induction m; clear m;
    [ idtac
      | intros v
      | intros c
      | intros m1 IHm1 m2 IHm2
      | intros m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1 ];
    simpl;
      intros u;
        try (intros; contradiction);
          try (trivial);
            try (apply IHm1);
              try (
                intros [ EQv | InMML ];
                  [ left; assumption
                    | right; apply IHm1; assumption ]).

  intros InMS.
  cut (List.In u (message_vars_matchL m1) \/ List.In u (message_vars_matchL m2)).
  intros [ InM1 | InM2 ];
    apply in_or_app.
  left. apply IHm1. assumption.
  right. apply IHm2. assumption.
  apply in_app_or. assumption.
Qed.

Hint Resolve message_match_in_varsL.

Fixpoint vars_msg vs :=
  match vs with
    | nil =>
      MNil
    | v0 :: vs' =>
      Join (MVar v0) (vars_msg vs')
  end .

Lemma vars_msg_match:
  forall (l:list var),
    Equal (message_vars_match (vars_msg l)) empty.
Proof.
  simple induction l. simpl.
  unfold Equal. intuition.
  simpl. intros. rewrite H.
  unfold Equal. intro. rewrite union_iff. intuition.
Qed.

Hint Resolve vars_msg_match.
Hint Rewrite -> vars_msg_match : ppl.

Lemma vars_msg_bound:
  forall (l:list var),	  
    Equal (message_vars_bound (vars_msg l)) (varlist_set l).
Proof.
  simple induction l; simpl.
  apply equal_refl.
  intros. rewrite H.
  unfold Equal; intro; set_iff; intuition.
Qed.

Hint Resolve vars_msg_match.
Hint Rewrite -> vars_msg_bound : ppl.

(* Type Check *)
Definition exist_var vs (v:var) := In (varid v) vs.
Definition exist_vars vs evs := Subset evs vs.

Hint Unfold exist_var.
Hint Unfold exist_vars.

Definition type_varT nv T :=
  match nv with
    | Var nvi nvt =>
      if type_eq_dec nvt T then
	True
        else
	  False
  end.

Lemma type_varT_dec:
  forall (nv:var) (T:type),
    {type_varT nv T} + {~ type_varT nv T}.
Proof.
  intros [vi vt] T.
  simpl.
  case (type_eq_dec vt T);
    intuition.
Qed.

Fixpoint type_msg_send vs (m:msg) {struct m} : Prop :=
  match m with
    | MNil =>
      True
    | MVar v =>
      exist_var vs v
    | MConst c =>
      True
    | Join m0 m1 =>
      (type_msg_send vs m0) /\ (type_msg_send vs m1)
    | Hash m =>
      type_msg_send vs m
    | SymEnc v m =>
      (type_varT v Symkey) /\ (exist_var vs v) /\ (type_msg_send vs m)
    | SymSign v m =>
      (type_varT v Symkey) /\ (exist_var vs v) /\ (type_msg_send vs m)
    | PubEnc v m =>
      (type_varT v Pubkey) /\ (exist_var vs v) /\ (type_msg_send vs m)
    | PubSign v m =>
      (type_varT v Pubkey) /\ (exist_var vs v) /\ (type_msg_send vs m)
    | Bind v m =>
      (type_varT v Msg) /\ (type_msg_send vs m)
  end .

Lemma type_msg_send_dec :
  forall (vs:IdSet.t) (m:msg),
    {type_msg_send vs m} + {~ type_msg_send vs m}.
Proof.
  intros vs m. generalize vs. clear vs.
  induction m; intro vs; simpl;
    unfold exist_var;
      [ left; trivial
        | apply In_dec
        | left; trivial
        | destruct (IHm1 vs); clear IHm1;
          destruct (IHm2 vs); clear IHm2;
            [ left | right | right | right];
            intuition
        | apply IHm
        | idtac
        | idtac
        | idtac
        | idtac
        | idtac ];
      destruct v; simpl;
        [ destruct (type_eq_dec t0 Symkey); case (In_dec i vs); intros InH
          | destruct (type_eq_dec t0 Symkey); case (In_dec i vs); intros InH
          | destruct (type_eq_dec t0 Pubkey); case (In_dec i vs); intros InH
          | destruct (type_eq_dec t0 Pubkey); case (In_dec i vs); intros InH
          | destruct (type_eq_dec t0 Msg)];
        case (IHm vs); clear IHm; intro IHmH;
          intuition.
Qed.       

Lemma type_msg_send_subset:
  forall (vs:IdSet.t) (m:msg),
    type_msg_send vs m ->
    forall (vs':IdSet.t),      
      Subset vs vs' ->
      type_msg_send vs' m .
Proof.
  unfold Subset.
  induction m; simpl; unfold exist_var, exist_vars; set_iff; intuition.
Qed.        

Hint Resolve type_msg_send_subset.

Lemma type_msg_send_ext :
  forall (vs:IdSet.t) (m:msg),
    type_msg_send vs m ->
    forall (vs':IdSet.t),
      type_msg_send (union vs vs') m .
Proof.
  intros vs m Tvs vs'.
  apply (type_msg_send_subset vs).
  assumption.
  intros a. set_iff. intuition.
Qed.

Hint Resolve type_msg_send_ext.

Lemma type_msg_send_vars :
  forall (vs:IdSet.t) (m:msg),
    type_msg_send vs m ->
    Subset (message_vars m) vs .
Proof.
  induction m; simpl; intuition.
  unfold exist_var in |- *.
  intro.
  unfold Subset in |- *.
  intros.
  assert (E.eq (varid v) a).
  apply singleton_1.
  exact H0.
  rewrite <- H1 in |- *.
  exact H.
Qed.

Hint Resolve type_msg_send_vars.

Lemma type_msg_send_vars_msg:
  forall (vs:IdSet.t) (l:list var),
    type_msg_send vs (vars_msg l) <->
    Subset (varlist_set l) vs.
Proof.
  simple induction l.
  simpl. intuition.
  simpl. intros. elim H. clear H. intros. split.
  intros. elim H1. clear H1. intros.
  unfold exist_var in H1.
  assert (Subset (varlist_set l0) vs).
  apply H. exact H2.
  clear H H2. generalize H0 H1. clear H0 H1.
  set_iff. intuition.
  unfold exist_var.
  intros. split.
  generalize H1. unfold Subset. intro.
  apply H2.
  set_iff. intuition.
  apply H0. unfold Subset.
  intro. unfold Subset in H1.
  intro. apply H1.
  rewrite add_iff. right. exact H2.
Qed.

Hint Rewrite -> type_msg_send_vars_msg : ppl.

Fixpoint type_msg_recv vs (m:msg) {struct m} : Prop :=
  match m with
    | MNil => True
    | MVar v => True
    | MConst c => True
    | Join m0 m1 =>
      (type_msg_recv vs m0) /\ (type_msg_recv vs m1)
    | Hash m =>
      type_msg_send vs m
    | SymEnc v m =>
      (type_varT v Symkey) /\ (exist_var vs v) /\ (type_msg_recv vs m)
    | SymSign v m =>
      (type_varT v Symkey) /\ (exist_var vs v) /\ (type_msg_recv vs m)
    | PubEnc v m =>
      (type_varT v Pubkey) /\ (exist_var vs v) /\ (type_msg_recv vs m)
    | PubSign v m =>
      (type_varT v Pubkey) /\ (exist_var vs v) /\ (type_msg_recv vs m)
    | Bind v m =>
      (type_varT v Msg) /\ (type_msg_recv vs m)
  end .

Lemma type_msg_recv_dec:
  forall (vs:IdSet.t) (m:msg),    
    {type_msg_recv vs m} + {~ type_msg_recv vs m}.
Proof.
  intros vs m. generalize vs. clear vs.
  induction m; intro vs; simpl;
    unfold exist_var;
      [ left; trivial
        | left; trivial
        | left; trivial
        | destruct (IHm1 vs); clear IHm1;
          destruct (IHm2 vs); clear IHm2;
            [ left | right | right | right];
            intuition
        | apply type_msg_send_dec
        | idtac
        | idtac
        | idtac
        | idtac
        | idtac ];
      destruct v; simpl;
        [ destruct (type_eq_dec t0 Symkey); case (In_dec i vs); intros InH
          | destruct (type_eq_dec t0 Symkey); case (In_dec i vs); intros InH
          | destruct (type_eq_dec t0 Pubkey); case (In_dec i vs); intros InH
          | destruct (type_eq_dec t0 Pubkey); case (In_dec i vs); intros InH
          | destruct (type_eq_dec t0 Msg)];
        case (IHm vs); clear IHm; intro IHmH;
          intuition.
Qed.

Lemma type_msg_recv_vars_msg:
  forall (vs:IdSet.t) (l:list var),
    type_msg_recv vs (vars_msg l).
Proof.
  simple induction l; simpl; intuition.
Qed.

Hint Resolve type_msg_recv_vars_msg.

Lemma type_msg_send_recv :
  forall (vs:IdSet.t) (m:msg),
    type_msg_send vs m -> type_msg_recv vs m.
Proof.
  intros vs m. generalize vs. clear vs.
  induction m; simpl; intuition.
Qed.

Hint Resolve type_msg_send_recv.

Lemma type_msg_recv_ext :
  forall (vs:IdSet.t) (m:msg),
    type_msg_recv vs m ->
    forall (vs':IdSet.t),
      type_msg_recv (union vs vs') m .
Proof.
  induction m; simpl; unfold exist_var; intuition; 
    try (apply type_msg_send_ext);
      try (assert (In (varid v) (union vs vs') <-> In (varid v) vs \/ In (varid v) vs');
	try (apply union_iff)); intuition.
Qed.

Hint Resolve type_msg_recv_ext.

Lemma type_msg_recv_subset:  
  forall (vs:IdSet.t) (m:msg),
    type_msg_recv vs m ->
    forall (vs':IdSet.t),
      Subset vs vs' ->
      type_msg_recv vs' m .
Proof.
  unfold Subset.
  induction m; simpl;
    unfold exist_var, exist_vars; set_iff; intuition.
  apply type_msg_send_subset with vs.
  exact H. exact H0.
Qed. 

Hint Resolve type_msg_recv_subset.

Lemma type_msg_recv_vars :
  forall (vs:IdSet.t) (m:msg),
    type_msg_recv vs m ->
    Subset (message_vars_match m) vs .
Proof.
  induction m; simpl; intuition.
Qed.

Hint Resolve type_msg_recv_vars.

Lemma formulas_vars_smallest:
  forall (fs:list formula),
    SmallestSetWithList
    (fun env => Subset (formulas_vars_prove fs) env)
    (formulas_vars_prove fs) (formulas_vars_proveL fs).
Proof.
  unfold SmallestSetWithList.
  unfold SmallestSet, VarListSet.
  intros fs.
  split. split.
  apply subset_refl.
  intros env'. trivial.
  intros a. rewrite <- formulas_vars_proveL_equiv.
  intuition.
Qed.

Hint Resolve formulas_vars_smallest.

Fixpoint type_msgT m :=
  match m with
    | MNil => True
    | MVar v => True
    | MConst c => True
    | Join m0 m1 => (type_msgT m0) /\ (type_msgT m1)
    | Hash m => type_msgT m
    | SymEnc v m => (type_varT v Symkey) /\ (type_msgT m)
    | SymSign v m => (type_varT v Symkey) /\ (type_msgT m)
    | PubEnc v m => (type_varT v Pubkey) /\ (type_msgT m)
    | PubSign v m => (type_varT v Pubkey) /\ (type_msgT m)
    | Bind v m => (type_varT v Msg) /\ (type_msgT m)
  end.

Lemma type_msg_sendT:
  forall (m:msg) (env:t),
    type_msg_send env m -> type_msgT m.
Proof.
  intros m env.
  induction m; simpl; intuition.
Qed.

Hint Resolve type_msg_sendT.

Lemma type_msg_recvT:
  forall (m:msg) (env:t),
    type_msg_recv env m -> type_msgT m.
Proof.
  intros m env.
  induction m; simpl; intuition.
Qed.

Hint Resolve type_msg_recvT.

Lemma message_vars_smallest:
  
  forall (m:msg),
    type_msgT m ->
    SmallestSetWithList
    (fun env => type_msg_send env m)
    (message_vars m) (message_varsL m).
Proof.
  unfold SmallestSetWithList.
  unfold SmallestSet, VarListSet.
  simple induction m; clear m;
    [ idtac
      | intros v
      | intros c
      | intros m1 IHm1 m2 IHm2
      | intros m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1 ];
    simpl; unfold exist_var, Subset;
      intros TM; set_iff;
        unfold E.eq, IOT.eq;
          try (intuition; fail).

  split. split. trivial.
  intros s'. intros T a.
  set_iff. intros. contradiction.
  intros a. set_iff.
  intuition. destruct H. contradiction.

  split. split. reflexivity.
  intro s'. intros T a.
  set_iff. unfold E.eq, IOT.eq.
  intros EQ. subst a. assumption.
  intros a. set_iff. unfold E.eq, IOT.eq.
  split. intros EQ. subst a.
  destruct v as [vi vt]. simpl.
  exists vt. left. reflexivity.
  intros [vt [EQ | F]]. subst v. simpl. reflexivity.
  contradiction.

  split. split. trivial.
  intros s'. intros T a.
  set_iff. intros. contradiction.
  intros a. set_iff.
  intuition. destruct H. contradiction.

  destruct TM as [TM1 TM2].
  destruct (IHm1 TM1) as [[T1 S1] H1]. clear IHm1.
  destruct (IHm2 TM2) as [[T2 S2] H2]. clear IHm2.
  clear TM1 TM2.
  split. split. split.
  apply type_msg_send_subset with (message_vars m1).
  assumption.
  unfold Subset. intros a. set_iff. intuition.
  apply type_msg_send_subset with (message_vars m2).
  assumption.
  unfold Subset. intros a. set_iff. intuition.
  intros s' [TMS1 TMS2] a. set_iff.
  intros [IN1 | IN2].
  apply S1; assumption.
  apply S2; assumption.
  clear S1 S2 T1 T2.
  intros a. set_iff.
  rewrite H1. rewrite H2. clear H1 H2.
  split. intros [[vt IN1] | [vt IN2]].
  exists vt. rewrite <- in_app_iff. left. assumption.
  exists vt. rewrite <- in_app_iff. right. assumption.
  intros [vt H]. rewrite <- in_app_iff in H.
  destruct H as [IN1 | IN2].
  left. exists vt. assumption.
  right. exists vt. assumption.

  destruct TM as [TV TM1].
  destruct (IHm1 TM1) as [[T1 S1] H1].
  split. split. split.
  assumption.
  split. set_iff. intuition.
  apply type_msg_send_subset with (message_vars m1). assumption.
  unfold Subset. intro a. set_iff. intuition.
  intros s' [Tv [InS' TMS1]] a.
  set_iff. intros [EQ | In1].
  rewrite <- EQ. assumption.
  apply S1; assumption.
  clear IHm1 TV TM1 T1 S1.
  intros a. set_iff. rewrite H1. clear H1.
  unfold E.eq, IOT.eq. split.
  intros [EQ | [vt In1]].
  subst a.
  destruct v as [vi vt]. simpl.
  exists vt. left. reflexivity.
  exists vt. simpl. right. assumption.
  simpl. intros [vt [EQ | In1]].
  destruct v as [vi vt']. simpl.
  left. inversion EQ. reflexivity.
  right. exists vt. assumption.

  destruct TM as [TV TM1].
  destruct (IHm1 TM1) as [[T1 S1] H1].
  split. split. split.
  assumption.
  split. set_iff. intuition.
  apply type_msg_send_subset with (message_vars m1). assumption.
  unfold Subset. intro a. set_iff. intuition.
  intros s' [Tv [InS' TMS1]] a.
  set_iff. intros [EQ | In1].
  rewrite <- EQ. assumption.
  apply S1; assumption.
  clear IHm1 TV TM1 T1 S1.
  intros a. set_iff. rewrite H1. clear H1.
  unfold E.eq, IOT.eq. split.
  intros [EQ | [vt In1]].
  subst a.
  destruct v as [vi vt]. simpl.
  exists vt. left. reflexivity.
  exists vt. simpl. right. assumption.
  simpl. intros [vt [EQ | In1]].
  destruct v as [vi vt']. simpl.
  left. inversion EQ. reflexivity.
  right. exists vt. assumption.

  destruct TM as [TV TM1].
  destruct (IHm1 TM1) as [[T1 S1] H1].
  split. split. split.
  assumption.
  split. set_iff. intuition.
  apply type_msg_send_subset with (message_vars m1). assumption.
  unfold Subset. intro a. set_iff. intuition.
  intros s' [Tv [InS' TMS1]] a.
  set_iff. intros [EQ | In1].
  rewrite <- EQ. assumption.
  apply S1; assumption.
  clear IHm1 TV TM1 T1 S1.
  intros a. set_iff. rewrite H1. clear H1.
  unfold E.eq, IOT.eq. split.
  intros [EQ | [vt In1]].
  subst a.
  destruct v as [vi vt]. simpl.
  exists vt. left. reflexivity.
  exists vt. simpl. right. assumption.
  simpl. intros [vt [EQ | In1]].
  destruct v as [vi vt']. simpl.
  left. inversion EQ. reflexivity.
  right. exists vt. assumption.

  destruct TM as [TV TM1].
  destruct (IHm1 TM1) as [[T1 S1] H1].
  split. split. split.
  assumption.
  split. set_iff. intuition.
  apply type_msg_send_subset with (message_vars m1). assumption.
  unfold Subset. intro a. set_iff. intuition.
  intros s' [Tv [InS' TMS1]] a.
  set_iff. intros [EQ | In1].
  rewrite <- EQ. assumption.
  apply S1; assumption.
  clear IHm1 TV TM1 T1 S1.
  intros a. set_iff. rewrite H1. clear H1.
  unfold E.eq, IOT.eq. split.
  intros [EQ | [vt In1]].
  subst a.
  destruct v as [vi vt]. simpl.
  exists vt. left. reflexivity.
  exists vt. simpl. right. assumption.
  simpl. intros [vt [EQ | In1]].
  destruct v as [vi vt']. simpl.
  left. inversion EQ. reflexivity.
  right. exists vt. assumption.

  destruct TM as [TV TM1].
  destruct (IHm1 TM1) as [[T1 S1] H1]. clear IHm1.
  split. split. split; assumption.
  intros s' [Tv TMS1] a.
  intros IN. apply S1; assumption.
  intros a. rewrite H1.
  intuition.
Qed.

Hint Resolve message_vars_smallest.

Lemma message_vars_match_smallest:
  
  forall (m:msg),
    type_msgT m ->
    SmallestSetWithList
    (fun env => type_msg_recv env m)
    (message_vars_match m) (message_vars_matchL m).
Proof.
  unfold SmallestSetWithList.
  unfold SmallestSet, VarListSet.
  simple induction m; clear m;
    [ idtac
      | intros v
      | intros c
      | intros m1 IHm1 m2 IHm2
      | intros m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1
      | intros v m1 IHm1 ];
    simpl; unfold exist_var, Subset;
      intros TM; set_iff;
        unfold E.eq, IOT.eq;
          try (intuition; fail).

  split. split. trivial.
  intros s'. intros T a.
  set_iff. intros. contradiction.
  intros a. set_iff.
  intuition. destruct H. contradiction.

  split. split. trivial.
  intros s'. intros T a.
  set_iff. intros. contradiction.
  intros a. set_iff.
  intuition. destruct H. contradiction.

  split. split. trivial.
  intros s'. intros T a.
  set_iff. intros. contradiction.
  intros a. set_iff.
  intuition. destruct H. contradiction.

  destruct TM as [TM1 TM2].
  destruct (IHm1 TM1) as [[T1 S1] H1]. clear IHm1.
  destruct (IHm2 TM2) as [[T2 S2] H2]. clear IHm2.
  clear TM1 TM2.
  split. split. split.
  apply type_msg_recv_subset with (message_vars_match m1).
  assumption.
  unfold Subset. intros a. set_iff. intuition.
  apply type_msg_recv_subset with (message_vars_match m2).
  assumption.
  unfold Subset. intros a. set_iff. intuition.
  intros s' [TMS1 TMS2] a. set_iff.
  intros [IN1 | IN2].
  apply S1; assumption.
  apply S2; assumption.
  clear S1 S2 T1 T2.
  intros a. set_iff.
  rewrite H1. rewrite H2. clear H1 H2.
  split. intros [[vt IN1] | [vt IN2]].
  exists vt. rewrite <- in_app_iff. left. assumption.
  exists vt. rewrite <- in_app_iff. right. assumption.
  intros [vt H]. rewrite <- in_app_iff in H.
  destruct H as [IN1 | IN2].
  left. exists vt. assumption.
  right. exists vt. assumption.

  clear IHm1. destruct (message_vars_smallest m1 TM).
  unfold SmallestSet, VarListSet in *.
  intuition.

  destruct TM as [TV TM1].
  destruct (IHm1 TM1) as [[T1 S1] H1]. clear IHm1.
  split. split. split.
  assumption.
  split. set_iff. intuition.
  apply type_msg_recv_subset with (message_vars_match m1).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  intros s' [Tv [InS' TMS1]] a.
  set_iff. intros [EQ | In1].
  rewrite <- EQ. assumption.
  apply S1; assumption.
  clear TV TM1 T1 S1.
  intros a. set_iff. rewrite H1. clear H1.
  unfold E.eq, IOT.eq. split.
  intros [EQ | [vt In1]].
  subst a.
  destruct v as [vi vt]. simpl.
  exists vt. left. reflexivity.
  exists vt. simpl. right. assumption.
  simpl. intros [vt [EQ | In1]].
  destruct v as [vi vt']. simpl.
  left. inversion EQ. reflexivity.
  right. exists vt. assumption.

  destruct TM as [TV TM1].
  destruct (IHm1 TM1) as [[T1 S1] H1]. clear IHm1.
  split. split. split.
  assumption.
  split. set_iff. intuition.
  apply type_msg_recv_subset with (message_vars_match m1).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  intros s' [Tv [InS' TMS1]] a.
  set_iff. intros [EQ | In1].
  rewrite <- EQ. assumption.
  apply S1; assumption.
  clear TV TM1 T1 S1.
  intros a. set_iff. rewrite H1. clear H1.
  unfold E.eq, IOT.eq. split.
  intros [EQ | [vt In1]].
  subst a.
  destruct v as [vi vt]. simpl.
  exists vt. left. reflexivity.
  exists vt. simpl. right. assumption.
  simpl. intros [vt [EQ | In1]].
  destruct v as [vi vt']. simpl.
  left. inversion EQ. reflexivity.
  right. exists vt. assumption.
  destruct TM as [TV TM1].
  destruct (IHm1 TM1) as [[T1 S1] H1]. clear IHm1.
  split. split. split.
  assumption.
  split. set_iff. intuition.
  apply type_msg_recv_subset with (message_vars_match m1).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  intros s' [Tv [InS' TMS1]] a.
  set_iff. intros [EQ | In1].
  rewrite <- EQ. assumption.
  apply S1; assumption.
  clear TV TM1 T1 S1.
  intros a. set_iff. rewrite H1. clear H1.
  unfold E.eq, IOT.eq. split.
  intros [EQ | [vt In1]].
  subst a.
  destruct v as [vi vt]. simpl.
  exists vt. left. reflexivity.
  exists vt. simpl. right. assumption.
  simpl. intros [vt [EQ | In1]].
  destruct v as [vi vt']. simpl.
  left. inversion EQ. reflexivity.
  right. exists vt. assumption.
  destruct TM as [TV TM1].
  destruct (IHm1 TM1) as [[T1 S1] H1]. clear IHm1.
  split. split. split.
  assumption.
  split. set_iff. intuition.
  apply type_msg_recv_subset with (message_vars_match m1).
  assumption.
  unfold Subset. intro a. set_iff. intuition.
  intros s' [Tv [InS' TMS1]] a.
  set_iff. intros [EQ | In1].
  rewrite <- EQ. assumption.
  apply S1; assumption.
  clear TV TM1 T1 S1.
  intros a. set_iff. rewrite H1. clear H1.
  unfold E.eq, IOT.eq. split.
  intros [EQ | [vt In1]].
  subst a.
  destruct v as [vi vt]. simpl.
  exists vt. left. reflexivity.
  exists vt. simpl. right. assumption.
  simpl. intros [vt [EQ | In1]].
  destruct v as [vi vt']. simpl.
  left. inversion EQ. reflexivity.
  right. exists vt. assumption.

  destruct TM as [TV TM1].
  destruct (IHm1 TM1) as [[T1 S1] H1]. clear IHm1.
  split. split. split; assumption.
  intros s' [Tv TMS1] a.
  intros IN. apply S1; assumption.
  intros a. rewrite H1.
  intuition.
Qed.

Hint Resolve message_vars_match_smallest.

Theorem type_msg_send_iff:
  forall (vs:t) (m:msg),
    type_msg_send vs m <-> (Subset (message_vars m) vs /\ type_msgT m).
Proof.
  intros vs m.
  split.
  intros TM. split. apply type_msg_send_vars. assumption.
  apply type_msg_sendT with vs. assumption.

  intros [SUB TM].
  generalize (message_vars_smallest m TM).
  unfold SmallestSetWithList, SmallestSet.
  intros [[TMS Min] VLS].
  apply (type_msg_send_subset (message_vars m) m); assumption.
Qed.

Hint Rewrite -> type_msg_send_iff : ppl.

Theorem type_msg_recv_iff:
  forall (vs:t) (m:msg),
    type_msg_recv vs m <-> 
    (Subset (message_vars_match m) vs /\ type_msgT m).
Proof.
  intros vs m.
  split.
  intros TM. split. apply type_msg_recv_vars. assumption.
  apply type_msg_recvT with vs. assumption.

  intros [SUB TM].
  generalize (message_vars_match_smallest m TM).
  unfold SmallestSetWithList, SmallestSet.
  intros [[TMS Min] VLS].
  apply (type_msg_recv_subset (message_vars_match m) m); assumption.
Qed.

Hint Rewrite -> type_msg_recv_iff : ppl.

Add Morphism type_msg_send : type_msg_send_m.
Proof.
  intros x1 x2 H; induction x; simpl;
    unfold exist_var; intuition; 
      [ rewrite <- H
	| rewrite H
	| rewrite <- H
	| rewrite H
	| rewrite <- H
	| rewrite H
	| rewrite <- H
	| rewrite H
	| rewrite <- H
	| rewrite H]; intuition.
Qed.      

Add Morphism type_msg_recv : type_msg_recv_m.
Proof.
  intros x1 x2 H; induction x; simpl; unfold exist_var;
    [ intuition 
      | intuition
      | intuition
      | intuition
      | apply type_msg_send_m; exact H
      | intuition; [rewrite <- H | rewrite H]; intuition
      | intuition; [rewrite <- H | rewrite H]; intuition
      | intuition; [rewrite <- H | rewrite H]; intuition
      | intuition; [rewrite <- H | rewrite H]; intuition
      | intuition ] .
Qed.

(* Strands *)
Inductive signs : Set := 
| Outgoing : signs
| Incoming : signs .

Hint Constructors signs.

Lemma signs_eq_dec :
  forall u v : signs, {u = v} + {u <> v} .
Proof.
  decide equality.
Qed.

Definition strand_desc :=
  prod (prod signs msg) (list formula) .

Hint Unfold strand_desc.

Lemma strand_desc_eq_dec :
  forall x y : strand_desc, {x = y} + {x <> y}.
Proof.
  decide equality. apply (list_eq_dec formula_eq_dec).
  decide equality. apply msg_eq_dec. apply signs_eq_dec.
Qed.

  (* Those variables introduced by sd *)
Definition strand_desc_vars_bound sd :=
  match sd with
    | (Outgoing, m, fs) =>
      (formulas_vars_bound fs)
    | (Incoming, m, fs) =>
      (union (message_vars_bound m) (formulas_vars_bound fs))
  end.

Definition type_strand_desc vs (sd:strand_desc) := 
  match sd with
    | (Outgoing, m, fs) => 
      
      type_msg_send (union vs (formulas_vars_bound fs)) m
      /\ Subset (formulas_vars_prove fs) vs
    | (Incoming, m, fs) =>		  	
      type_msg_recv vs m
      /\ Subset (formulas_vars_prove fs)
      (union vs (message_vars_bound m))
  end.

  (* Those variables needed to execute sd *)
Definition strand_desc_vars (sd:strand_desc) :=
  match sd with
    | (Outgoing, m, f) =>
      (union (diff (message_vars m) (formulas_vars_bound f))
        (formulas_vars_prove f))
    | (Incoming, m, f) =>
      (union (message_vars_match m)
        (diff (formulas_vars_prove f) (message_vars_bound m)))
  end.

Definition strand_desc_varsL (sd:strand_desc) :=
  match sd with
    | (Outgoing, m, f) =>
      (List.filter 
        (fun v => negb (mem (varid v)
          (formulas_vars_bound f)))
        (message_varsL m))
      ++ (formulas_vars_proveL f)
    | (Incoming, m, f) =>
      (List.filter 
        (fun v => negb (mem (varid v)
          (message_vars_bound m)))
        (formulas_vars_proveL f))
      ++ (message_vars_matchL m)
  end.

Definition strand := list strand_desc .

Hint Unfold strand.

Lemma strand_eq_dec: forall x y : strand, {x = y} + {x <> y}.
Proof.
  unfold strand.
  apply (list_eq_dec strand_desc_eq_dec).
Qed.

Fixpoint strand_vars (s:strand) :=
  match s with
    | nil => empty
    | sd :: s' =>
      (union (strand_desc_vars sd)
	(diff (strand_vars s') (strand_desc_vars_bound sd)))
  end.

Definition strand_vars_bound s :=
  List.fold_right
  (fun e a => union a (strand_desc_vars_bound e))
  empty s.

Fixpoint strand_varsL (s:strand) :=
  match s with
    | nil => nil
    | sd :: s' =>
      (strand_desc_varsL sd)
      ++ 
      (List.filter 
        (fun v => negb (mem (varid v)
          (strand_desc_vars_bound sd)))
        (strand_varsL s'))
  end.

Fixpoint type_strand vs s :=
  match s with
    | nil => True
    | sd :: s' =>
      (type_strand_desc vs sd)
      /\ (type_strand (union vs (strand_desc_vars_bound sd)) s')
  end .

Add Morphism type_strand_desc : type_strand_desc_m.
Proof.
  intros.
  destruct x. destruct p.
  destruct s; simpl;
    rewrite H; intuition.
Qed.

Add Morphism type_strand : type_strand_m.
Proof.
  intros.
  generalize x1 x2 H.
  clear x1 x2 H.
  induction x.
  intros. simpl. intuition.
  intros. simpl. intuition.

  rewrite <- H. exact H1.
  assert ((type_strand (union x1 (strand_desc_vars_bound a)) x -> 
    type_strand (union x2 (strand_desc_vars_bound a)) x)
  /\ (type_strand (union x2 (strand_desc_vars_bound a)) x -> 
    type_strand (union x1 (strand_desc_vars_bound a)) x)).
  apply IHx.
  assert (Equal (union x1 (strand_desc_vars_bound a)) (union x2 (strand_desc_vars_bound a))).
  rewrite H. intuition. exact H0.
  decompose [and] H0; clear H0. apply H3. exact H2.

  rewrite H. exact H1.
  assert ((type_strand (union x1 (strand_desc_vars_bound a)) x -> 
    type_strand (union x2 (strand_desc_vars_bound a)) x)
  /\ (type_strand (union x2 (strand_desc_vars_bound a)) x -> 
    type_strand (union x1 (strand_desc_vars_bound a)) x)).
  apply IHx.
  assert (Equal (union x1 (strand_desc_vars_bound a)) (union x2 (strand_desc_vars_bound a))).
  rewrite H. intuition. exact H0.
  decompose [and] H0; clear H0.  apply H4. exact H2.
Qed.

Lemma type_strand_desc_subset:
  
  forall (vs vs':t) (sd:strand_desc),
    Subset vs vs' ->
    type_strand_desc vs sd ->
    
    type_strand_desc vs' sd.
Proof.
  intros vs vs' sd SUB.
  destruct sd as [[r m] f]. destruct r;
  simpl; intros [TM TF].
  split. apply type_msg_send_subset 
  with (union vs (formulas_vars_bound f)).
  assumption.
  intro. set_iff. intuition.
  apply subset_trans with vs; assumption.
  split. apply type_msg_recv_subset with vs;
  assumption.
  apply subset_trans 
    with (union vs (message_vars_bound m)).
  assumption.
  intro. set_iff. intuition.
Qed.

Hint Resolve type_strand_desc_subset.

Lemma type_strand_desc_ext :
  forall (vs:IdSet.t) (sd:strand_desc),
    type_strand_desc vs sd ->
    forall (vs':IdSet.t),
      type_strand_desc (union vs vs') sd.
Proof.
  intros vs sd TSD vs'.
  apply type_strand_desc_subset with vs.
  intro a. set_iff. intuition.
  assumption.
Qed.

Hint Resolve type_strand_desc_ext.

Lemma type_strand_ext :
  forall (vs:IdSet.t) (s:strand),
    type_strand vs s ->
    forall (vs':IdSet.t),
      type_strand (union vs vs') s .
Proof.
  intros vs s.
  generalize vs. clear vs.
  induction s; simpl; intuition.
  assert (Equal (union (union vs vs') (strand_desc_vars_bound a))
    (union (union vs (strand_desc_vars_bound a)) vs')).
  unfold Equal. intros. set_iff. intuition.
  rewrite H. apply IHs. exact H1.
Qed.

Hint Resolve type_strand_ext.

Lemma type_strand_subset:
  
  forall (vs vs':t) (s:strand),
    Subset vs vs' ->
    type_strand vs s ->
    
    type_strand vs' s.
Proof.
  intros vs vs' s. revert vs vs'.
  induction s as [|sd s]; intros vs vs' SUB;
    simpl. trivial.
  intros [TSD TS]. split. 
  apply type_strand_desc_subset with vs; assumption.
  apply IHs with (union vs (strand_desc_vars_bound sd)).
  unfold Subset. intros a. set_iff. intuition.
  assumption.
Qed.

Hint Resolve type_strand_subset.

(* Semantics Help *)
Fixpoint message_replace x xm m {struct m} :=
  match m with
    | MNil => MNil
    | MVar v =>
      if ME.eq_dec (varid x) (varid v) then
        Bind x xm 
        else 
          MVar v
    | MConst c => MConst c
    | Join m0 m1 => 
      Join (message_replace x xm m0) (message_replace x xm m1)
    | Hash m => Hash (message_replace x xm m)
    | SymEnc v m => SymEnc v (message_replace x xm m)
    | SymSign v m => SymSign v (message_replace x xm m)
    | PubEnc v m => PubEnc v (message_replace x xm m)
    | PubSign v m => PubSign v (message_replace x xm m)
    | Bind v m =>	  
      if ME.eq_dec (varid x) (varid v) then	    
        Bind v m	    
        else
          Bind v (message_replace x xm m)	      
  end .

Lemma message_replace_eq:
  forall (x:var) (xm:msg) (m:msg),
    ~ In (varid x) (message_vars m) ->
    (m = (message_replace x xm m)).
Proof.
  induction m; simpl; set_iff; intuition.
  destruct (ME.eq_dec (varid x) (varid v)); intros.
  symmetry in e; contradiction.
  reflexivity.
  rewrite <- H; rewrite <- H2; reflexivity.
  rewrite <- H0; reflexivity.
  rewrite <- H; reflexivity.
  rewrite <- H; reflexivity.
  rewrite <- H; reflexivity.
  rewrite <- H; reflexivity.
  destruct (ME.eq_dec (varid x) (varid v)); intros.
  reflexivity.
  rewrite <- H0; reflexivity.
Qed.

Hint Rewrite <- message_replace_eq : ppl.

Lemma message_replace_type_send :
  forall (vs:IdSet.t) (xm:msg),
    type_msg_send vs xm ->
    forall (m:msg),
      type_msg_send vs m ->
      forall (x:var),
        type_varT x Msg ->
	type_msg_send vs (message_replace x xm m) .
Proof.
  intros vs xm Hxm.
  induction m; simpl; intuition.
  destruct (ME.eq_dec (varid x) (varid v)).
  simpl. split. exact H0. exact Hxm.
  simpl. exact H.
  destruct (ME.eq_dec (varid x) (varid v)).
  simpl. split. exact H0. exact H1.
  simpl. split. exact H0. 
  apply H. exact H2.
Qed.

Hint Resolve message_replace_type_send.

Lemma message_replace_type_recv :
  forall (vs:IdSet.t) (xm:msg),
    type_msg_recv vs xm ->
    type_msg_send vs xm ->
    forall (m:msg),
      type_msg_recv vs m ->
      forall (x:var),
        type_varT x Msg ->
	type_msg_recv vs (message_replace x xm m) .
Proof.
  intros vs xm Hxmr Hxms.
  induction m; simpl; intuition.
  destruct (ME.eq_dec (varid x) (varid v)).
  simpl. split. exact H0. exact Hxmr.
  simpl. trivial.
  destruct (ME.eq_dec (varid x) (varid v)).
  simpl. split. exact H0. exact H1.
  simpl. split. exact H0. apply H. exact H2.
Qed.

Hint Resolve message_replace_type_recv.

Lemma message_replace_vars_ext:
  forall (x:var) (xm:msg) (m:msg),
    exists ext:IdSet.t,
      (Equal (message_vars_bound (message_replace x xm m))
        (union (message_vars_bound m) ext)).
Proof.
  intros x xm m.
  induction m; simpl; intuition.
  exists empty. unfold Equal; intro; set_iff; intuition.
  case (ME.eq_dec (varid x) (varid v)); intros.
  exists (message_vars_bound xm). rewrite <- e.
  simpl. unfold Equal. intros. set_iff; intuition.
  exists empty. simpl. unfold Equal. intros. set_iff; intuition.
  exists empty. simpl. unfold Equal. intros. set_iff; intuition.
  elim IHm1. intros. clear IHm1.
  elim IHm2. intros. clear IHm2.
  exists (union x0 x1).
  rewrite H. rewrite H0. clear H H0.
  unfold Equal. intros. set_iff. intuition.
  exists empty. simpl. unfold Equal. intros. set_iff; intuition.
  case (ME.eq_dec (varid x) (varid v)); intros.
  exists empty. simpl. unfold Equal. intros. set_iff; intuition.
  elim IHm. intros. clear IHm.
  simpl.
  exists x0. rewrite H.
  unfold Equal. intros. set_iff. intuition.
Qed.

Hint Resolve message_replace_vars_ext.

Definition strand_desc_replace x xm (sd:strand_desc) :=
  match sd with
    | (dir, m, f) =>
      let m' := message_replace x xm m in
        (dir, m', f)
  end.

Lemma strand_desc_replace_type:
  forall (vs:IdSet.t) (xm:msg),
    type_msg_recv vs xm ->
    type_msg_send vs xm ->
    forall (sd:strand_desc),
      type_strand_desc vs sd ->
      forall (x:var),
        type_varT x Msg ->
        type_strand_desc vs (strand_desc_replace x xm sd).
Proof.
  intros vs xm TMR TMS sd.
  destruct sd as [[r m] f].
  destruct r; simpl;
    intros [TM SS] x TV.
  split. apply message_replace_type_send.
  apply type_msg_send_subset with vs.
  assumption. intro. set_iff. intuition.
  assumption. assumption. assumption.
  split. apply message_replace_type_recv; assumption.
  destruct (message_replace_vars_ext x xm m) as [ext EQ].
  rewrite EQ. clear EQ TV TM TMR TMS.
  apply subset_trans with (union vs (message_vars_bound m)).
  assumption. intro. set_iff. intuition.
Qed.

Hint Resolve strand_desc_replace_type.

Lemma strand_desc_replace_vars_ext:
  forall (x:var) (xm:msg) (sd:strand_desc),
    exists ext:IdSet.t,
      (Equal (strand_desc_vars_bound (strand_desc_replace x xm sd))
	(union (strand_desc_vars_bound sd) ext)).
Proof.
  intros x xm sd.
  destruct sd as [[r m] f].
  destruct r; simpl.
  exists empty. intro. set_iff. intuition.
  destruct (message_replace_vars_ext x xm m) as [ext EQ].
  exists ext. rewrite EQ. clear EQ.
  intro. set_iff. intuition.
Qed.

Hint Resolve strand_desc_replace_vars_ext.

Fixpoint strand_replace (x:var) (xm:msg) (s:strand) {struct s} :=
  match s with
    | nil => nil
    | sd :: s =>
      (strand_desc_replace x xm sd)
      :: (strand_replace x xm s)
  end .      

Lemma strand_replace_type :
  forall (vs:IdSet.t) (xm:msg),
    type_msg_recv vs xm ->
    type_msg_send vs xm ->
    forall (s:strand),
      type_strand vs s ->
      forall (x:var),
        type_varT x Msg ->
	type_strand vs (strand_replace x xm s) .
Proof.
  intros vs xm Hxmr Hxms s.
  generalize vs xm Hxmr Hxms. clear vs xm Hxmr Hxms.
  induction s; simpl; intros; intuition.
  apply IHs.
  apply type_msg_recv_ext. exact Hxmr.
  apply type_msg_send_ext. exact Hxms. 
  assert (exists ext:t,
    (Equal (strand_desc_vars_bound (strand_desc_replace x xm a))
      (union (strand_desc_vars_bound a) ext))).
  apply strand_desc_replace_vars_ext.
  elim H. clear H. intros.
  rewrite H. clear H.
  assert (Equal (union vs (union (strand_desc_vars_bound a) x0))
    (union (union vs (strand_desc_vars_bound a)) x0)).
  unfold Equal. intro. set_iff. intuition.
  rewrite H. clear H.
  apply type_strand_ext. exact H2.
  exact H0.
Qed.

Hint Resolve strand_replace_type.

Definition ai_id := d1.
Definition pr_id := d2.

Definition ai := Var ai_id Nonce .
Definition pr := Var pr_id Name .
Hint Unfold ai.
Hint Unfold pr.

Definition faildesc : strand_desc := 
  ((Outgoing, (Join (MConst C.FAIL) (MVar ai))), nil).
Hint Unfold faildesc.

Lemma strand_desc_replace_faildesc:
  forall (sd:strand_desc),
    sd <> faildesc ->
    forall (x:var) (xm:msg),
      strand_desc_replace x xm sd <> faildesc .
Proof.
  destruct sd as [[r m] fs].
  destruct r; unfold not; simpl;
    intros H x xm SDR.

  inversion SDR. clear SDR.
  subst fs. revert H1.
  destruct m; simpl; intros MR;
    try (discriminate MR); revert MR.

  case (ME.eq_dec (varid x) (varid v));
    intros EQ MR; discriminate MR.

  intros MR. inversion MR. clear MR.
  destruct m1; simpl in H1; try (discriminate H1);
    revert H1.

  case (ME.eq_dec (varid x) (varid v));
    intros EQ MR; discriminate MR.

  intros EQ. inversion EQ. subst c. clear EQ.
  destruct m2; simpl in H2; try (discriminate H2);
    revert H2.

  case (ME.eq_dec (varid x) (varid v));
    intros EQ MR; try (discriminate MR).
  inversion MR. subst v. clear MR.
  apply H. reflexivity.

  case (ME.eq_dec (varid x) (varid v));
    intros EQ MR; discriminate MR.

  case (ME.eq_dec (varid x) (varid v));
    intros EQ MR; discriminate MR.

  case (ME.eq_dec (varid x) (varid v));
    intros EQ MR; discriminate MR.

  discriminate SDR.
Qed.

Hint Resolve strand_desc_replace_faildesc.

Lemma strand_replace_faildesc:
  forall (s:strand),
    last s faildesc <> faildesc ->
    forall (x:var) (xm:msg),
      last (strand_replace x xm s) faildesc <> faildesc .
Proof.
  induction s.
  simpl. intuition.
  destruct s.
  simpl. intros.
  apply strand_desc_replace_faildesc.
  exact H.
  simpl. exact IHs.
Qed.

Hint Resolve strand_replace_faildesc.

Definition call_msg n xs :=
  (Join (MConst C.CALL)
    (Join (MVar pr)
      (Join (MConst n)
	(Join (MVar ai)
	  (vars_msg xs))))) .

Hint Unfold call_msg.