(* ListOrder *)
Require Export WPPL.
Open Scope list_scope .

Section list_order.
  Variable A : Set.
  Variable f : t -> A -> bool.
  Variable f_subset : forall (s s':t) (a:A),
    f s a = true -> Subset s s' -> f s' a = true.
  Variable f_equal : forall (s s':t) (a:A),
    Equal s s' -> f s a = f s' a.

  Variable C : A -> t.

  Lemma partition_filter:
    forall (l:list A) (g:A -> bool),
      List.partition g l
      = ((List.filter g l),
         (List.filter (fun a => negb (g a)) l)).
  Proof.
    intros l g.
    induction l as [|a l]; simpl; intuition.
    destruct (List.partition g l) as [o c].
    case (g a); simpl;
    inversion IHl; reflexivity.
  Qed.
    
  Lemma partition_perm:
    forall (l:list A) (f:A -> bool),
      let oc := List.partition f l in
        Permutation l ((fst oc) ++ (snd oc)).
  Proof.
    intros l g.
    induction l as [|a l]; simpl; intuition.
    apply perm_nil.
    destruct (List.partition g l) as [o c].
    simpl in IHl.
    case (g a); simpl.
    apply perm_skip. assumption.
    apply Permutation_cons_app. assumption.
  Qed.

  Lemma partition_fst:
    forall (l:list A) (g:A -> bool) (a:A),
      List.In a (fst (List.partition g l)) ->
      g a = true.
  Proof.
    intros l g a.
    rewrite partition_filter.
    simpl.
    rewrite filter_In.
    intuition.
  Qed.

  Lemma partition_snd:
    forall (l:list A) (g:A -> bool) (a:A),
      List.In a (snd (List.partition g l)) ->
      g a = false.
  Proof.
    intros l g a.
    rewrite partition_filter.
    simpl.
    rewrite filter_In.
    case (g a); intuition.
  Qed.

  Lemma partition_snd_len:
    forall (l:list A) (g:A -> bool),
      List.length (snd (List.partition g l))
      = List.length l - List.length (fst (List.partition g l)).
  Proof.
    intros l g.
    generalize (partition_perm l g).
    destruct (List.partition g l) as [o c].
    simpl. intros PR.
    generalize (Permutation_length PR).
    intros EQ. rewrite EQ.
    rewrite app_length.
    auto with arith.
  Qed.
    
  Definition LC l :=
    List.fold_right (fun e a => union (C e) a) empty l.

  Lemma LC_comm:
    forall l l',
      Equal (union (LC l) (LC l')) (LC (l ++ l')).
  Proof.
    induction l as [|a l]; simpl; intros l'.
    intuition.
    rewrite union_assoc. rewrite IHl. reflexivity.
  Qed.
   
  Fixpoint list_order (n:nat) (i:t) (l:list A) : (list A * list A):=
    match n with
      | O => (nil, l)
      | S n =>
        let (o,c) := List.partition (f i) l in
          match o with
            | nil => (o,c)
            | _ =>
              let i' := union i (LC o) in
              let (o', c') := list_order n i' c in
                (o ++ o', c')
          end
    end.
    
  Lemma list_order_perm:
    forall n i l,
      let oc := list_order n i l in
        Permutation l ((fst oc) ++ (snd oc)).
  Proof.
    induction n as [|n]; intros i l; simpl.
    apply Permutation_refl.
    generalize (partition_perm l (f i)).
    destruct (List.partition (f i) l) as [o c].
    simpl.
    destruct o as [|oa oo].
    simpl. trivial.
    set (i' := union i (LC (oa :: oo))).
    generalize (IHn i' c).
    destruct (list_order n i' c) as [o' c'].
    simpl.    
    intros Pc Pl.
    apply Permutation_trans with (oa :: oo ++ c).
    assumption. clear Pl.
    apply perm_skip.
    rewrite app_ass.
    apply Permutation_app_head.
    assumption.
  Qed.

  Lemma partition_snd_len_n:
    forall l n i oa oo c,
      length l <= S n ->
      List.partition (f i) l = (oa :: oo, c) ->
      length c <= n.
  Proof.
    intros l n i oa oo c LE EQl.
    generalize (partition_snd_len l (f i)).
    rewrite EQl. simpl. intros EQc.
    rewrite EQc.
    revert LE.
    set (x := length l).
    set (y := length oo).
    intros LE.
    generalize (Arith.Le.le_pred x (S n) LE).
    simpl. intros LE2.
    apply Arith.Le.le_trans with (pred x).
    rewrite Arith.Minus.pred_of_minus.
    cut (x - S y = (x - 1) - y). intros EQ.
    rewrite EQ. apply Arith.Minus.le_minus.
    pattern x, y. apply nat_double_ind; auto with arith.
    intros x' y'. simpl.
    rewrite <- Arith.Minus.minus_n_O. reflexivity.
    assumption.
  Qed.

  Lemma partition_equal:
    forall (s s':t) l,
      Equal s s' ->
      List.partition (f s) l =
      List.partition (f s') l.
  Proof.
    intros s s' l EQs.
    induction l as [|a l]; simpl; intuition.
    rewrite IHl. rewrite (f_equal s s' a EQs).
    reflexivity.
  Qed.

  Theorem list_order_closed:
    forall n i l,
      List.length l <= n ->
      forall o c,
        list_order n i l = (o,c) ->
        forall a,
          List.In a c ->
          f (union i (LC o)) a = false.
  Proof.
    induction n as [|n]; intros i l.
    destruct l as [|e l]; simpl; intros LE.
    intros o c EQ. inversion EQ. simpl. intuition.
    absurd_hyp LE.
    apply Arith.Le.le_Sn_O. assumption.
    intros LE o' c'. simpl.
    case_eq (List.partition (f i) l).
    intros o c EQl.
    destruct o as [|oa oo].
    simpl. intros EQ. inversion EQ.
    subst o' c'. clear EQ.
    simpl. intros a INa.
    apply (partition_snd l).
    rewrite (partition_equal (union i empty) i).
    rewrite EQl. simpl. assumption.
    intuition.
    case_eq (list_order n (union i (LC (oa::oo))) c).
    intros o'' c'' EQl' EQ.
    inversion EQ. subst c' o'. clear EQ.
    intros a INa.
    set (i' := (union i (LC (oa :: oo)))).
    cut (List.length c <= n). intros LE2.
    rewrite (f_equal (union i (LC (oa :: oo ++ o''))) (union i' (LC o''))).
    apply (IHn i' c LE2 o'' c'' EQl' a INa).
    unfold i'. rewrite union_assoc. rewrite LC_comm.
    simpl. reflexivity.
    apply (partition_snd_len_n l n i oa oo c LE EQl).
  Qed.

  Theorem in_split_or:
    forall (a:A) o o' pre post,
      o ++ o' = pre ++ a :: post ->
      (* In o *)
      (exists post':list A,
        o = pre ++ a :: post'
        /\ post = post' ++ o')
      \/ (* In o' *)
      (exists pre':list A,
        pre = o ++ pre'
        /\ o' = pre' ++ a :: post).
  Proof.
   intros a.
   induction o as [|oe o].
   simpl. intros o' pr po EQ.
   right. exists pr. intuition.
   
   intros o' pre. revert o'.
   induction pre as [|pre pr].
   simpl. intros o' po EQ. inversion EQ. subst oe po. clear EQ.
   left. exists o. intuition.

   intros o' po.
   simpl. intros EQ.
   inversion EQ. subst oe. clear EQ. rename H1 into EQ.
   destruct (IHo o' pr po EQ)
     as [[po' [EQo EQp]]
        |[pr' [EQp EQo]]].
   subst o po.
   left. exists po'. intuition.
   subst pr o'.
   right. exists pr'. intuition.
 Qed.

 Theorem list_order_opened:
   forall n i l,
     List.length l <= n ->
     forall o c,
       list_order n i l = (o,c) ->
       forall a pre post,
         o = pre ++ a :: post ->
         f (union i (LC pre)) a = true.
 Proof.
   induction n as [|n]; intros i l.
   destruct l as [|e l]; simpl; intros LE.
   intros o c EQlo. inversion EQlo. simpl.
   intros a pre post. subst o c. clear EQlo.
   intros EQ. absurd_hyp EQ.
   apply app_cons_not_nil. assumption.
   absurd_hyp LE.
   apply Arith.Le.le_Sn_O. assumption.
   intros LE _o _c. simpl.
   case_eq (List.partition (f i) l).
   intros o c EQl.
   destruct o as [|oa oo].
   simpl. intros EQ. inversion EQ.
   subst _o _c. clear EQ.
   intros a pre post.
   intros EQ. absurd_hyp EQ.
   apply app_cons_not_nil. assumption.
   case_eq (list_order n (union i (LC (oa::oo))) c).
   intros o' c' EQlo EQ.
   inversion EQ.
   subst _c _o. clear EQ.
   intros a pre post EQo.
   
   destruct (in_split_or a (oa::oo) o' pre post EQo)
     as [[post' [EQo' EQp]]
       |[pre' [EQp EQo']]].
   
   generalize (partition_fst l (f i) a).
   rewrite EQl. rewrite EQo'. simpl.
   intros INa. apply f_subset with i; intuition.
   
   set (i' := (union i (LC (oa :: oo)))).
   cut (List.length c <= n). intros LE2.
   generalize (IHn i' c LE2 o' c' EQlo).
   clear IHn. intros IHn. clear LE2 LE EQlo c'.
   generalize (IHn a pre' post EQo').
   unfold i'. rewrite EQp.
   
   rewrite (f_equal 
     (union (union i (LC (oa :: oo))) (LC pre'))
     (union i (LC ((oa::oo) ++ pre')))). trivial.
   rewrite union_assoc. rewrite LC_comm. reflexivity.
   
   apply (partition_snd_len_n l n i oa oo c LE EQl).
 Qed.

 Lemma list_order_closed_nil_perm:
   forall n i l o,
     list_order n i l = (o, nil) ->
     Permutation l o.
 Proof.
   intros n i l o EQlo.
   generalize (list_order_perm n i l).
   rewrite EQlo. simpl.
   rewrite <- app_nil_end.
   trivial.
 Qed.

  Definition list_order_auto i l :=
    list_order (List.length l) i l.

  Lemma list_order_auto_perm:
    forall i l,
      let oc := list_order_auto i l in
        Permutation l ((fst oc) ++ (snd oc)).
  Proof.
    unfold list_order_auto.
    intros i l.
    apply list_order_perm.
  Qed.

  Theorem list_order_auto_closed:
    forall i l o c,
      list_order_auto i l = (o,c) ->
      forall a,
        List.In a c ->
        f (union i (LC o)) a = false.
  Proof.
    intros i l. unfold list_order_auto.
    apply list_order_closed.
    apply le_refl.
  Qed.

 Theorem list_order_auto_opened:
   forall i l o c,
     list_order_auto i l = (o,c) ->
     forall a pre post,
       o = pre ++ a :: post ->
       f (union i (LC pre)) a = true.
 Proof.
   intros i l. unfold list_order_auto.
   apply list_order_opened.
   apply le_refl.
 Qed.
End list_order.


