(* WPPL *)
Require Export CPPL.
Open Scope list_scope.

Require Constants.
Module C := Constants.

(* Types *)
Inductive new_type : Set :=
      | New_nonce : new_type
      | New_symkey : new_type
      | New_pubkey : new_type .
Hint Constructors new_type.

    Lemma new_type_eq_dec: 
     forall x y : new_type, {x = y} + {x <> y}.
    Proof.
     decide equality.
    Qed.

    Definition new_type_to_type nt :=
      match nt with
	| New_nonce => Nonce
	| New_symkey => Symkey
	| New_pubkey => Pubkey
      end.
  
    Definition type_to_new_type t :=
      match t with
	| Nonce => Some New_nonce
	| Symkey => Some New_symkey
	| Pubkey => Some New_pubkey
	| _ => None
      end.

Inductive action : Set :=
| Return : action
| Comm : id -> id -> list formula -> msg -> list formula -> action -> action
| New : id -> var -> new_type -> action -> action
| Bind : id -> var -> msg -> action -> action
| Match : id -> var -> msg -> action -> action
| Derive : id -> list formula -> action -> action.
Hint Constructors action.

    Lemma action_eq_dec:
     forall x y : action, {x = y} + {x <> y}.
    Proof.
     decide equality;
     try (apply id_eq_dec);
     try (apply msg_eq_dec);
     try (apply var_eq_dec);
     try (apply new_type_eq_dec);
     try (apply (list_eq_dec formula_eq_dec)).
    Qed.

    Fixpoint action_varsL a {struct a} :=
     match a with
     | Return =>
       nil
     | Comm from_r to_r from_fs msg to_fs sa =>
       (message_varsL msg) ++ (action_varsL sa) 
       (*++ (formulas_vars_proveL from_fs)
       ++ (formulas_vars_proveL to_fs)*)
     | New r v t sa =>
       v :: (action_varsL sa)
     | Bind r v msg sa =>
       v :: (message_varsL msg) ++ (action_varsL sa)
     | Match r v msg sa =>
       v :: (message_varsL msg) ++ (action_varsL sa)
     | Derive r f sa =>
       (formulas_vars_proveL f) ++ (action_varsL sa)
     end.

    Inductive role_spec : Set :=
      | RSpec : id -> list var -> list var -> role_spec .
Hint Constructors role_spec.

    Lemma role_spec_eq_dec:
     forall x y : role_spec, {x = y} + {x <> y}.
    Proof.
     decide equality;
     try (apply id_eq_dec);
     try (apply (list_eq_dec var_eq_dec)).
    Qed.

    Inductive spec : Set :=
      | Spec : list role_spec -> action -> spec .

Hint Constructors spec.

    Lemma spec_eq_dec: forall x y : spec, {x = y} + {x <> y}.
    Proof.
     decide equality;
     [ apply action_eq_dec
     | apply (list_eq_dec role_spec_eq_dec)].
    Qed.

 Fixpoint rspecs_varsL rss :=
 match rss with
 | nil => nil
 | r :: rss =>
 match r with
 | RSpec _ start_decls end_decls =>
 (start_decls ++ end_decls) ++ (rspecs_varsL rss)
 end
 end.

 Definition spec_varsL s :=
 match s with
 | Spec rspecs a =>
   (rspecs_varsL rspecs) ++ (action_varsL a)
 end.

(* Patterns *)
    Fixpoint rspec_find r rss {struct rss} :=
      match rss with 
	| nil => None
	| rs :: rss' =>
	  match rs with
	    | RSpec r' _ _ =>
	      if (ME.eq_dec r r') then
		Some rs
		else
		  rspec_find r rss'
	  end
      end .

  Lemma in_rspec_find:
   forall (r:id) (bs es:list var) (rss:list role_spec),
    List.In (RSpec r bs es) rss ->
    exists bs':list var, exists es':list var,
     rspec_find r rss = Some (RSpec r bs' es').
  Proof.
   intros r bs es rss.
   set (rs := RSpec r bs es).
   induction rss as [| rs' rss ]; simpl.
   intro. contradiction.
   intros [ EQ | InRSS ].
   subst rs'.
   simpl.
   destruct (ME.eq_dec r r).
   exists bs. exists es.
   reflexivity.
   unfold E.eq, IOT.eq in *.
   absurd (r = r). assumption. reflexivity.
   destruct rs' as [ r' bs' es' ].
   case (ME.eq_dec r r'); unfold E.eq, IOT.eq.
   intro EQ. subst r'.
   exists bs'. exists es'.
   reflexivity.
   intro NEQ. clear r' NEQ bs' es'.
   apply IHrss. assumption.
  Qed.

Hint Resolve in_rspec_find.

(* Type Check *)
Definition type_newtype v nt :=
 match nt with
 | New_nonce => type_varT v Nonce
 | New_pubkey => type_varT v Pubkey
 | New_symkey => type_varT v Symkey
 end.

Lemma type_newtype_dec :
 forall (v:var) (nt:new_type),
  {type_newtype v nt} + {~ type_newtype v nt}.
Proof.
 destruct v as [ vi vt ]. destruct nt; simpl.
 case (type_eq_dec vt Nonce); intro C; simpl; intuition.
 case (type_eq_dec vt Symkey); intro C; simpl; intuition.
 case (type_eq_dec vt Pubkey); intro C; simpl; intuition.
Qed.

Fixpoint type_action ns vs r es a {struct a} :=
 match a with
 | Return => 
   exist_vars vs (varlist_set es)
 | Comm f t ffs m tfs a =>
   if ME.eq_dec r f then
    (In t ns \/ In t vs) 
    /\ exist_vars vs (formulas_vars_prove ffs)
    /\ (type_msg_send (union vs (formulas_vars_bound ffs)) m) 
    /\ (type_action (add t ns) (union vs (formulas_vars_bound ffs)) r es a)
   else if ME.eq_dec r t then
    (type_msg_recv vs m)
    /\ exist_vars (union vs (message_vars_bound m)) (formulas_vars_prove tfs)
    /\ (type_action (add f ns)
         (union (union vs (message_vars_bound m)) (formulas_vars_bound tfs))
         r es a)
   else
    type_action ns vs r es a
 | New ar v nt a =>
   if ME.eq_dec r ar then
    (type_newtype v nt) 
    /\ (type_action ns (add (varid v) vs) r es a)
   else
    type_action ns vs r es a
 | Bind ar v m a =>
   if ME.eq_dec r ar then
    (type_varT v Msg) /\ (type_msg_send vs m)
    /\ (type_action ns (add (varid v) vs) r es a)
   else
    type_action ns vs r es a
 | Match ar v m a =>
   if ME.eq_dec r ar then
    (type_varT v Msg) /\ (exist_var vs v) /\ (type_msg_recv vs m)
    /\ (type_action ns (union vs (message_vars_bound m)) r es a)
   else
    type_action ns vs r es a
 | Derive ar f a =>
   if ME.eq_dec r ar then
    exist_vars vs (formulas_vars_prove f)
    /\ type_action ns (union vs (formulas_vars_bound f)) r es a
   else
    type_action ns vs r es a
 end.

Lemma type_action_dec :
 forall (ns vs:t) (r:id) (es:list var) (a:action),
  {type_action ns vs r es a} + {~ type_action ns vs r es a}.
Proof.
 intros ns vs r es a.
 revert a ns vs r es.
 simple induction a; clear a; simpl;
 unfold exist_vars, type_varT;
 [ idtac
 | intros fr tr ffs m tfs a IHa
 | intros ar v nt a IHa
 | intros ar v m a IHa
 | intros ar v m a IHa
 | intros ar f a IHa ];
 intros ns vs r es;
 [ idtac
 | case (ME.eq_dec r fr); [intro EQ | intro NEQ;
   case (ME.eq_dec r tr); intro EQ ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]].

 apply Subset_dec.

 destruct (In_dec tr vs).
 destruct (Subset_dec (formulas_vars_prove ffs) vs).
 destruct (type_msg_send_dec (union vs (formulas_vars_bound ffs)) m).
 destruct (IHa (add tr ns) (union vs (formulas_vars_bound ffs)) r es).
 left. eauto.
 right. intros [ F1 [ F2 [ F3 F4] ] ]. contradiction.
 right. intros [ F1 [ F2 [ F3 F4] ] ]. contradiction.
 right. intros [ F1 [ F2 [ F3 F4] ] ]. contradiction.
 destruct (In_dec tr ns).
 destruct (Subset_dec (formulas_vars_prove ffs) vs).
 destruct (type_msg_send_dec (union vs (formulas_vars_bound ffs)) m).
 destruct (IHa (add tr ns) (union vs (formulas_vars_bound ffs)) r es).
 left. eauto.
 right. intros [ F1 [ F2 [ F3 F4] ] ]. contradiction.
 right. intros [ F1 [ F2 [ F3 F4] ] ]. contradiction.
 right. intros [ F1 [ F2 [ F3 F4] ] ]. contradiction.
 right. intros [ [F11 | F12] [ F2 [ F3 F4] ] ]; contradiction.

 destruct (type_msg_recv_dec vs m).
 destruct (Subset_dec (formulas_vars_prove tfs)
  (union vs (message_vars_bound m))).
 destruct (IHa (add fr ns) 
  (union (union vs (message_vars_bound m)) (formulas_vars_bound tfs)) r es).
 eauto.
 right; intros [ F1 [F2 F3] ]. contradiction.
 right. intros [ F1 [F2 F3] ]. contradiction.
 right. intros [ F1 [F2 F3] ]. contradiction.

 apply IHa.

 destruct (type_newtype_dec v nt).
 destruct (IHa ns (add (varid v) vs) r es).
 left. split; assumption.
 right. intros [ F1 F2 ]. contradiction.
 right. intros [ F1 F2 ]. contradiction.

 apply IHa.

 destruct v as [ vid vt ].
 destruct (type_eq_dec vt Msg).
 destruct (type_msg_send_dec vs m).
 destruct (IHa ns (add (varid (Var vid vt)) vs) r es).
 left. split. trivial. split; assumption.
 right. intros [ F1 [ F2 F3 ] ]. contradiction.
 right. intros [ F1 [ F2 F3 ] ]. contradiction.
 right. intros [ F1 [ F2 F3 ] ]. contradiction.

 apply IHa.

 unfold exist_var.
 destruct v as [ vid vt ].
 destruct (type_eq_dec vt Msg).
 destruct (In_dec (varid (Var vid vt)) vs).
 destruct (type_msg_recv_dec vs m).
 destruct (IHa ns (union vs (message_vars_bound m)) r es).
 left. split. trivial. split. assumption. split; assumption.
 right. intros [ F1 [ F2 [ F3 F4 ] ] ]. contradiction.
 right. intros [ F1 [ F2 [ F3 F4 ] ] ]. contradiction.
 right. intros [ F1 [ F2 [ F3 F4 ] ] ]. contradiction.
 right. intros [ F1 [ F2 [ F3 F4 ] ] ]. contradiction.

 apply IHa.
 
 destruct (IHa ns (union vs (formulas_vars_bound f)) r es);
 destruct (Subset_dec (formulas_vars_prove f) vs);
 intuition.

 apply IHa.
Qed.

Definition type_rspecs r rss := 
 exists bs:list var, exists es:list var,
 rspec_find r rss = Some (RSpec r bs es).

Hint Unfold type_rspecs.

Lemma type_rspecs_dec :
 forall (r:id) (rss:list role_spec),
  {type_rspecs r rss} + {~ type_rspecs r rss}.
Proof.
 unfold type_rspecs.
 intros r rss.
 induction rss as [| a rss].
 
 right. intros [ bs [ es EQ ] ].
 simpl in EQ. discriminate EQ.

 destruct a as [ ar bs es ]. simpl.
 case (ME.eq_dec r ar); intro EQ.
 unfold E.eq in EQ. unfold IOT.eq in EQ.
 subst r.
 left. exists bs. exists es. reflexivity.
 apply IHrss.
Qed.

Definition type_spec r s :=
 match s with
 | Spec rss a =>
   (type_rspecs r rss)
   /\ (match rspec_find r rss with
       | None => False
       | Some (RSpec r bs es) =>
         type_action empty (varlist_set bs) r es a
       end)
 end.

Lemma type_spec_dec :
 forall (r:id) (s:spec),
  {type_spec r s} + {~ type_spec r s}.
Proof.
 intros r s.
 destruct s as [ rss a ].
 simpl.
 destruct (type_rspecs_dec r rss).
 case_eq (rspec_find r rss); 
 [ destruct r0 as [ r0 bs0 es0 ]
 | idtac ];
 intro EQ.
 destruct (type_action_dec empty (varlist_set bs0) r0 es0 a).
 left. split; assumption.
 right. intros [ F1 F2 ]; contradiction.
 right. intros [ F1 F2 ]; contradiction.
 right. intros [ F1 F2 ]; contradiction.
Qed.

Fixpoint type_spec_roles rs s :=
 match rs with
 | nil => True
 | r :: rs =>
   (type_spec r s) /\ (type_spec_roles rs s)
 end.

Lemma type_spec_roles_dec :
 forall (rs:list id) (s:spec),
  {type_spec_roles rs s} + {~ type_spec_roles rs s}.
Proof.
 intros rs s.
 induction rs as [| r rs ]; simpl.
 left. trivial.
 destruct IHrs as [ S | F ].
 destruct (type_spec_dec r s) as [ S1 | F1 ].
 left. split; assumption.
 right. intros [ X1 X2 ]. contradiction.
 right. intros [ X1 X2 ]. contradiction.
Qed.

Definition role_list := List.map (fun rs => match rs with RSpec r _ _ => r end).

Hint Unfold role_list.

Definition spec_roles s :=
 match s with
 | Spec rss _ => role_list rss
 end.

Hint Unfold spec_roles.

  Lemma spec_roles_type:
   forall (rss:list role_spec) (a:action) (r:id),
    List.In r (spec_roles (Spec rss a)) -> type_rspecs r rss.
  Proof.
   intros rss a r.
   unfold spec_roles, type_rspecs, role_list.
   intro InM.
   rewrite in_map_iff in InM.
   destruct InM as [ [ r' bs es ] [ EQr InRSS ] ].
   subst r'.
   apply in_rspec_find with bs es.
   assumption.
  Qed.

Hint Resolve spec_roles_type.

Definition type_spec_all s :=
 type_spec_roles (spec_roles s) s.

Hint Unfold type_spec_all.

Lemma type_spec_all_dec :
 forall (s:spec),
  {type_spec_all s} + {~ type_spec_all s}.
Proof.
 destruct s as [ rss a ].
 unfold type_spec_all.
 apply type_spec_roles_dec.
Qed.

Add Morphism type_action : type_action_m.
Proof.
 intros ns1 ns2 EQns vs1 vs2 EQvs r es a.
 revert a ns1 ns2 EQns vs1 vs2 EQvs.
 simple induction a; clear a; simpl;
 unfold exist_vars;
 [ idtac
 | intros fr tr ffs m tfs a IHa
 | intros ar v nt a IHa
 | intros ar v m a IHa
 | intros ar v m a IHa
 | intros ar f a IHa ];
 intros ns1 ns2 EQns vs1 vs2 EQvs;
 [ idtac
 | case (ME.eq_dec r fr); [intro EQ | intro NEQ;
   case (ME.eq_dec r tr); intro EQ ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
 | case (ME.eq_dec r ar); [intro EQ | intro NEQ]];
 unfold exist_var;
 try (rewrite EQvs in *);
 try (rewrite EQns in *);
 intuition;
 try (rewrite EQvs in *);
 try (rewrite EQns in *);
 try (destruct (IHa ns1 ns2 EQns vs1 vs2 EQvs) as [ IHa1 IHa2 ]);
 try (assert (Equal (union vs1 (message_vars_bound m))
              (union vs2 (message_vars_bound m))) as EQm
      by (apply union_equal_1; assumption));
 try (assert (Equal (add fr ns1) (add fr ns2)) as EQnaf
      by (rewrite EQns; apply equal_refl));
 try (assert (Equal 
        (union (union vs1 (message_vars_bound m)) (formulas_vars_bound tfs))
        (union (union vs2 (message_vars_bound m)) (formulas_vars_bound tfs)))
      as EQmfs
      by (apply union_equal_1; assumption));
 try (destruct (IHa (add fr ns1) (add fr ns2) EQnaf 
        (union (union vs1 (message_vars_bound m)) (formulas_vars_bound tfs))
        (union (union vs2 (message_vars_bound m)) (formulas_vars_bound tfs))
        EQmfs)
      as [ IHa1mnf IHa2mnf ]);
 try (destruct (IHa ns1 ns2 EQns
                (union vs1 (message_vars_bound m))
                (union vs2 (message_vars_bound m)) EQm)
      as [ IHa1m IHa2m ]);
 try (assert (Equal (add tr ns1) (add tr ns2)) as EQnat
      by (rewrite EQns; apply equal_refl));
 try (assert (Equal
       (union vs1 (formulas_vars_bound ffs))
       (union vs2 (formulas_vars_bound ffs))) as EQvsfs
      by (rewrite EQvs; apply equal_refl));
 try (destruct (IHa (add tr ns1) (add tr ns2) EQnat 
       (union vs1 (formulas_vars_bound ffs))
       (union vs2 (formulas_vars_bound ffs)) EQvsfs)
      as [ IHa1n IHa2n ]);
 try (assert (Equal (add (varid v) vs1) (add (varid v) vs2))
      as EQv
      by (repeat (rewrite add_union_singleton);
          apply union_equal_2; assumption));
 try (destruct (IHa ns1 ns2 EQns (add (varid v) vs1)
                 (add (varid v) vs2) EQv)
      as [ IHa1v IHa2v ]);
 try (assert (Equal (union vs1 (formulas_vars_bound f))
               (union vs2 (formulas_vars_bound f))) as EQf
      by (apply union_equal_1; assumption));
 try (destruct (IHa ns1 ns2 EQns 
                (union vs1 (formulas_vars_bound f))
                (union vs2 (formulas_vars_bound f)) EQf)
      as [ IHa1f IHa2f ]);
 intuition.
Qed. 

  Lemma type_action_subset:
   forall (ns vs:IdSet.t) (r:id) (es:list var) (a:action),
    type_action ns vs r es a ->
    forall (vs':IdSet.t),
     Subset vs vs' ->
      type_action ns vs' r es a.
  Proof.
   intros ns vs r es a. revert a ns vs r es.
   simple induction a; clear a; simpl;
   unfold exist_vars;
   [ idtac
   | intros fr tr ffs m tfs a IHa
   | intros ar v nt a IHa
   | intros ar v m a IHa
   | intros ar v m a IHa
   | intros ar f a IHa ];
   intros ns vs r es;
   [ idtac
   | case (ME.eq_dec r fr); [intro EQ | intro NEQ;
     case (ME.eq_dec r tr);
     intro EQ ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]];
   intros TAvs vs' SUB;
   intuition;
   try (apply IHa with vs; assumption);
   try (apply type_msg_send_subset with vs; assumption);
   try (apply type_msg_recv_subset with vs; assumption);
   try (apply IHa with (union vs (message_vars_bound m));
   [ assumption | unfold Subset; intro; set_iff; intuition ]);
   try (apply IHa with (add (varid v) vs);
   [ assumption | unfold Subset; intro; set_iff; intuition ]).

   apply subset_trans with vs; assumption.
   apply subset_trans with vs; assumption.
   apply type_msg_send_subset with (union vs (formulas_vars_bound ffs));
   [ assumption | unfold Subset; intro; set_iff; intuition ].
   apply IHa with (union vs (formulas_vars_bound ffs));
   [ assumption | unfold Subset; intro; set_iff; intuition ].
   apply subset_trans with vs; assumption.
   apply type_msg_send_subset with (union vs (formulas_vars_bound ffs));
   [ assumption | unfold Subset; intro; set_iff; intuition ].
   apply IHa with (union vs (formulas_vars_bound ffs));
   [ assumption | unfold Subset; intro; set_iff; intuition ].
   apply subset_trans with (union vs (message_vars_bound m));
   [ assumption | unfold Subset; intro; set_iff; intuition ].
   apply IHa with (union (union vs (message_vars_bound m))
     (formulas_vars_bound tfs));
   [ assumption | unfold Subset; intro; set_iff; intuition ].
   apply subset_trans with vs; assumption.
   apply IHa with (union vs (formulas_vars_bound f));
   [ assumption | unfold Subset; intro; set_iff; intuition ].
  Qed.

Hint Resolve type_action_subset.

  Lemma type_spec_rspec_find:
   forall (r:id) (rss:list role_spec) (a:action),
    type_spec r (Spec rss a) ->
    exists bs:list var, exists es:list var,
    rspec_find r rss = Some (RSpec r bs es).
  Proof.
   simpl. unfold type_rspecs. intros.
   decompose [and] H. exact H0.
  Qed.

Hint Resolve type_spec_rspec_find.

(* Connected *)
Definition connected_role (mr:option id) (r:id) :=
 match mr with
 | None => True
 | Some r' => r = r'
 end.

Lemma connected_role_dec:
 forall (mr:option id) (r:id),
  {connected_role mr r} + {~ connected_role mr r}.
Proof.
 intros mr r.
 destruct mr as [| r']; simpl.
 apply id_eq_dec.
 left. trivial.
Qed.

Fixpoint connected_action mr a {struct a} :=
 match a with
 | Return => 
   True
 | Comm f t ffs m tfs a =>
   connected_role mr f
   /\ connected_action (Some t) a
 | New ar v nt a =>
   connected_role mr ar
   /\ connected_action (Some ar) a
 | Bind ar v m a =>
   connected_role mr ar
   /\ connected_action (Some ar) a
 | Match ar v m a =>
   connected_role mr ar
   /\ connected_action (Some ar) a
 | Derive ar f a =>
   connected_role mr ar
   /\ connected_action (Some ar) a
 end.

Lemma connected_action_dec :
 forall (mr:option id) (a:action),
  {connected_action mr a} + {~ connected_action mr a}.
Proof.
 intros mr a. revert a mr.
 simple induction a; clear a;
 simpl;
 [ idtac
 | intros fr tr ffs m tfs a IHa
 | intros ar v nt a IHa
 | intros ar v m a IHa
 | intros ar v m a IHa
 | intros ar f a IHa ];
 intros mr;
 [ left; trivial
 | destruct (connected_role_dec mr fr) as [ CR | NCR ];
   [ destruct (IHa (Some tr)) as [ CA | NCA ];
     [ left; split; assumption
     | right; intros [ F1 F2 ]; contradiction ]
   | right; intros [ F1 F2 ]; contradiction ]
 | idtac
 | idtac
 | idtac
 | idtac ];
 (destruct (connected_role_dec mr ar) as [ CR | NCR ];
  [ destruct (IHa (Some ar)) as [ CA | NCA ];
    [ left; split; assumption
    | right; intros [ F1 F2 ]; contradiction ]
  | right; intros [ F1 F2 ]; contradiction ]).
Qed.

Definition connected_spec s :=
 match s with
 | Spec _ a =>
   connected_action None a
 end.

Lemma connected_spec_dec :
 forall (s:spec),
  {connected_spec s} + {~ connected_spec s}.
Proof.
 intros s. destruct s as [ rss a ].
 simpl.
 apply connected_action_dec.
Qed.

(* Well formed *)
Definition wellformed_spec s :=
 type_spec_all s
 /\ connected_spec s.

Hint Unfold wellformed_spec.

Lemma wellformed_spec_dec :
 forall (s:spec),
  {wellformed_spec s} + {~ wellformed_spec s}.
Proof.
 intros s.
 unfold wellformed_spec.
 destruct (type_spec_all_dec s) as [ TS | NTS ].
 destruct (connected_spec_dec s) as [ CS | NCS ].
 left; split; assumption.
 right; intros [ F1 F2 ]; contradiction.
 right; intros [ F1 F2 ]; contradiction.
Qed. 

(* Semantics *)
 Definition nf := @nil formula.
 Definition role_chan i := Var (role_chanid i) Channel.

Hint Unfold nf.
Hint Unfold role_chan.

 Inductive eval_action : 
 id -> list var -> t -> action ->
 strand -> t -> Prop :=
 | FAIL :
   forall (r:id) (es:list var) (cs:t)
          (a:action),
    eval_action r es cs a (faildesc :: nil) empty
 | RETURN :
   forall (r:id) (es:list var) (cs:t),
    eval_action r es cs
     Return
     (((Outgoing, (Join (MConst C.RETURN) 
                  (Join (MVar ai) (vars_msg es)))), nf)
	    :: nil)
      empty
 | COMM_SEND_CONNECTED :
   forall (r:id) (es:list var) (cs:t)
          (to:id) (ffs tfs:list formula) 
          (m:msg) (a:action) (s:strand) (v:t),
   In to cs -> to <> r ->
   (eval_action r es (add to cs) a s v) ->
    let s' := 
      ((Outgoing, (Join (MConst C.MSG)
        (Join (MVar (role_chan to)) m))), ffs)
      :: s
    in
    eval_action r es cs
                (Comm r to ffs m tfs a)
                s' v
 | COMM_CONNECT :
   forall (r:id) (es:list var) (cs:t)
          (to:id) (ffs tfs:list formula)
          (m:msg) (a:action) (s:strand) (v:t),
   ~ In to cs -> to <> r ->
   let x := role_chan to in
   eval_action r es (add to cs) a s v ->
    let s' := (Outgoing, (MConst C.CONNECT), 
              (Formula C.CONNECT 
               ((DataV (Var to Name)) :: 
                (LogicV (role_chan to)) :: nil) :: nil))
              :: ((Outgoing, 
                  (Join (MConst C.MSG)
                    (Join (MVar (role_chan to)) m))), ffs)
              :: s
    in
    eval_action r es cs
                (Comm r to ffs m tfs a)
                s' v
 | COMM_RECV_CONNECTED :
   forall (r:id) (es:list var) (cs:t)
          (from:id) (ffs tfs:list formula)
          (m:msg) (a:action) (s:strand) (v:t),
   In from cs -> from <> r ->
   let x := role_chan from in
   (eval_action r es (add from cs) a s v) ->
    let s' := ((Incoming,
 	        (Join (MConst C.MSG)
 		 (Join (MVar x) m))), tfs) :: s
    in
    eval_action r es cs
                (Comm from r ffs m tfs a)
                s' v
  | COMM_ACCEPT :
   forall (r:id) (es:list var) (cs:t)
          (from:id) (ffs tfs:list formula)
          (m:msg) (a:action) (s:strand) (v:t),
   ~ In from cs -> from <> r ->
   let x := role_chan from in
   (eval_action r es (add from cs) a s v) ->
    let s' := (Outgoing, (MConst C.ACCEPT), 
               (Formula C.ACCEPT 
                ((LogicV (role_chan from)) :: nil)) :: nil)
              :: ((Incoming,
 	        (Join (MConst C.MSG)
 		 (Join (MVar x) m))), tfs)
              :: s
    in
    eval_action r es cs
                (Comm from r ffs m tfs a)
                s' v
  | COMM_NEXT :
    forall (r:id) (es:list var) (cs:t)
           (to from:id) (ffs tfs:list formula)
           (m:msg) (a:action) (s:strand) (v:t),
     from <> r /\ to <> r ->
     eval_action r es cs a s v ->
     eval_action r es cs (Comm from to ffs m tfs a) s v
  | NEW_NONCE :
    forall (r:id) (es:list var) (cs:t)
           (x:var) (a:action) (s:strand) (v:t),
      eval_action r es cs a s v ->
       let s' := (Outgoing, (MConst C.NEW_NONCE),
                   (Formula C.NEW_NONCE ((LogicV x) :: nil))::nil)
                 :: s in
        eval_action r es cs (New r x New_nonce a)
                    s' (add (varid x) v)
  | NEW_SYMKEY :
    forall (r:id) (es:list var) (cs:t)
           (x:var) (a:action) (s:strand) (v:t),
      eval_action r es cs a s v ->
       let s' := (Outgoing, (MConst C.NEW_SYMKEY),
                   (Formula C.NEW_SYMKEY ((LogicV x) :: nil))::nil)
                 :: s in
        eval_action r es cs (New r x New_symkey a)
                    s' (add (varid x) v)
  | NEW_PUBKEY :
    forall (r:id) (es:list var) (cs:t)
           (x:var) (a:action) (s:strand) (v:t),
      eval_action r es cs a s v ->
       let s' := (Outgoing, (MConst C.NEW_PUBKEY),
                   (Formula C.NEW_PUBKEY ((LogicV x) :: nil))::nil)
                 :: s in
        eval_action r es cs (New r x New_pubkey a)
                    s' (add (varid x) v)
  | NEW_NEXT :
    forall (r:id) (es:list var) (cs:t)
           (ar:id) (x:var) (nt:new_type) (a:action)
           (s:strand) (v:t),
     ar <> r ->
     eval_action r es cs a s v ->
     eval_action r es cs (New ar x nt a) s v
  | BIND :
    forall (r:id) (es:list var) (cs:t)
           (x:var) (m:msg) (a:action) (s:strand) (v:t),
      eval_action r es cs a s v ->
      let s' := (Outgoing, (MConst C.BIND),
                 (Formula C.BIND ((LogicV x) :: nil))::nil)
                :: strand_replace x m s in                  
       eval_action r es cs (Bind r x m a) s' v
  | BIND_NEXT :
    forall (r:id) (es:list var) (cs:t)
           (ar:id) (x:var) (m:msg) (a:action) (s:strand) (v:t),
     ar <> r ->
     eval_action r es cs a s v ->
     eval_action r es cs (Bind ar x m a) s v
  | MATCH :
    forall (r:id) (es:list var) (cs:t)
           (x:var) (m:msg) (a:action) (s:strand) (v:t),
      eval_action r es cs a s v ->
       let s' := ((Outgoing, (Join (MConst C.MATCH) (MVar x))), nf)
	  	 :: ((Incoming, (Join (MConst C.MATCH) m)), nf)	
                 (* XXX Use strand_replace on the rest? *)
		 :: s
       in
        eval_action r es cs (Match r x m a) s' v
  | MATCH_NEXT :
    forall (r:id) (es:list var) (cs:t)
           (ar:id) (x:var) (m:msg) (a:action) (s:strand) (v:t),
     ar <> r ->
     eval_action r es cs a s v ->
     eval_action r es cs (Match ar x m a) s v
  | ASSERT :
    forall (r:id) (es:list var) (cs:t)
           (phi:list formula) (a:action) (s:strand) (v:t),
      eval_action r es cs a s v ->
       let s' := ((Outgoing, MConst C.ASSERT), phi) :: s in
        eval_action r es cs (Derive r phi a) s' v
  | ASSERT_NEXT :
    forall (r:id) (es:list var) (cs:t)
           (ar:id) (phi:list formula) (a:action) (s:strand) (v:t),
     ar <> r ->
     eval_action r es cs a s v ->
     eval_action r es cs (Derive ar phi a) s v.
Hint Constructors eval_action.

 Inductive eval_spec : 
 id -> const -> spec -> strand -> t -> Prop :=
 | SPEC_NONE :
   forall (r:id) (rc:const)
          (rss:list role_spec) (a:action) (s:strand) (v:t),
    rspec_find r rss = None ->
    eval_spec r rc (Spec rss a) (faildesc :: nil) empty
 | SPEC_SOME :
   forall (r:id) (rc:const)
          (rss:list role_spec) (bs es:list var)
          (a:action) (s:strand) (v:t),
    rspec_find r rss = Some (RSpec r bs es) ->
     (eval_action r es empty a s v) ->
     let s' := ((Incoming, (call_msg rc bs)), nf) :: s in
      (eval_spec r rc (Spec rss a) s' v).

Hint Constructors eval_spec.

(* Compiler *)
  Definition let_connect (r:id) (cs:IdSet.t) (c:cont) :=
    if mem r cs then
      c
    else
      Let (role_chan r) (Connect (Var r Name)) c .
 
  Lemma let_connect_type_notin:
   forall (r:id) (cs:IdSet.t) (c:cont) (vs:IdSet.t),
    ~ In r cs -> In r vs ->
    type_cont (add (varid (role_chan r)) vs) c ->
    type_cont vs (let_connect r cs c).
  Proof.
   intros r cs c vs NINc INv.
   unfold let_connect.
   case_eq (mem r cs).
   rewrite <- mem_iff. intro INc. contradiction.
   intro X. clear X.
   simpl. intro Tc.
   split. split.
   case (type_eq_dec Channel Channel); intuition.
   split. unfold exist_var. simpl.
   assumption.
   case (type_eq_dec Name Name); intuition.
   assumption.
  Qed.

Hint Resolve let_connect_type_notin.

  Lemma let_connect_type_in:
   forall (r:id) (cs:IdSet.t) (c:cont) (vs:IdSet.t),
    In r cs ->
    type_cont vs c ->
    type_cont vs (let_connect r cs c).
  Proof.
   intros r cs v vs INc.
   unfold let_connect.
   case_eq (mem r cs).
   intro X. clear X.
   trivial.
   rewrite <- not_mem_iff. intro NINc. contradiction.
  Qed.

Hint Resolve let_connect_type_in.

  Definition let_accept (r:id) (cs:IdSet.t) (c:cont) :=
    if mem r cs then
      c
    else
      Let (role_chan r) Accept c .
 
  Lemma let_accept_type_notin:
   forall (r:id) (cs:IdSet.t) (c:cont) (vs:IdSet.t),
    ~ In r cs -> 
    type_cont (add (varid (role_chan r)) vs) c ->
    type_cont vs (let_accept r cs c).
  Proof.
   intros r cs c vs NINc.
   unfold let_accept.
   case_eq (mem r cs).
   rewrite <- mem_iff. intro INc. contradiction.
   intro X. clear X.
   simpl. intro Tc.
   split. 
   case (type_eq_dec Channel Channel); intuition.
   assumption.
  Qed.

Hint Resolve let_accept_type_notin.

  Lemma let_accept_type_in:
   forall (r:id) (cs:IdSet.t) (c:cont) (vs:IdSet.t),
    In r cs ->
    type_cont vs c ->
    type_cont vs (let_accept r cs c).
  Proof.
   intros r cs v vs INc.
   unfold let_accept.
   case_eq (mem r cs).
   intro X. clear X.
   trivial.
   rewrite <- not_mem_iff. intro NINc. contradiction.
  Qed.

Hint Resolve let_accept_type_in.

  Fixpoint compile_action r es cs (a:action) {struct a} :=
    match a with
      | Return =>
	(CPPL.Return nf es)
      | Comm f t ffs m tfs a' =>
	if ME.eq_dec r f then
	  let tv := role_chan t in
	    let c' := compile_action r es (add t cs) a' in
	      let_connect t cs
              (Send ffs tv m c' Fail)
	  else if ME.eq_dec r t then
	    let fv := role_chan f in
	      let c' := compile_action r es (add f cs) a' in
		let_accept f cs		
		(Recv fv m tfs c' Fail)
	    else
	      compile_action r es cs a'
      | New ar v nt a' =>
	if ME.eq_dec r ar then
	  let c' := compile_action r es cs a' in
	    let lv :=
	      match nt with
		| New_nonce => CPPL.New_nonce
		| New_symkey => CPPL.New_symkey
		| New_pubkey => CPPL.New_pubkey
	      end
	      in
	      Let v lv c'
	  else
	    compile_action r es cs a'
      | Bind ar v m a' =>
	let c' := compile_action r es cs a' in
	  if ME.eq_dec r ar then
	    Let v (LVMsg m) c'
	    else
	      c'
      | Match ar v m a' =>
	let c' := compile_action r es cs a' in
	  if ME.eq_dec r ar then	    
   	    CPPL.Match v m nf c' Fail
	    else
	      c'
      | Derive ar f a' =>
	let c' := compile_action r es cs a' in
	  if ME.eq_dec r ar then
	    CPPL.Derive f c' Fail
	    else
	      c'
    end .

  Lemma compile_action_fail:
   forall (r:id) (es:list var) (cs:IdSet.t) (a:action),
    compile_action r es cs a <> Fail.
  Proof.
   intros r es cs a. generalize a r es cs. clear a r es cs.
   simple induction a; clear a; simpl;
   [ idtac
   | intros fr tr ffs m tfs a IHa
   | intros ar v nt a IHa
   | intros ar v m a IHa
   | intros ar v m a IHa
   | intros ar f a IHa ];
   intros r es cs;
   [ idtac
   | case (ME.eq_dec r fr); [intro EQ | intro NEQ;
     case (ME.eq_dec r tr); intro EQ ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]];
   simpl;
   try (discriminate);
   try (apply IHa; assumption);
   unfold let_accept, let_connect;
   [ case (mem tr cs)
   | case (mem fr cs) ];
   discriminate.
  Qed.

Hint Resolve compile_action_fail.

  Lemma compile_action_type:
   forall (ns vs:IdSet.t) (r:id) (es:list var) (a:action),
    (forall a:id, In a ns -> In (varid (role_chan a)) vs) ->
    type_action ns vs r es a ->
    type_cont vs (compile_action r es ns a).
  Proof.
   intros ns vs r es a. revert a ns vs r es.
   simple induction a; clear a; simpl;
   [ idtac
   | intros fr tr ffs m tfs a IHa
   | intros ar v nt a IHa
   | intros ar v m a IHa
   | intros ar v m a IHa
   | intros ar f a IHa ];
   intros ns vs r es SUB;
   [ idtac
   | case (ME.eq_dec r fr); [intro EQ | intro NEQ;
     case (ME.eq_dec r tr); intro EQ ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]];
   simpl;
   try (unfold exist_vars);
   try (unfold exist_var);
   intro TA;
   try (apply IHa; assumption).

   split. apply subset_trans with vs. assumption.
   intro a. set_iff. intuition.
   intro a. set_iff. intuition.
   
   destruct TA as [ [ InNS | InVS ] [ Sfs [ Tm Ta ] ] ].
   apply let_connect_type_in; eauto. simpl. unfold exist_vars, exist_var.
   destruct (type_eq_dec Channel Channel) as [Q | Q]. set_iff.
   simpl. split. trivial. split. eauto. split. trivial. split; eauto.
   split; eauto. apply IHa. intros a'. set_iff. clear Q.
   intros [Q | IN]; subst; eauto. trivial.
   absurd (Channel = Channel); auto.

   destruct (In_dec tr ns) as [ InNS | NInNS ].
   apply let_connect_type_in; eauto. simpl. unfold exist_vars, exist_var.
   destruct (type_eq_dec Channel Channel) as [Q | Q]. set_iff.
   simpl. split; eauto. split; eauto. split; eauto. split; eauto.
   split; eauto. apply IHa. intros a'. set_iff. clear Q.
   intros [Q | IN]; subst; eauto. trivial.
   absurd (Channel = Channel); auto.

   apply let_connect_type_notin; eauto. simpl. unfold exist_vars, exist_var.
   destruct (type_eq_dec Channel Channel) as [Q | Q]. set_iff.
   split. intros a'. set_iff. eauto. split; eauto. split; trivial. split. 
   apply type_msg_send_subset with (union vs (formulas_vars_bound ffs)).
   trivial. intros a'. set_iff. intros [IN1 | IN2]; eauto.
   split; trivial. apply IHa. intros a'. set_iff. clear Q.
   intros [Q | IN]; subst; eauto. 
   apply type_action_subset with (union vs (formulas_vars_bound ffs)); eauto.
   intros a'. set_iff. intros [IN1 | IN2]; eauto.
   absurd (Channel = Channel); auto.

   destruct TA as [ Tm [ Sfs Ta ] ].
   destruct (In_dec fr ns) as [ InNS | NInNS ].
   apply let_accept_type_in; eauto. simpl. unfold exist_vars, exist_var.
   destruct (type_eq_dec Channel Channel) as [Q | Q];
     [ idtac | absurd (Channel = Channel); auto ]. simpl.
   split; eauto. split; eauto. split; eauto. split; eauto. split; eauto.
   apply IHa. intros a'. set_iff. clear Q. intros [Q | IN]; subst; eauto.
   trivial.

   apply let_accept_type_notin; eauto. simpl. unfold exist_vars, exist_var.
   destruct (type_eq_dec Channel Channel) as [Q | Q];
     [ idtac | absurd (Channel = Channel); auto ]. simpl. set_iff.
   split. intros a'. set_iff. intros IN. apply Sfs in IN. revert IN.
   set_iff. intros [IN1|IN2]; eauto.
   split; eauto. split; eauto. split.
   apply type_msg_recv_subset with vs; eauto.
   intros a'; set_iff; eauto. split.
   apply IHa. intros a'. set_iff. clear Q. intros [Q | IN]; subst; eauto.
   apply type_action_subset with (union (union vs (message_vars_bound m))
     (formulas_vars_bound tfs)); eauto.
   intros a'. set_iff. intros [[N1 | N2]|N3]; eauto. trivial.

   destruct TA as [ Tnt Ta ].
   split.
   destruct nt; simpl; simpl in Tnt; assumption.

   apply IHa;
   [ intro b; set_iff;
     intro InB; right;
     apply SUB; assumption 
   | assumption ].

   destruct TA as [ Tv [ Tm Ta ] ].
   split. split; assumption.
   apply IHa.
   intro b. set_iff. intro InB. right. apply SUB. assumption.
   assumption.

   destruct TA as [ Tv [ InV [ Tm Ta ] ] ].
   split. unfold exist_vars. intro. set_iff. intuition.
   split. assumption.
   split. assumption.
   split. assumption.
   split. apply IHa.
   intro b. intro InB. set_iff. left. left. apply SUB. assumption.
   apply type_action_subset with (union vs (message_vars_bound m)).
   assumption.
   unfold Subset; intro; set_iff; intuition.
   trivial.

   destruct TA as [SUBf TA].
   split. assumption.
   split. apply IHa.
   intro b. intro InB. set_iff. left. apply SUB. assumption.
   assumption. trivial.
  Qed.

Hint Resolve compile_action_type.

  Theorem compile_action_correct:
   forall (r:id) (es:list var) (cs:t) (a:action)
          (s:strand) (v:t),
    eval_action r es cs a s v ->
    eval_cont (compile_action r es cs a) s v.
  Proof.
   intros r es cs a s v.
   revert a r es cs s v.
   simple induction a; clear a; simpl;
   [ idtac
   | intros fr tr ffs m tfs a IHa
   | intros ar x nt a IHa
   | intros ar x m a IHa
   | intros ar x m a IHa
   | intros ar phi a IHa ];
   intros r es cs s v EQeval;
   inversion EQeval; try (subst);
   try (apply CPPL.FAIL).

   apply CPPL.RETURN.

   subst s. unfold let_connect, let_accept.
   destruct (ME.eq_dec fr fr) as [Q|NQ];
     [ clear Q | absurd (fr = fr); auto ].
   case_eq (mem tr cs); [ intros MEM; clear MEM
     | intros NMEM; rewrite <- not_mem_iff in NMEM; contradiction ].
   apply CPPL.SEND_WITH_GUARANTEE. apply IHa. trivial.

   subst s. unfold let_connect, let_accept.
   destruct (ME.eq_dec fr fr) as [Q|NQ];
     [ clear Q | absurd (fr = fr); auto ].
   case_eq (mem tr cs); [ 
        intros MEM; rewrite <- mem_iff in MEM; contradiction
     | intros NMEM; clear NMEM ].
   apply CPPL.LET_CONNECT.
   apply CPPL.SEND_WITH_GUARANTEE. apply IHa. trivial.

   subst v0 s' a0 m0 r cs0 es0 fr r0 ffs0 tfs0.
   unfold x0 in *. clear x0. unfold x in *. clear x. subst s.
   destruct (ME.eq_dec tr from) as [Q|NQ];
     [ subst tr; absurd (from=from); auto | clear NQ ].
   destruct (ME.eq_dec tr tr) as [Q|NQ];
     [ clear Q | absurd (tr = tr); auto ]. unfold let_accept.
   case_eq (mem from cs); 
   [ intro MEM 
   | intro NMEM; rewrite <- not_mem_iff in NMEM; contradiction ].
   apply CPPL.RECEIVE_AND_RELY. apply IHa. trivial.

   case (ME.eq_dec tr fr); 
   [ intro EQ; absurd (fr = tr); [ assumption | symmetry; assumption ]
   | intro NEQ; clear NEQ ]. 
   case (ME.eq_dec tr tr); 
   [ intro EQ; clear EQ 
   | intro NEQ; absurd (tr = tr); [ assumption | reflexivity ] ].
   unfold let_accept.
   case_eq (mem fr cs); 
   [ intro MEM; rewrite <- mem_iff in MEM; contradiction 
   | intro NMEM; clear NMEM ]. subst s.
   apply CPPL.LET_ACCEPT.
   apply CPPL.RECEIVE_AND_RELY. apply IHa. trivial.

   destruct H10 as [ NEQf NEQt ].
   case (ME.eq_dec r fr); 
   [ intro EQ; absurd (fr = r); [ assumption | symmetry; assumption ]
   | intro NEQ; clear NEQ ]. 
   case (ME.eq_dec r tr); 
   [ intro EQ; absurd (tr = r); [ assumption | symmetry; assumption ]
   | intro NEQ; clear NEQ ].
   apply IHa. assumption.

   case (ME.eq_dec ar ar); 
   [ intro EQ; clear EQ
   | intro NEQ; absurd (ar = ar); [ assumption | reflexivity ] ].
   apply CPPL.LET_NEW_NONCE.
   apply IHa. assumption.

   case (ME.eq_dec ar ar); 
   [ intro EQ; clear EQ
   | intro NEQ; absurd (ar = ar); [ assumption | reflexivity ] ].
   apply CPPL.LET_NEW_SYMKEY.
   apply IHa. assumption.

   case (ME.eq_dec ar ar); 
   [ intro EQ; clear EQ
   | intro NEQ; absurd (ar = ar); [ assumption | reflexivity ] ].
   apply CPPL.LET_NEW_PUBKEY.
   apply IHa. assumption.

   case (ME.eq_dec r ar); 
   [ intro EQ; absurd (ar = r); [ assumption | symmetry; assumption ]
   | intro NEQ; clear NEQ ].
   apply IHa. assumption.

   case (ME.eq_dec ar ar); 
   [ intro EQ; clear EQ
   | intro NEQ; absurd (ar = ar); [ assumption | reflexivity ] ].
   apply CPPL.LET_LVMSG.
   apply IHa. assumption.

   case (ME.eq_dec r ar); 
   [ intro EQ; absurd (ar = r); [ assumption | symmetry; assumption ]
   | intro NEQ; clear NEQ ].
   apply IHa. assumption.

   case (ME.eq_dec ar ar); 
   [ intro EQ; clear EQ
   | intro NEQ; absurd (ar = ar); [ assumption | reflexivity ] ].
   apply CPPL.MATCH_AND_RELY.
   apply IHa. assumption.

   case (ME.eq_dec r ar); 
   [ intro EQ; absurd (ar = r); [ assumption | symmetry; assumption ]
   | intro NEQ; clear NEQ ].
   apply IHa. assumption.

   case (ME.eq_dec ar ar); 
   [ intro EQ; clear EQ
   | intro NEQ; absurd (ar = ar); [ assumption | reflexivity ] ].
   apply CPPL.ASSERT.
   apply IHa. assumption.

   case (ME.eq_dec r ar); 
   [ intro EQ; absurd (ar = r); [ assumption | symmetry; assumption ]
   | intro NEQ; clear NEQ ].
   apply IHa. assumption.
  Qed.

Hint Resolve compile_action_correct.

  Lemma compile_action_returnpath :
   forall (r:id) (es:list var) (cs:IdSet.t) (a:action),
    return_path (compile_action r es cs a).
  Proof.
   intros r es cs a.
   revert a es cs.
   unfold return_path.
   simple induction a; clear a; simpl;
   [ idtac
   | intros fr tr ffs m tfs a IHa
   | intros ar x nt a IHa
   | intros ar x m a IHa
   | intros ar x m a IHa
   | intros ar phi a IHa ];
   intros es cs;
   [ idtac
   | case (ME.eq_dec r fr); [intro EQ | intro NEQ;
     case (ME.eq_dec r tr); intro EQ ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]];
   simpl;
   try (apply IHa).

   apply Arith.Gt.gt_Sn_O.

   unfold let_connect.
   case (mem tr cs); simpl;
   rewrite app_length; simpl;
   rewrite Arith.Plus.plus_0_r;
   apply IHa.

   unfold let_accept.
   case (mem fr cs); simpl; 
   rewrite app_length; simpl;
   rewrite Arith.Plus.plus_0_r;
   apply IHa.

   rewrite app_length; simpl;
   rewrite Arith.Plus.plus_0_r;
   apply IHa.

   rewrite <- app_nil_end.
   apply IHa.
  Qed.

Hint Resolve compile_action_returnpath.

  Lemma compile_action_returnssame :
   forall (r:id) (es:list var) (cs:IdSet.t) (a:action),
    returns_same (compile_action r es cs a).
  Proof.
   intros r es cs a.
   revert a es cs.
   unfold returns_same.
   simple induction a; clear a; simpl;
   [ idtac
   | intros fr tr ffs m tfs a IHa
   | intros ar x nt a IHa
   | intros ar x m a IHa
   | intros ar x m a IHa
   | intros ar phi a IHa ];
   intros es cs;
   [ idtac
   | case (ME.eq_dec r fr); [intro EQ | intro NEQ;
     case (ME.eq_dec r tr); intro EQ ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]
   | case (ME.eq_dec r ar); [intro EQ | intro NEQ]];
   simpl;
   try (rewrite <- app_nil_end);
   try (apply IHa).

   reflexivity.

   unfold let_connect.
   case (mem tr cs);
   unfold returns_same; simpl;
   set (crs := cont_returns (compile_action r es (add tr cs) a));
   (cut (crs ++ nil = crs);
   [ intro EQc;
     rewrite EQc;
     unfold crs; apply IHa
   | symmetry; apply app_nil_end ]).

   unfold let_accept.
   case (mem fr cs);
   unfold returns_same; simpl;
   set (crs := cont_returns (compile_action r es (add fr cs) a));
   (cut (crs ++ nil = crs);
   [ intro EQc;
     rewrite EQc;
     unfold crs; apply IHa
   | symmetry; apply app_nil_end ]).
  Qed.

Hint Resolve compile_action_returnssame.

  Lemma eval_action_type :
   forall (sig:t) (s:strand) (v:t)
          (ns:t) (r:id) (es:list var) (a:action),
    (forall a:id, In a ns -> In (varid (role_chan a)) sig) ->
    type_action ns sig r es a ->
    In (varid pr) sig ->
    In (varid ai) sig ->
     eval_action r es ns a s v ->
     type_strand sig s .
  Proof.
   intros sig s v ns r es a.
   intros NVS Ta InPR InAI EA.
   apply eval_cont_type with (compile_action r es ns a) v.
   assumption. assumption.
   apply compile_action_type. assumption.
   assumption.
   apply compile_action_correct.
   assumption.
  Qed.

Hint Resolve eval_action_type.

  Definition compile_spec r rc s :=
    match s with
      | Spec rss a =>
	match rspec_find r rss with
	  | None => None
	  | Some (RSpec _ bs es) =>
	    let c := compile_action r es empty a in
	      Some (Proc rc bs nf c)
	end
    end .
  
  Theorem compile_spec_type :
    forall (r:id) (rc:const) (s:spec),
      type_spec r s ->
       exists p:proc,
        compile_spec r rc s = Some p /\
        type_proc p .
  Proof.
   simple destruct s. clear s.
   intros rss a.
   simpl. 
   intros [ Trs RF ].
   unfold type_rspecs in Trs.
   destruct Trs as [bs [es RF']].
   rewrite RF' in *.
   exists (Proc rc bs nf (compile_action r es empty a)). 
   split. reflexivity.
   simpl.
   split. unfold exist_vars. intro. set_iff. intuition.
   split.
   apply compile_action_type.
   unfold Subset; intro; set_iff; intuition.
   apply type_action_subset with (varlist_set bs).
   assumption.
   unfold Subset; intro; set_iff; intuition.
   split.
   apply compile_action_returnpath.
   apply compile_action_returnssame.
  Qed.

Hint Resolve compile_spec_type.

  Theorem compile_spec_correct:
   forall (r:id) (rc:const) (sp:spec) (p:proc)
          (s:strand) (v:t),
    eval_spec r rc sp s v ->
    compile_spec r rc sp = Some p ->
    eval_proc p s v.
  Proof.
   intros r rc sp p s v EV CS.
   inversion EV. subst.
   inversion CS.
   revert H H1.
   case_eq (rspec_find r rss).
   intros r0 EQr EQs. discriminate EQs.
   intros EQr EQs EQs'. discriminate EQs'.

   subst.
   inversion CS.
   revert H H2.
   case_eq (rspec_find r rss).
   destruct r0 as [ r' bs' es' ].
   intros EQr EQs EQp.
   clear EQr. inversion EQs. subst. clear EQs.
   inversion EQp. subst. clear EQp.
   apply CPPL.PROC.
   apply compile_action_correct.
   assumption.
   intros EQr EQs EQp.
   discriminate EQp.
  Qed.

Hint Resolve compile_spec_correct.

  Theorem eval_spec_type :
   forall (r:id) (sp:spec),
    type_spec r sp ->
     forall (sig:t) (rc:const) (s:strand) (v:t),
      eval_spec r rc sp s v ->
      type_strand sig s.
  Proof.
   intros r sp Ts sig rc s v ES.
   destruct (compile_spec_type r rc sp) as [p [EQ Tp]].
   assumption.
   apply eval_proc_type with p v.
   assumption.
   apply compile_spec_correct with r rc sp.
   assumption. assumption.
  Qed.

Hint Resolve eval_spec_type.