(* CPPL Compiler
   Copyright (C) 2008 Jay McCarthy
*)

open VComp.CPPL
open VComp.Common
open Format
open Printer

let cmp_var ppf (Var (id, _)) =
  fprintf ppf "\"%a\""
    pr_id id

let cmp_list cmp_elem ppf es =
  let rec h ppf = function
      VComp.List.Coq_nil -> 
	pp_print_string ppf "[]"
    | VComp.List.Coq_cons (e, es) ->
	fprintf ppf "%a::%a" 
	  cmp_elem e
	  h es
  in
    h ppf es

let cmp_vars = cmp_list cmp_var

let cmp_fvars ppf fs =
  cmp_vars ppf
    (VComp.List.map (function | DataV v -> v | LogicV v -> v) fs)

let cmp_formula ppf = function
  | Formula (id, vs) ->
      fprintf ppf "@[<1>(Formula@ \"%a\"@ (%a))@]"
	pp_print_string id
	cmp_fvars vs

let rec cmp_msg ppf = function
  | MNil ->
      pp_print_string ppf "PNil"
  | MVar v ->
      fprintf ppf "@[<1>(PVar@ %a)@]"
	cmp_var v
  | MConst s ->
      fprintf ppf "@[<1>(PConst@ %a)@]"
	pp_print_string (Printf.sprintf "%S" s)
  | Join (m1, m2) ->
      fprintf ppf "@[<1>(PJoin@ (%a,@ %a))@]"
	cmp_msg m1
	cmp_msg m2
  | Hash m ->
      fprintf ppf "@[<1>(PHash@ %a)@]"
	cmp_msg m
  | SymEnc (v, m) ->
      fprintf ppf "@[<1>(PSymEnc@ (%a,@ %a))@]"
	cmp_var v
	cmp_msg m
  | SymSign (v, m) ->
      fprintf ppf "@[<1>(PSymSign@ (%a,@ %a))@]"
	cmp_var v
	cmp_msg m
  | PubEnc (v, m) ->
      fprintf ppf "@[<1>(PPubEnc@ (%a,@ %a))@]"
	cmp_var v
	cmp_msg m
  | PubSign (v, m) ->
      fprintf ppf "@[<1>(PPubSign@ (%a,@ %a))@]"
	cmp_var v
	cmp_msg m
  | Bind (v, m) ->
      fprintf ppf "@[<1>(PBind@ (%a,@ %a))@]"
	cmp_var v
	cmp_msg m

let cmp_formulas = cmp_list cmp_formula

let cmp_letval v ppf = function
  | New_nonce ->
      fprintf ppf "@[<1>(LetNewNonce %a)@]"
	cmp_var v
  | New_symkey ->
      fprintf ppf "@[<1>(LetNewSymkey %a)@]"
	cmp_var v
  | New_pubkey ->
      fprintf ppf "@[<1>(LetNewPubkey %a)@]"
	cmp_var v
  | Accept ->
      fprintf ppf "@[<1>(LetAccept %a)@]"
	cmp_var v
  | Connect cv ->
      fprintf ppf "@[<1>(LetConnect (%a,%a))@]"
	cmp_var v
	cmp_var cv
  | LVMsg m ->
      fprintf ppf "@[<1>(LetBind (%a,%a))@]"
	cmp_var v
	cmp_msg m

let rec cmp_cont pth ppf = function
  | Fail ->
      fprintf ppf "@[| %a ->@ @[<1>Fail@]@]"
	pp_print_string pth
  | Return (fs,is) ->
      fprintf ppf "@[| %a ->@ @[<1>(Return (%a,@ (%a)))@]@]"
	pp_print_string pth	
	cmp_formulas fs 
	cmp_vars is
  | Let (v, lv, c) ->
      fprintf ppf "@[| %a ->@ %a@]@ %a"
	pp_print_string pth	
	(cmp_letval v) lv
	(cmp_cont (Printf.sprintf "L::%s" pth)) c
  | Derive (f, sk, fk) ->
      fprintf ppf "@[| %a ->@ @[<1>(Assert %a)@]@]@ %a@ %a"
	pp_print_string pth	
	cmp_formulas f
	(cmp_cont (Printf.sprintf "L::%s" pth)) sk
	(cmp_cont (Printf.sprintf "R::%s" pth)) fk
  | Send (f, v, m, sk, fk) ->
      fprintf ppf "@[| %a ->@ @[<1>(Send (%a,@ %a,@ %a))@]@]@ %a@ %a"
	pp_print_string pth	
	cmp_formulas f
	cmp_var v
	cmp_msg m
	(cmp_cont (Printf.sprintf "L::%s" pth)) sk
	(cmp_cont (Printf.sprintf "R::%s" pth)) fk
  | Recv (v, m, f, sk, fk) ->
      fprintf ppf "@[| %a ->@ @[<1>(Recv (%a,@ %a,@ %a))@]@]@ %a@ %a"
	pp_print_string pth	
	cmp_var v
	cmp_msg m
	cmp_formulas f
	(cmp_cont (Printf.sprintf "L::%s" pth)) sk
	(cmp_cont (Printf.sprintf "R::%s" pth)) fk
  | Call _ ->
      failwith "Can't handle Call"
  | Match (v, m, f, sk, fk) ->
      fprintf ppf "@[| %a ->@ @[<1>(Match (%a,@ %a,@ %a))@]@]@ %a@ %a"
	pp_print_string pth	
	cmp_var v
	cmp_msg m
	cmp_formulas f
	(cmp_cont (Printf.sprintf "L::%s" pth)) sk
	(cmp_cont (Printf.sprintf "R::%s" pth)) fk

let compile_proc proc = 
  let Proc (p,xs,phi,c) = proc in
    fprintf std_formatter
      "open Cppl@\n@\n@[<1>let %a : protocol@ = @[<1>{@[args = %a;@]@ @[assumptions = %a;@]@ @[handler =@ function@\n%a@\n| _ -> Fail;}@]@]@];;@\n"
      pp_print_string p
      cmp_vars xs
      cmp_formulas phi
      (cmp_cont "[]") c
