(* Inference
   Copyright (C) 2008 Jay McCarthy
*)

open VComp.CPPL (* Also binds 'Return' *)
open VComp.Common (* Also binds 'Bind' *)
open VComp.WPPL 
open VComp.List

let the_prefix = ref ""

let cst s = VComp.IdSet.IOT.str (Printf.sprintf "%s%s" !the_prefix s)
let tsc = VComp.IdSet.IOT.rts

(* HELPERS *)
let id_type_to_coq_type t = 
  match t with
  | Ast_t.Name _ -> Name
  | Ast_t.Text _ -> Text
  | Ast_t.Msg _ -> Msg
  | Ast_t.Nonce _ -> Nonce
  | Ast_t.Pubkey _ -> Pubkey
  | Ast_t.Symkey _ -> Symkey
  | Ast_t.Channel _ -> Channel

let new_type_to_coq_new_type = function
  | Ast_t.New_nonce _ -> New_nonce
  | Ast_t.New_symkey _ -> New_symkey
  | Ast_t.New_pubkey _ -> New_pubkey

let decls2vars l =
  List.fold_left
    (fun a d ->
      let (id, t) = d in
      let (_, ids) = id in
      let t' = id_type_to_coq_type t in
      let v = Var ((cst ids), t') in
      Coq_cons (v,a))
    Coq_nil
    l

(* WPPL INFERENCE *)
open Pos
open Env
open Ast_ut

let type_id renv id =
  let (_, ids) = id in
  try 
    match (PMap.find ids renv) with
	None ->
	  failwith (Printf.sprintf "Cannot type %s" ids)
      | Some t' ->
	  t'
  with
      Not_found ->
	failwith (Printf.sprintf "Unknown id: %s" ids)

let update_decls renv decls =
 List.fold_left
    (fun new_decls d ->
      let (id, _) = d in
      let t' = type_id renv id in
	(id, t') :: new_decls)
    []
    decls

let infer_var renv d mt =
  let (id, _) = d in
  let renv' = extend_decl cmb_merge d renv in
  let d' = id, mt in
  let renv'' = extend_decl cmb_merge d' renv' in
  let t' = type_id renv'' id in
  let (_,ids) = id in
  let v = (Var ((cst ids), (id_type_to_coq_type t'))) in
    renv'', v

let infer_fvar renv fd mt =
  match fd with
    | UDataV d ->
	let renv', v = infer_var renv d mt in
	  renv', (DataV v)
    | ULogicV d ->
	let renv', v = infer_var renv d mt in
	  renv', (LogicV v)

let rec infer_msg renv msg =
  match msg with
      UDecl d ->
	let renv', v = infer_var renv d None in	
	  renv', MVar v
    | UCat (m1, m2) ->
	let renv', m1' = infer_msg renv m1 in
	let renv'', m2' = infer_msg renv' m2 in
	  renv'', Join (m1', m2')
    | UConst (_, c) ->
	renv, MConst c
    | UPubEncrypt (m, d) ->
	let renv', m' = infer_msg renv m in
	let renv'', v = infer_var renv' d (Some (Ast_t.Pubkey empty_pos)) in
	  renv'', PubEnc (v, m')
    | UPubSign (m, d) ->
	let renv', m' = infer_msg renv m in
	let renv'', v = infer_var renv' d (Some (Ast_t.Pubkey empty_pos)) in
	  renv'', PubSign (v, m')
    | USymEncrypt (m, d) ->
	let renv', m' = infer_msg renv m in
	let renv'', v = infer_var renv' d (Some (Ast_t.Symkey empty_pos)) in
	  renv'', SymEnc (v, m')
    | USymSign (m, d) ->
	let renv', m' = infer_msg renv m in
	let renv'', v = infer_var renv' d (Some (Ast_t.Symkey empty_pos)) in
	  renv'', SymSign (v, m')
    | UHash m ->
	let renv', m' = infer_msg renv m in
	  renv', Hash m'
    | UMBind (v, m) ->
	let renv', v' = infer_var renv v (Some (Ast_t.Msg empty_pos)) in
	let renv'', m' = infer_msg renv' m in
	  renv'', VComp.Common.Bind (v', m')

(* XXX fix *)
let update_denv_msg denv msg = denv

let infer_formula denv renv = function
  | UTerm (id, uds) ->
      let denv', renv', vs =
	List.fold_left
	  (fun (denv', renv', vs) ud ->
	    let renv'', fv = infer_fvar renv' ud None in
	      (* XXX denv' update *)
	      denv', renv'', (Coq_cons (fv, vs)))
	  (denv, renv, Coq_nil)
	  uds
      in
      let (_,ids) = id in
	denv', renv', Formula (ids, (rev vs))

let rec infer_formulas denv renv = function
  | [] -> denv, renv, Coq_nil
  | f :: fs ->
      let denv', renv', f' = infer_formula denv renv f in
      let denv'', renv'', fs' = infer_formulas denv renv' fs in
	denv'', renv'', (Coq_cons (f', fs'))

let rec infer_action denv renv = function
  | UZero ->
      renv, Return
  | UComm (from_r, to_r, ffs, msg, tfs, a) ->
      let (_,from_rs) = from_r in
      let (_,to_rs) = to_r in
      let renv', _ =
	infer_var renv (to_r, None) (Some (Ast_t.Name empty_pos)) in
      let renv'', msg' = infer_msg renv' msg in
      let denv1, renv3, ffs' = infer_formulas denv renv'' ffs in
      let denv2, renv4, tfs' = infer_formulas denv1 renv3 tfs in
      let renv5, a' = infer_action denv2 renv4 a in
	renv5, Comm ((cst from_rs), (cst to_rs), ffs', msg', tfs', a')
  | UNew (r, d, nt, a) ->
      let (_,rs) = r in
      let t = Ast_t.new_type_to_id_type nt in
      let renv', v = infer_var renv d (Some t) in	
      let renv'' = extend_decl cmb_merge d renv' in
      let denv' = pmap_add_decl d () denv in
      let renv''', a' = infer_action denv' renv'' a in
      let nt' = new_type_to_coq_new_type nt in
	renv''', New ((cst rs), v, nt', a')
  | UBind (r, d, msg, a) ->
      let (_,rs) = r in
      let renv', v = infer_var renv d (Some (Ast_t.Msg empty_pos)) in
      let renv'', msg' = infer_msg renv' msg in
      let denv' = pmap_add_decl d () denv in
      let renv''', a' = infer_action denv' renv'' a in
	renv''', Bind ((cst rs), v, msg', a')
  | UMatch (r, d, msg, a) ->
      let (_,rs) = r in
      let renv', v = infer_var renv d (Some (Ast_t.Msg empty_pos)) in
      let renv'', msg' = infer_msg renv' msg in
      let renv''', a' = infer_action denv renv'' a in
	renv''', Match ((cst rs), v, msg', a')
  | UDerive (r, t, a) ->
      let (_,rs) = r in
      let denv', renv', t' = infer_formulas denv renv t in
      let renv'', a' = infer_action denv' renv' a in
	renv'', Derive ((cst rs), t', a')

let infer_spec (us : unspec) : spec =
  let (rspecs, a) = us in
  let renv =
    List.fold_left
      (fun renv rspec ->
	let (_, start_decls, end_decls) = rspec in
	let renv' = extend_decls cmb_merge start_decls renv in
	  extend_decls cmb_merge end_decls renv')
      PMap.empty
      rspecs
  in
  let denv =
    List.fold_left
      (fun denv rspec ->
	let (_, start_decls, _) = rspec in
	  add_decls start_decls denv)
      PMap.empty
      rspecs
  in
  let renv', a' = infer_action denv renv a in
  let rspecs' =
    rev
      (List.fold_left
	  (fun new_rspecs rspec ->
	    let (role, start_decls, end_decls) = rspec in
	    let (_,rs) = role in
	    let start_decls' = update_decls renv' start_decls in
	    let end_decls' = update_decls renv' end_decls in
	      Coq_cons ((RSpec ((cst rs), (decls2vars start_decls'),
			       (decls2vars end_decls'))),
		       new_rspecs))
	  Coq_nil
	  rspecs)
  in
  let s : spec = Spec (rspecs', a') in
    s

(* CPPL Inference *)
let rec infer_vars renv = function
  | [] -> renv, Coq_nil
  | uv :: uvs ->
      let renv', v = infer_var renv uv None in
      let renv'', vs = infer_vars renv' uvs in
	renv'', Coq_cons (v, vs)

let infer_letval renv = function
  | UNew_nonce ->
      renv, VComp.CPPL.New_nonce, (Ast_t.Nonce empty_pos)
  | UNew_symkey ->
      renv, VComp.CPPL.New_symkey, (Ast_t.Symkey empty_pos)
  | UNew_pubkey ->
      renv, VComp.CPPL.New_pubkey, (Ast_t.Pubkey empty_pos)
  | UAccept ->
      renv, Accept, (Ast_t.Channel empty_pos)
  | UConnect ud -> 
      let renv', d = infer_var renv ud (Some (Ast_t.Name empty_pos)) in
	renv', Connect d, (Ast_t.Channel empty_pos)
  | UMsg um ->
      let renv', m = infer_msg renv um in
	renv', LVMsg m, (Ast_t.Msg empty_pos)

let rec infer_cont outs proves denv renv = function
  | UFail -> renv, Fail
  | UReturn ->
      let denv', renv', proves' = infer_formulas denv renv proves in
      let renv'', outs' = infer_vars renv' outs in
	renv'', VComp.CPPL.Return (proves', outs')
  | ULet (ud, ulv, usk) ->
      let renv', lv, t = infer_letval renv ulv in
      let renv'', d = infer_var renv' ud (Some t) in
      let denv' = pmap_add_decl ud () denv in
      let renv''', sk = infer_cont outs proves denv' renv'' usk in
	renv''', Let (d, lv, sk)
  | UCDerive (ut, usk, ufk) ->
      let denv', renv', t = infer_formulas denv renv ut in
      let renv'', sk = infer_cont outs proves denv' renv' usk in
      let renv3, fk = infer_cont outs proves denv renv'' ufk in
	renv3,  VComp.CPPL.Derive (t, sk, fk)
  | USend (ut, uv, um, usk, ufk) ->
      let denv', renv', t = infer_formulas denv renv ut in
      let renv'', v = infer_var renv' uv (Some (Ast_t.Channel empty_pos)) in
      let renv''', m = infer_msg renv'' um in
      let renv4, sk = infer_cont outs proves denv' renv''' usk in
      let renv5, fk = infer_cont outs proves denv renv4 ufk in
	renv5, Send (t, v, m, sk, fk)
  | UReceive (uv, um, ut, usk, ufk) ->
      let renv1, v = infer_var renv uv (Some (Ast_t.Channel empty_pos)) in
      let renv2, m = infer_msg renv1 um in
      let denv' = update_denv_msg denv um in
      let denv2, renv3, t = infer_formulas denv' renv2 ut in
      let renv4, sk = infer_cont outs proves denv2 renv3 usk in
      let renv5, fk = infer_cont outs proves denv renv4 ufk in
	renv5, Recv (v, m, t, sk, fk)
  | UCall (ubt, p, ubs, ues, uet, usk, ufk) ->
      let _, pc = p in
      let denv1, renv1, bt = infer_formulas denv renv ubt in
      let renv2, bs = infer_vars renv1 ubs in
      let renv3, es = infer_vars renv2 ues in
      let denv2 = add_decls ues denv in
      let denv3, renv4, et = infer_formulas denv2 renv3 uet in
      let renv5, sk = infer_cont outs proves denv3 renv4 usk in
      let renv6, fk = infer_cont outs proves denv renv5 ufk in
	renv6, Call (bt, pc, bs, es, et, sk, fk)
  | UCMatch (uv, um, ut, usk, ufk) ->
      let renv1, v = infer_var renv uv (Some (Ast_t.Msg empty_pos)) in
      let renv2, m = infer_msg renv1 um in
      let denv1 = update_denv_msg denv um in
      let denv2, renv3, t = infer_formulas denv1 renv2 ut in
      let renv4, sk = infer_cont outs proves denv2 renv3 usk in
      let renv5, fk = infer_cont outs proves denv renv4 ufk in
	renv5, VComp.CPPL.Match (v, m, t, sk, fk)

let infer_proc prfx (up:unproc) : proc =
  the_prefix := prfx ;
  let UProc (n, ins, outs, assumes, proves, c) = up in
  let renv = extend_decls cmb_merge ins
             (extend_decls cmb_merge outs PMap.empty) in
  let denv = add_decls ins PMap.empty in
  let denv', renv', assumes' = infer_formulas denv renv assumes in
  let renv'', c' = infer_cont outs proves denv' renv' c in
  let _, ins' = infer_vars renv'' ins in
  let _, nc = n in
    Proc (nc, ins', assumes', c')

