(* CPPL Interpreter
   Copyright (C) 2008 Jay McCarthy
*)

open Sexpr
open Crypto
open Message

type v =
  | Pub of pubkey
  | Sym of symkey
  | Nonc of nonce
  | Txt of text
  | Nm of name
  | Mess of msg

let v_equals v v' =
  match v, v' with
    Pub x, Pub x' ->
      pubkey_equals x x'
  | Sym x, Sym x' ->
      symkey_equals x x'
  | Nonc x, Nonc x' ->
      nonce_equals x x'
  | Txt x, Txt x' ->
      text_equals x x'
  | Nm x, Nm x' ->
      name_equals x x'
  | Mess x, Mess x' ->
      msg_equals x x'
  | _ -> false

include Datalog.Make(
  struct
    type t = v
    let equal = v_equals
    let hash = Hashtbl.hash
  end)

open Infer
open Printer
open Format

let cst = VComp.IdSet.IOT.str
let tsc = VComp.IdSet.IOT.rts

module VD = VComp.Datatypes
module VL = VComp.List

open VComp.CPPLlb
open VComp.CPPLss
open VComp.CPPLc
open VComp.CPPL
open VComp.Common

type 'a env = id -> 'a option
let empty_env : 'a env = fun _ -> None
let ext_env : id -> 'a -> 'a env -> 'a env =
  fun i v e i' ->
    if i = i' then Some v else e i'

let id_to_val env i =
  match env i with Some v -> v | None -> raise (Cppl_abort "id unbound")

let val_to_symkey = function
  | Sym s -> s
  | _ -> raise (Cppl_abort "not symkey")

let val_to_pubkey = function
  | Pub s -> s
  | _ -> raise (Cppl_abort "not pubkey")

let val_to_name = function
  | Nm s -> s
  | _ -> raise (Cppl_abort "not name")

let val_to_msg = function
  | Pub p -> message_of_pubkey p
  | Sym s -> message_of_symkey s
  | Nonc n -> message_of_nonce n
  | Txt t -> message_of_text t
  | Nm n -> message_of_name n
  | Mess m -> message_of_msg m

let rec msg_to_message env (m:msg) = 
  match m with
  | MNil -> Nil
  | MVar (Var (i,_)) -> val_to_msg (id_to_val env i)
  | MConst s -> message_of_text s
  | Join (p1,p2) ->
      concatenate (msg_to_message env p1) (msg_to_message env p2)
  | Hash p -> hash (msg_to_message env p)
  | SymEnc (Var (i,_),p) -> 
      symkey_encrypt (msg_to_message env p) (val_to_symkey (id_to_val env i))
  | SymSign (Var (i,_),p) -> 
      symkey_sign (msg_to_message env p) (val_to_symkey (id_to_val env i))
  | PubEnc (Var (i,_),p) -> 
      pubkey_encrypt (msg_to_message env p) (val_to_pubkey (id_to_val env i))
  | PubSign (Var (i,_),p) -> 
      pubkey_sign (msg_to_message env p) (val_to_pubkey (id_to_val env i))
  | Bind (_,p) -> msg_to_message env p

let rec match_msg env (m:msg) (c:message) =
  match m with
    | MNil ->
	cknil c ; env
    | MVar (Var (i,t)) ->
	(match t with
	   | Text ->
	       let s = text_of_message c in
		 ext_env i (Txt s) env
	   | Msg ->
	       let m = msg_of_message c in
		 ext_env i (Mess m) env
	   | Name ->
	       let n = name_of_message c in
		 ext_env i (Nm n) env
	   | Symkey ->
	       let s = symkey_of_message c in
		 ext_env i (Sym s) env
	   | Pubkey ->
	       let p = pubkey_of_message c in
		 ext_env i (Pub p) env
	   | Nonce ->
	       let n = nonce_of_message c in
		 ext_env i (Nonc n) env
	   | Channel ->
	       raise (Cppl_abort "Cannot send/recv channels"))
    | MConst mc ->
	text_same mc (text_of_message c) ;
	env
    | Join (m1,m2) ->
	let (c1, c2) = separate c in
	let env' = match_msg env m1 c1 in
	  match_msg env' m2 c2
    | Hash m1 ->
	message_same (msg_to_message env m1) c ;
	env
    | SymEnc (Var (i,_), m1) ->
	let k = val_to_symkey (id_to_val env i) in
	let c1 = symkey_decrypt c k in
	  match_msg env m1 c1
    | SymSign (Var (i,_), m1) ->
	let k = val_to_symkey (id_to_val env i) in
	  symkey_verify c k ;
	  match_msg env m1 (symkey_signed_message c)
    | PubEnc (Var (i,_), m1) ->
	let k = val_to_pubkey (id_to_val env i) in
	let c1 = pubkey_decrypt c k in
	  match_msg env m1 c1
    | PubSign (Var (i,_), m1) ->
	let k = val_to_pubkey (id_to_val env i) in
	  pubkey_verify c k ;
	  match_msg env m1 (pubkey_signed_message c)
    | Bind (Var (i,_), m1) ->
	let env' = match_msg env m1 c in
	  ext_env i (Mess (Submsg c)) env'

(* Trust management interface *)
let id_to_term env i =
  match env i with
    | None -> mkvar (tsc i)
    | Some v -> mkval v

let fvar_id = function
  | DataV (Var (i,_)) -> i
  | LogicV (Var (i,_)) -> i

let l2l l =
  VComp.List.fold_right
    (fun e a -> e :: a)
    [] l

let formula_to_qry : v env -> formula -> literal =
  fun env f ->
    let Formula (s,is) = f in
    mkliteral s 
      (l2l (VComp.List.map (fun fv -> id_to_term env (fvar_id fv)) is))

let answer_ext : v env -> literal -> literal -> v env =
  fun env qry ans ->
    let rec loop env qts ats =
      match (qts, ats) with
	| ([], []) -> env
	| (qt :: qts, at :: ats) ->
	    let env' =
	      spreadterm
		(fun qi -> 
		   spreadterm
		     (fun _ -> 
			raise (Cppl_abort "variable should be not in term"))
		     (fun av -> ext_env (cst qi) av env)
		     at)
		(fun _ -> env)
		qt
	    in
	      loop env' qts ats
	| _ -> raise (Cppl_abort "terms not the same length")
    in
      loop env (getterms qry) (getterms ans)

let prove_in : theory -> v env -> formula VComp.List.list 
  -> (theory * v env) option =
  fun thy env fs ->
    let thy' = copy thy in
    let rec loop fs env = 
      match fs with
	| [] -> Some (thy', env)
	| f :: fs ->
	    let qry = formula_to_qry env f in
	      (match prove thy' qry with
		 | [] -> None
		 | ans :: _ ->
		     loop fs (answer_ext env qry ans))
    in
      loop (l2l fs) env

(* Interfaces *)
type 'a interface = {
  accept : unit -> 'a;
  connect : name -> 'a;
  recv : 'a -> message;
  send : 'a -> message -> unit
}

type 'a session =
  | Session of theory * v env * 'a

type 'a result =
  | Err
  | Ret of v list
  | Cont of 'a session * cont

type 'a execute =
    ('a env) session -> cont -> (('a env) result)

let mk_execute : 'a interface -> 'a execute  =
  fun iface ses cont ->
    let { accept = accept; connect = connect;
	  recv = recv; send = send; } = iface in
    let Session (thy, env, cenv) = ses in
      match cont with
	 | Fail -> Err
	 | Return (f,vrs) -> 
	     (match prove_in thy env f with
		| None -> Err
		| Some (_, env') ->
		    let is = VComp.List.map (function Var (i,_) -> i) vrs in
		    let vs = List.map (id_to_val env') (l2l is) in
		      Ret vs)
	 | Let ((Var (i,_)), lv, k) ->
	     (match lv with
		| New_nonce ->
		    let n = new_nonce () in
		    let env' = ext_env i (Nonc n) env in
		      Cont (Session (thy, env', cenv), k)
		| New_symkey ->
		    let n = new_symkey () in
		    let env' = ext_env i (Sym n) env in
		      Cont (Session (thy, env', cenv), k)
		| New_pubkey ->
		    let n = new_pubkey () in
		    let env' = ext_env i (Pub n) env in
		      Cont (Session (thy, env', cenv), k)
		| Accept ->
		    let c = accept () in
		    let cenv' = ext_env i c cenv in
		      Cont (Session (thy, env, cenv'), k)
		| Connect (Var (ci,_)) ->
		    let cn = val_to_name (id_to_val env ci) in
		    let c = connect cn in
		    let cenv' = ext_env i c cenv in
		      Cont (Session (thy, env, cenv'), k)
		| LVMsg p ->
		    let m' = msg_to_message env p in
		    let env' = ext_env i (Mess (Submsg m')) env in
		      Cont (Session (thy, env', cenv), k))
	 | Assert (f, sk, fk) -> 
	     (match prove_in thy env f with
		| None -> Cont (Session (thy,env,cenv), fk)
		| Some (thy',env') -> Cont (Session (thy',env', cenv), sk))
	 | Send (f, Var (i,_), p, sk, fk) -> 
	     (match prove_in thy env f with
		| None -> Cont (Session (thy,env,cenv), fk)
		| Some (thy',env') ->
		    send (id_to_val cenv i) (msg_to_message env p) ;
		    Cont (Session (thy',env',cenv), sk))
	 | Recv (Var (i,_), p, f, sk, fk) -> (* XXX How to get fk? *)
	     let m = recv (id_to_val cenv i) in
	     let env' = match_msg env p m in
	       (match prove_in thy env' f with
		  | None -> Err
		  | Some (thy',env'') ->
		      Cont (Session (thy',env'', cenv), sk))
	 | Call _ ->
	     raise (Cppl_abort "Can't interpret call")
	 | Match (Var (i,_), p, f, sk, fk) -> 
	     let m = val_to_msg (id_to_val env i) in
	       try 
		 let env' = match_msg env p m in
		   (match prove_in thy env' f with
		      | None -> Cont (Session (thy, env, cenv), fk)
		      | Some (thy',env'') -> 
			  Cont (Session (thy',env'', cenv), sk))
	       with
		   _ ->
		     Cont (Session (thy, env, cenv), fk)

let simple_run : theory -> v env -> cont -> 'a execute -> (v list) option =
  fun thy env k exec ->
    let rec loop ses k =
      match exec ses k with
	| Err -> None
	| Ret vs -> Some vs
	| Cont (ses',k') -> loop ses' k'
    in
      loop (Session (thy,env,empty_env)) k

open Unix
type channel = {
  fd : file_descr;
  input : in_channel;
  output : out_channel
}

let rec try_n_times n f =
  if n > 0 then
    (try 
       f ()
     with	
	 _ ->
	   Unix.sleep 1 ;
	   try_n_times (n-1) f)
  else
    f ()

let unixInterface port =
  let server = ref None in
  let myself =
    try
      let name = gethostname () in
      let { h_addr_list = addr_list; } = gethostbyname name in
	Array.get addr_list 0
    with
	Not_found -> inet_addr_loopback
  in
  let self = (ADDR_INET (myself, port)) in
  let timeout = 30.0 in
  let mkchannel fd =
    setsockopt_float fd SO_RCVTIMEO timeout;
    let input = in_channel_of_descr fd in
    let output = out_channel_of_descr fd in
      { fd = fd; input = input; output = output }
  in
  let link () =
    match !server with
	Some fd -> fd
      | None ->
	  let domain = domain_of_sockaddr self in
	  let fd = socket domain SOCK_STREAM 0 in
	    setsockopt fd SO_REUSEADDR true ;
	    try_n_times 10 (fun () -> bind fd self);
	    listen fd 1000;
	    setsockopt_optint fd SO_LINGER (Some 1);
	    server := Some fd;
	    fd
  in
  let accept () =
    let fd, _ = Unix.accept (link ()) in   
      mkchannel fd
  in
  let connect addr =
    let domain = domain_of_sockaddr addr in
    let fd = socket domain SOCK_STREAM 0 in
      Unix.connect fd addr;
      mkchannel fd
  in
  let send_sexpr { output = output } s =
    output_sexpr output s;
    flush output
  in
  let receive_sexpr { input = input } =
    try
      input_sexpr input
    with Unix_error (e, s, _) ->
      let msg = Printf.sprintf "%s: %s" s (error_message e) in
	raise (Cppl_abort msg)
  in
  let recv chan =
    message_of_sexpr (receive_sexpr chan)
  in
  let send chan message =
    send_sexpr chan (sexpr_of_message message)
  in
    { accept = accept;
      connect = connect;
      recv = recv;
      send = send }

let unix_simple_run thy env k port = 
  simple_run thy env k (mk_execute (unixInterface port))

let multi_run : (theory * v env * cont * 'a execute) list -> (v list) list =
  fun ps ->
    let rec loop q res =
      match q with
	| [] -> res
	| (exec, ses, k) :: q ->
	    match exec ses k with
	      | Err -> loop q res
	      | Ret vs -> loop q (vs :: res)
	      | Cont (ses', k') -> loop (q @ [(exec, ses', k')]) res
    in
      loop 
	(List.map (fun (thy, env, k, exec) ->
		     (exec, Session (thy, env, empty_env), k))
	   ps)
	[]

open Thread
open Event
exception NoComm
exception MsgUsed

let forever : 'a interface -> theory -> v env -> cont -> unit =
  (* Run continuations like k forever *)
  fun iface init_thy init_env k ->
    let { accept = iaccept; connect = iconnect;
	  recv = irecv; send = isend; } = iface in
      (* There is one thread per accepted channel reading off messages *)
    let recv_ch = new_channel () in
    let recv_thr c =
      create (fun () ->
		while true do
		  let m = irecv c in
		    send recv_ch (c,m)
		done)
	()
    in
      (* There is one thread constantly accepting *)
    let accept_thr =
      create (fun () ->
		while true do
		  recv_thr (iaccept ())
		done)
	()
    in
      (* This kind of execution allows send/recv without use of shared
	 port *)
    let nocomm_exec =
      mk_execute
	{ accept = (fun () -> None) ;
	  connect = (fun n -> Some (iconnect n));
	  recv = (fun c ->
		    match c with
		      | Some ch -> irecv ch
		      | None -> raise NoComm);
	  send = (fun c m ->
		    match c with
		      | Some ch -> isend ch m
		      | None -> failwith "Accept can't send first") }
    in
      (* This kind allows the reception of one message *) 
    let recv_once_exec m =
      let the_m = ref true in
	mk_execute
	  { accept = (fun () -> failwith "Can't accept twice") ;
	    connect = (fun n -> Some (iconnect n));
	    recv = (fun c ->
		      match c with
			| Some ch -> irecv ch
			| None ->
			    if !the_m then (
			      the_m := false ;
			      m
			    ) else (
			      raise MsgUsed
			    )) ;
	    send = (fun c m ->
		      match c with
			| Some ch -> isend ch m
			| None -> raise NoComm) }
    in
      (* This allos many sends, but no recvs *)
    let send_many_exec mc =
	mk_execute
	  { accept = (fun () -> failwith "Can't accept twice") ;
	    connect = (fun n -> Some (iconnect n));
	    recv = (fun c ->
		      match c with
			| Some ch -> irecv ch
			| None -> raise MsgUsed) ;
	    send = (fun c m ->
		      match c with
			| Some ch -> isend ch m
			| None -> isend mc m) }
    in
      (* Main loop *)
    let wait_ch = new_channel () in
      (* XXX *)
    let create_new_session_or_drop m = () in
      (* XXX *)
    let try_to_recv ctxt = false in
    let rec loop waiting =
      match 
	select
	  [ (wrap wait_ch (fun x -> Left x)) ;
	    (wrap recv_ch (fun x -> Right x)) ]
      with
	  (* A running thread has been moved to the waiting step *)
	| Left x -> loop (x :: waiting)
	| Right m ->
	    let rec iloop ws =
	      match ws with 
		| [] -> 
		    create_new_session_or_drop m ;
		    []
		| w :: ws ->
		    if try_to_recv w then
		      ws
		    else
		      w :: (iloop ws)
	    in
	      loop (iloop waiting)
    in
      loop []
      
    (* XXX Turn proc into thy, env and cont *)

let load f =
  let ast1_ut = Reader.read_proc f in
  let ast1_t = infer_proc "" ast1_ut in
  let ast1_tf = foldm_proc ast1_t in
    ast1_tf
