(* Runtime messages
   Copyright (C) 2008 Jay McCarthy
*)

open Crypto
open Unix
open Sexpr

exception Cppl_abort of string

type text = string
type name = sockaddr

type message =
  | Nil
  | Str of string
  | Name of name
  | Nonce of nonce
  | Pub of pubkey
  | Sym of symkey
  | Concatenate of message * message
  | Hash of string
  | Pub_signed of string * message
  | Sym_signed of string * message
  | Pub_encrypted of string * string
  | Sym_encrypted of string

let message_type = function
  | Nil -> "nil"
  | Str _ -> "str"
  | Name _ -> "name"
  | Nonce _ -> "nonce"
  | Pub _ -> "pub"
  | Sym _ -> "sym"
  | Concatenate _ -> "concatenate"
  | Hash _ -> "hash"
  | Pub_signed _ -> "pub_signed"
  | Sym_signed _ -> "sym_signed"
  | Pub_encrypted _ -> "pub_encrypted"
  | Sym_encrypted _ -> "sym_encrypted"

let expect s m =
  let msg =
    Printf.sprintf
      "Expecting %s, but got %s"
      s (message_type m) in
  raise (Cppl_abort msg)

type msg = Submsg of message

let rec sexpr_of_message m =
  match m with
    | Nil -> List [ String "Nil" ]
    | Str s ->
	List [ String "Str"; String s ]
    | Name (ADDR_INET (addr, port)) ->
	List [ String "ADDR_INET"; String (string_of_inet_addr addr);
	       String (string_of_int port) ]
    | Name (ADDR_UNIX s) ->
	List [ String "ADDR_UNIX"; String s ]
    | Nonce n ->
	List [ String "Nonce"; String (string_of_nonce n) ]
    | Pub pk ->
	List [ String "Pub"; String (string_of_pubkey pk) ]
    | Sym s ->
	List [ String "Sym"; String (string_of_symkey s) ]
    | Concatenate (m1, m2) ->
	List [ String "Concatenate"; sexpr_of_message m1; sexpr_of_message m2 ]
    | Hash h ->
	List [ String "Hash"; String h ]
    | Pub_signed (signature, message) ->
	List [ String "Pub_signed"; String signature; sexpr_of_message message ]
    | Sym_signed (signature, message) ->
	List [ String "Sym_signed"; String signature; sexpr_of_message message ]
    | Pub_encrypted (s, d) ->
	List [ String "Pub_encrypted"; String s ; String d ]
    | Sym_encrypted s ->
	List [ String "Sym_encrypted"; String s ]

let rec message_of_sexpr s =
  match s with
    | List [ String "Nil" ] -> Nil
    | List [ String "Str"; String s ] ->
	Str s
    | List [ String "ADDR_INET"; String addr; String port ] ->
	Name (ADDR_INET (inet_addr_of_string addr, int_of_string port))
    | List [ String "ADDR_UNIX"; String s ] ->
	Name (ADDR_UNIX s)
    | List [ String "Nonce"; String n ] ->
	Nonce (nonce_of_string n)
    | List [ String "Pub"; String pk ] ->
	Pub (pubkey_of_string pk)
    | List [ String "Sym"; String s ] ->
	Sym (symkey_of_string s)
    | List [ String "Concatenate"; s1; s2 ] ->
	Concatenate (message_of_sexpr s1, message_of_sexpr s2)
    | List [ String "Hash"; String h ] ->
	Hash h
    | List [ String "Pub_signed"; String signature; s1 ] ->
	Pub_signed (signature, message_of_sexpr s1)
    | List [ String "Sym_signed"; String signature; s1 ] ->
	Sym_signed (signature, message_of_sexpr s1)
    | List [ String "Pub_encrypted"; String s ; String d ] ->
	Pub_encrypted (s, d)
    | List [ String "Sym_encrypted"; String s ] ->
	Sym_encrypted s
    | _ -> raise (Cppl_abort "message parsing failed")
	
let to_string message =
  string_of_sexpr (sexpr_of_message message)

let from_string string =
  message_of_sexpr (sexpr_of_string string)

let message_equals = (=)

let message_of_text s =
  Str s

let message_of_msg (Submsg m) =
  m

let message_of_name p =
  Name p

let message_of_pubkey k =
  Pub k

let message_of_symkey k =
  Sym k

let message_of_nonce n =
  Nonce n

let text_of_message m =
  match m with
      Str s -> s
    | _ -> expect "text" m

let msg_of_message m =
  Submsg m

let name_of_message m =
  match m with
      Name p -> p
    | _ -> expect "name" m

let symkey_of_message m =
  match m with
      Sym k -> k
    | _ -> expect "symkey" m

let pubkey_of_message m =
  match m with
      Pub k -> k
    | _ -> expect "pubkey" m

let nonce_of_message m =
  match m with
      Nonce n -> n
    | _ -> expect "nonce" m

let cknil m =
  match m with
    | Nil -> ()
    | _ -> expect "nil" m

let concatenate m1 m2 =
  Concatenate (m1, m2)

let separate m =
  match m with
      Concatenate (m1, m2) ->
	(m1, m2)
    | _ ->
	expect "concatenate" m

let hash m =
  Hash (Crypto.hash (to_string m))

let key_pairs = Hashtbl.create 32   (* Maps public keys -> full keys *)

let get_key_pair k =		     (* Get the full key previously *)
  try					(* stored here *)
    Hashtbl.find key_pairs k
  with Not_found ->
    raise (Cppl_abort "private portion of key not known")

let store_key_pair k =		       (* Store the full key in the *)
  let stripped = pubkey_of k in	     (* hash table, and return just *)
  Hashtbl.replace key_pairs stripped k;	(* the public part *)
  stripped

let new_pubkey () =
  store_key_pair (new_keypair())

let pubkey_sign m k =
  let priv_key = get_key_pair k in
  Pub_signed (Crypto.pub_sign (to_string m) priv_key, m)

let pubkey_verify m k =
  match m with
    Pub_signed (signature, message) ->
      let valid =
	try
	  Crypto.pub_verify signature (to_string message) k
	with _ ->
	  raise (Cppl_abort "Error during signature validation") in
      if not valid then
	raise (Cppl_abort "invalid signature")
  | _ ->
      expect "pub_signed" m

let pubkey_signed_message m =
  match m with
    Pub_signed (signature, message) ->
      message
  | _ ->
      expect "pub_signed" m

let pubkey_encrypt m k =
  let secret, data = Crypto.pub_encrypt (to_string m) k in
  Pub_encrypted (secret, data)

let pubkey_decrypt m k =
  match m with
  | Pub_encrypted (secret, data) ->
      let priv_key = get_key_pair k in
      let contents =
	try
	  Crypto.pub_decrypt (secret, data) priv_key
	with _ ->
	    raise (Cppl_abort "Error during asymmetric decryption") in
      from_string contents
  | _ ->
      expect "pub_encrypt" m

let symkey_sign m k =
  let h = Crypto.hash (to_string m) in
  Sym_signed (Crypto.sym_encrypt h k, m)

let symkey_verify m k =
  match m with
    Sym_signed (signature, message) ->
      let h = Crypto.hash (to_string m) in
      if h <> Crypto.sym_decrypt signature k then
	raise (Cppl_abort "invalid signature")
  | _ ->
      expect "sym_signed" m

let symkey_signed_message m =
  match m with
    Sym_signed (signature, message) ->
      message
  | _ ->
      expect "sym_signed" m

let symkey_encrypt m k =
  Sym_encrypted (Crypto.sym_encrypt (to_string m) k)

let symkey_decrypt m k =
  match m with
    Sym_encrypted data ->
      let contents =
	try
	  Crypto.sym_decrypt data k
	with _ ->
	  raise (Cppl_abort "Error during symmetric decryption") in
      from_string contents
  | _ ->
      expect "sym_encrypt" m

let save_key_pair_entry output name =
  let key = new_keypair() in
  let str = string_of_keypair key in
  let sexpr = List [ String "key pair"; String name; String str] in
  output_sexpr output sexpr;
  pubkey_of key

let save_pubkey_entry output name k =
  let str = string_of_pubkey k in
  let sexpr = List [ String "pubkey"; String name; String str] in
  output_sexpr output sexpr

let load_key_store_entry input =
  match input_sexpr input with
    List [ String "key pair"; String name; String str] ->
      let keypair = keypair_of_string str in
      name, store_key_pair keypair
  | List [ String "pubkey"; String name; String str] ->
      name, pubkey_of_string str
  | _ -> failwith "load_key_store"

let msg_equals = (=)
let text_equals = (=)
let name_equals = (=)

(* Raise Cppl_abort on unequal values. *)

let pubkey_same x y =
  if not (pubkey_equals x y) then
    raise (Cppl_abort "differing pub keys")

let symkey_same x y =
  if not (symkey_equals x y) then
    raise (Cppl_abort "differing sym keys")

let nonce_same x y =
  if not (nonce_equals x y) then
    raise (Cppl_abort "differing nonces")

let text_same x y =
  if not (text_equals x y) then
    raise (Cppl_abort "differing texts")

let name_same x y =
  if not (name_equals x y) then
    raise (Cppl_abort "differing names")

let msg_same x y =
  if not (msg_equals x y) then
    raise (Cppl_abort "differing msgs")

let message_same x y =
  if not (message_equals x y) then
    raise (Cppl_abort "differing messages")
