/* A parser for WPPL and CPPL programs.
   Copyright (C) 2007 Jay McCarthy
*/

%{

open Ast_ut
open Ast_t

%}

%token EOF
%token <Pos.pos * string> CONST ID
%token <Pos.pos> HASH MSG NAME NONCE SPEC CHANNEL FAIL IN SEND
%token <Pos.pos> PUBKEY SYMKEY TEXT END ASSUMES PROVES THEN RECEIVE
%token <Pos.pos> ARROW CANGLE CBRAC CBRACBAR CCURL CCURLBAR CPAREN
%token <Pos.pos> COLON COMMA OBRAC OBRACBAR OCURL OCURLBAR ELSE CONNECT
%token <Pos.pos> OPAREN OPBRAC CPBRAC UNDERSCORE PERIOD ASSERT ACCEPT
%token <Pos.pos> LET AT EQUAL NEW BIND TO MATCH WITH RETURN CALL
%right COMMA

%start spec

%type <Ast_ut.unspec> spec

%start proc

%type <Ast_ut.unproc> proc

%%

spec:
   OPAREN SPEC role_specs_list action CPAREN EOF
   { $3, $4 }
;

proc:
  ID decls_list decls_list assumes proves cont END
  { UProc ($1, $2, $3, $4, $5, $6) }
;

assumes:
  ASSUMES terms { $2 }
;

proves:
  PROVES terms { $2 }
;

role: ID { $1 }

role_specs_list:
  OPAREN CPAREN
  { [] }
  | OPAREN role_specs CPAREN
  { List.rev $2 }

role_specs:
  role_spec              
  { [$1] }
  | role_specs role_spec 
      { $2 :: $1 }
;

role_spec:
  OBRAC role decls_list decls_list CBRAC
  { $2, $3, $4 }

fdecl_list:
  OPAREN CPAREN
  { [] }
  | OPAREN fdecls CPAREN
  { List.rev $2 }

fdecls:
  fdecl                           
  { [$1] }
| fdecls fdecl
  { $2 :: $1 }
;

fdecl:
  decl              { UDataV $1 }
| OBRAC decl CBRAC  { ULogicV $2 }
;

decls_list:
  OPAREN CPAREN
  { [] }
  | OPAREN decls CPAREN
  { List.rev $2 }

decls:
  decl                           
  { [$1] }
| decls decl
  { $2 :: $1 }
;

decl:
  ID                { $1, None }
| ID COLON var_type { $1, Some $3 }
;

var_type:
  NAME                           { Name $1 }
| TEXT                           { Text $1 }
| MSG                            { Msg $1 }
| NONCE                          { Nonce $1 }
| PUBKEY                         { Pubkey $1 }
| SYMKEY                         { Symkey $1 }
| CHANNEL                        { Channel $1 }
;  

new_type:
  NONCE  { New_nonce $1 }
| SYMKEY { New_symkey $1 }
| PUBKEY { New_pubkey $1 }
;

action:
  PERIOD
  { UZero }
| OBRAC role ARROW role COLON msg CBRAC action
  { UComm ($2, $4, [], $6, [], $8) }
| OBRAC role ARROW role COLON terms ARROW msg ARROW terms CBRAC action
  { UComm ($2, $4, $6, $8, $10, $12) }
| LET decl AT role EQUAL NEW new_type action
  { UNew ($4, $2, $7, $8) }
| BIND decl AT role TO msg action
  { UBind ($4, $2, $6, $7) }
| MATCH decl AT role WITH msg action
  { UMatch ($4, $2, $6, $7) }
| ASSERT terms AT role action
  { UDerive ($4, $2, $5) }

cont:
  FAIL
  { UFail }
| RETURN 
  { UReturn }
| LET decl EQUAL letval IN cont
  { ULet ($2, $4, $6) }
| ASSERT terms THEN cont ELSE cont
  { UCDerive ($2, $4, $6) }
| SEND terms ARROW decl msg THEN cont ELSE cont
  { USend ($2, $4, $5, $7, $9) }
| RECEIVE decl msg ARROW terms THEN cont ELSE cont
  { UReceive ($2, $3, $5, $7, $9) }
| CALL terms ARROW ID decls_list decls_list ARROW terms THEN cont ELSE cont
  { UCall ($2, $4, $5, $6, $8, $10, $12) }
| MATCH decl msg ARROW terms THEN cont ELSE cont
  { UCMatch ($2, $3, $5, $7, $9) }
;

letval:
  NEW NONCE
  { UNew_nonce }
| NEW SYMKEY
  { UNew_symkey }
| NEW PUBKEY
  { UNew_pubkey }
| ACCEPT
  { UAccept }
| CONNECT decl
  { UConnect $2 }
| msg
  { UMsg $1 }
;

terms:
  UNDERSCORE
  { [] }
| term
  { [ $1 ] }
| term COMMA terms
  { $1 :: $3 }

term:
| ID fdecl_list
  { UTerm ($1, $2) }

msg:
  decl                          { UDecl $1 }
| OPAREN msg CPAREN             { $2 }
| msg COMMA msg                 { UCat ($1, $3) }
| CONST                         { let p, s = $1 in UConst (p, s) }
| OCURL msg CCURL decl          { UPubEncrypt ($2, $4) }
| OBRAC msg CBRAC decl          { UPubSign ($2, $4) }
| OCURLBAR msg CCURLBAR decl    { USymEncrypt ($2, $4) }
| OBRACBAR msg CBRACBAR decl    { USymSign ($2, $4) }
| HASH OPAREN msg CPAREN        { UHash $3 }
| OPBRAC decl EQUAL msg CPBRAC  { UMBind ($2, $4) }
;
