(* Printer
   Copyright (C) 2008 Jay McCarthy
*)

module VD = VComp.Datatypes

open VComp.CPPLc
open VComp.CPPL (* Also binds 'Return' *)
open VComp.Common (* Also binds 'Bind' *)
open VComp.WPPL 
open VComp.List

let cst = VComp.IdSet.IOT.str
let tsc = VComp.IdSet.IOT.rts

(* WPPL PRINTER *)
open Format

let pr_id ppf s = 
  pp_print_string ppf (tsc s)

let type2str = function
  | Text -> "text"
  | Msg -> "msg"
  | Nonce -> "nonce"
  | Channel -> "channel"
  | Name -> "name"
  | Symkey -> "symkey"
  | Pubkey -> "pubkey"

let pr_type ppf t =
  pp_print_string ppf (type2str t)

let pr_new_type ppf nt =
  pr_type ppf (new_type_to_type nt)

let pr_var ppf (Var (id, t)) =
  fprintf ppf "@[%a:%a@]"
    pr_id id
    pr_type t

let pr_list pr_elem d ppf es =
  let rec h ppf = function
      Coq_nil -> 
	pp_print_string ppf ""
    | Coq_cons (e, Coq_nil) ->
	fprintf ppf "%a"
	  pr_elem e
    | Coq_cons (e, es) ->
	fprintf ppf "%a%a@ %a" 
	  pr_elem e
	  pp_print_string d
	  h es
  in
    fprintf ppf "@[<1>(%a)@]" 
      h es

let pr_vars ppf ds = pr_list pr_var "" ppf ds

let pr_varsl ppf ds = pr_list pr_var "," ppf ds

let rec pr_msg ppf = function
  | MNil ->
      pp_print_string ppf "."
  | MVar v ->
      pr_var ppf v
  | MConst s ->
      pp_print_string ppf (Printf.sprintf "%S" s)
  | Join (m1, m2) ->
      fprintf ppf "@[%a,@ %a@]"
	pr_msg m1
	pr_msg m2
  | Hash m ->
      fprintf ppf "@[hash(%a)@]"
	pr_msg m
  | SymEnc (v, m) ->
      fprintf ppf "@[{|@ %a@ |}@ %a@]"
	pr_msg m
	pr_var v
  | SymSign (v, m) ->
      fprintf ppf "@[[|@ %a@ |]@ %a@]"
	pr_msg m
	pr_var v
  | PubEnc (v, m) ->
      fprintf ppf "@[{@ %a@ }@ %a@]"
	pr_msg m
	pr_var v
  | PubSign (v, m) ->
      fprintf ppf "@[[@ %a@ ]@ %a@]"
	pr_msg m
	pr_var v
  | VComp.Common.Bind (v, m) ->
      fprintf ppf "@[ <%a@ =@ %a>@]"
	pr_var v
	pr_msg m

let pr_fvar ppf = function
  | DataV v -> pr_var ppf v
  | LogicV v -> 
      fprintf ppf "@[[%a]@]"
	pr_var v

let pr_formula ppf = function
  | Formula (id, vs) ->
      fprintf ppf "@[%a%a@]"
	pp_print_string id
	(pr_list pr_fvar "") vs

let pr_formulas ppf es =
  let rec h ppf = function
      Coq_nil -> 
	pp_print_string ppf "_"
    | Coq_cons (e, Coq_nil) ->
	fprintf ppf "%a"
	  pr_formula e
    | Coq_cons (e, es) ->
	fprintf ppf "%a%a@ %a" 
	  pr_formula e
	  pp_print_string ","
	  h es
  in
    fprintf ppf "@[<1>%a@]" 
      h es

let pr_role = pr_id

let rec pr_action ppf = function
  | Return -> 
      pp_print_string ppf "."
  | Comm (fr, tr, ffs, msg, tfs, a) ->
      fprintf ppf "@[<1>[%a@ ->@ %a@ :@ %a@ ->@ %a@ ->@ %a]@]@\n%a"
	pr_role fr
	pr_role tr
	pr_formulas ffs
	pr_msg msg
	pr_formulas tfs
	pr_action a
  | New (r, v, nt, a) ->
      fprintf ppf "@[<1>let@ %a@ @@@ %a@ =@ new@ %a@]@\n%a"
	pr_var v
	pr_role r
	pr_new_type nt
	pr_action a
  | Bind (r, v, m, a) ->
      fprintf ppf "@[<1>bind@ %a@ @@@ %a@ to@ %a@]@\n%a"
	pr_var v
	pr_role r
	pr_msg m
	pr_action a
  | Match (r, v, m, a) ->
      fprintf ppf "@[<1>match@ %a@ @@@ %a@ with@ %a@]@\n%a"
	pr_var v
	pr_role r
	pr_msg m
	pr_action a
  | Derive (r, f, a) ->
      fprintf ppf "@[<1>assert@ %a@ @@@ %a@]@\n%a"
	pr_formulas f
	pr_role r
	pr_action a

let pr_rspec ppf (RSpec (r, sds, eds)) =
  fprintf ppf "@[<1>[%a@ %a@ %a]@]"
    pr_role r
    pr_vars sds
    pr_vars eds

let pr_rspecs ppf rspecs = pr_list pr_rspec "" ppf rspecs

let pr_spec ppf (Spec (rspecs, a)) =
  fprintf ppf "@[<1>(spec@ %a@ %a)@]@\n" 
    pr_rspecs rspecs
    pr_action a

let print_spec s =
  pr_spec std_formatter s

(* CPPL PRINTER *)
let pr_letval ppf = function
  | VComp.CPPL.New_nonce ->
      fprintf ppf "@[<1>new nonce@]"
  | VComp.CPPL.New_symkey ->
      fprintf ppf "@[<1>new symkey@]"
  | VComp.CPPL.New_pubkey ->
      fprintf ppf "@[<1>new pubkey@]"
  | Accept ->
      fprintf ppf "@[<1>accept@]"
  | Connect v ->
      fprintf ppf "@[<1>connect@ %a@]"
        pr_var v
  | LVMsg m ->
      pr_msg ppf m

let rec pr_cont ppf = function
  | Fail ->
      fprintf ppf "@[<1>fail@]"
  | VComp.CPPL.Return _ ->
      fprintf ppf "@[<1>return@]"
  | Let (v, lv, c) ->
      fprintf ppf "@[<1>let@ %a@ =@ %a@ in@]@\n%a"
	pr_var v
	pr_letval lv
	pr_cont c
  | VComp.CPPL.Derive (f, sk, fk) ->
      fprintf ppf "@[<1>assert@ %a@ then@ %a@ else@ %a@]"
	pr_formulas f
	pr_cont sk
	pr_cont fk
  | Send (f, v, m, sk, fk) ->
      fprintf ppf "@[<1>send@[@ %a -> %a@ (%a)@]@ then@ %a@ else@ %a@]"
	pr_formulas f
	pr_var v
	pr_msg m
	pr_cont sk
	pr_cont fk
  | Recv (v, m, f, sk, fk) ->
      fprintf ppf "@[<1>recv@[@ %a@ (%a) -> %a@]@ then@ %a@ else@ %a@]"
	pr_var v
	pr_msg m
	pr_formulas f
	pr_cont sk
	pr_cont fk
  | Call (bf, p, bs, es, ef, sk, fk) ->
      fprintf ppf "@[<1>call@ %a@ ->@ %a@ %a@ %a@ ->@ %a@ then@ %a@ else@ %a@]"
	pr_formulas bf
	pp_print_string p
	pr_vars bs
	pr_vars es
	pr_formulas ef
	pr_cont sk
	pr_cont fk
  | VComp.CPPL.Match (v, m, f, sk, fk) ->
      fprintf ppf "@[<1>match@ %a@ %a@ ->@ %a@ then@ %a@ else@ %a@]"
	pr_var v
	pr_msg m
	pr_formulas f
	pr_cont sk
	pr_cont fk        

let pr_proc ppf (Proc (p,xs,phi,c)) =
  let rs = cont_returns c in
    match rs with
      | Coq_cons (VComp.CPPL.Return (psi,ys), _) ->
	  fprintf ppf "@[<1>%a%a@ %a@ assumes@ %a@ proves@ %a@ %a@ end@]@\n@\n"
	    pp_print_string p
	    pr_vars xs
	    pr_vars ys 
	    pr_formulas phi
	    pr_formulas psi
	    pr_cont c
      | _ ->
	  failwith "Can only print CPPL programs with a return statement"

let print_proc p =
  pr_proc std_formatter p

(* STRAND PRINTER *)
let pr_strand_desc ppf = function
  | VD.Coq_pair (VD.Coq_pair (Outgoing, m), f) ->
      fprintf ppf "@[<1>[+(%a),@ %a]@]"
	pr_msg m
	pr_formulas f
  | VD.Coq_pair (VD.Coq_pair (Incoming, m), f) ->
      fprintf ppf "@[<1>[-(%a),@ %a]@]"
	pr_msg m
	pr_formulas f

let rec pr_strand ppf = function
  | Coq_nil ->
      pp_print_string ppf ".\n"
  | Coq_cons (sd, s) ->
      fprintf ppf "%a@ -->@ %a"
	pr_strand_desc sd
	pr_strand s

let print_strand s =
  pr_strand std_formatter s

let pr_ids ppf is = pr_list pr_id "" ppf is

let rec pr_strands ppf = function
  | Coq_nil ->
      pp_print_string ppf ""
  | Coq_cons (VD.Coq_pair (s,v),ss) ->
      fprintf ppf "@[%a:@\n%aEXT:@\n%a@]@\n%a"
	pr_ids v
	pr_strand s
	pr_strand (strand_ext_only s)
	pr_strands ss

let print_strands ss =
  pr_strands std_formatter ss

