(* A lexical analyzer for WPPL and CPPL input.
   Copyright (C) 2007 Jay McCarthy
*)

{

open Lexing
open Parser
open Pos

(* Update the current location with file name and line number. *)
let update_loc lexbuf file line absolute =
  let pos = lexbuf.lex_curr_p in
  let new_file =
    match file with
    | None -> pos.pos_fname
    | Some s -> s in
  let new_line =
    if absolute then line else pos.pos_lnum + line in
  lexbuf.lex_curr_p
  <- { pos with
       pos_fname = new_file;
       pos_lnum = new_line;
       pos_bol = pos.pos_cnum;
     }

let init lexbuf file_name =
  update_loc lexbuf (Some file_name) 1 true

let newline_found lexbuf =
  update_loc lexbuf None 1 false

(* Convert the external version of string into the internal representation *)
let strip str =
  Scanf.sscanf str "%S" (fun s -> s)

(* Keywords that scan as an identifier use a keyword table *)
let mktbl () =
  let keywords = [
      "spec", (fun pos -> SPEC pos) ;
      "assert", (fun pos -> ASSERT pos) ;
      "let", (fun pos -> LET pos) ;
      "bind", (fun pos -> BIND pos) ;
      "to", (fun pos -> TO pos) ;
      "match", (fun pos -> MATCH pos) ;
      "with", (fun pos -> WITH pos) ;
      "new", (fun pos -> NEW pos) ;
      "fail", (fun pos -> FAIL pos);
      "in", (fun pos -> IN pos);
      "send", (fun pos -> SEND pos);
      "assumes", (fun pos -> ASSUMES pos);
      "proves", (fun pos -> PROVES pos);
      "end", (fun pos -> END pos);
      "then", (fun pos -> THEN pos);
      "recv", (fun pos -> RECEIVE pos);
      "else", (fun pos -> ELSE pos);
      "connect", (fun pos -> CONNECT pos);
      "accept", (fun pos -> ACCEPT pos);
      "return", (fun pos -> RETURN pos);
      "call", (fun pos -> CALL pos);
      "hash", (fun pos -> HASH pos);
      "channel", (fun pos -> CHANNEL pos);
      "msg", (fun pos -> MSG pos);
      "name", (fun pos -> NAME pos);
      "nonce", (fun pos -> NONCE pos);
      "pubkey", (fun pos -> PUBKEY pos);
      "symkey", (fun pos -> SYMKEY pos);
      "text", (fun pos -> TEXT pos) ] in
  let table = Hashtbl.create(List.length keywords) in
  List.iter (fun (kwd, tok) -> Hashtbl.add table kwd tok) keywords;
  table

let table = mktbl()
let find_keyword_or_id lexbuf =
  let item = lexeme lexbuf in
  try
    Hashtbl.find table item (mkpos lexbuf)
  with Not_found ->
    ID(mkpos lexbuf, item)
}

let newline = ('\n' | '\r' | "\r\n")
let part = ['A'-'Z' 'a'-'z' '_' '0'-'9']

let identifier = ['a'-'z'] part*
let str_chars = [^ '"' '\\'] | "\\\\" | "\\\""  | "\\'"
| "\\n" | "\\r" | "\\t" | "\\b"
| "\\" [ '0'-'9' ]  [ '0'-'9' ]  [ '0'-'9' ]
let str = '"' str_chars* '"'

rule token = parse
  [' ' '\t']                { token lexbuf }
| newline                   { newline_found lexbuf; token lexbuf }
  (* and ML style comments *)
| "(*"			    { comment lexbuf; token lexbuf }
| "@"                       { AT (mkpos lexbuf) }
| "="                       { EQUAL (mkpos lexbuf) }
| "->"                      { ARROW (mkpos lexbuf) }
| "."                       { PERIOD (mkpos lexbuf) }
| ","                       { COMMA (mkpos lexbuf) }
| ')'                       { CPAREN (mkpos lexbuf) }
| ':'                       { COLON (mkpos lexbuf) }
| "["                       { OBRAC (mkpos lexbuf) }
| "{"                       { OCURL (mkpos lexbuf) }
| "]"                       { CBRAC (mkpos lexbuf) }
| "}"                       { CCURL (mkpos lexbuf) }
| "<"                       { OPBRAC (mkpos lexbuf) }
| ">"                       { CPBRAC (mkpos lexbuf) }
| "[|"                      { OBRACBAR (mkpos lexbuf) }
| "{|"                      { OCURLBAR (mkpos lexbuf) }
| "|]"                      { CBRACBAR (mkpos lexbuf) }
| "|}"                      { CCURLBAR (mkpos lexbuf) }
| '('                       { OPAREN (mkpos lexbuf) }
| "_"                       { UNDERSCORE (mkpos lexbuf) }
| identifier                { find_keyword_or_id lexbuf }
| str                       { CONST((mkpos lexbuf), strip(lexeme lexbuf)) }
| eof                       { EOF }
| _                         { raise Parsing.Parse_error }
and comment = parse
  "*)"			    { () }
| newline                   { newline_found lexbuf; comment lexbuf }
| "(*"			    { comment lexbuf; comment lexbuf }
| _	                    { comment lexbuf }
