(* Verified WPPL compiler 
   Copyright (C) 2007 Jay McCarthy
*)

open Infer
open Printer
open Format
open Compiler

module VD = VComp.Datatypes

open VComp.Constants
open VComp.Specif

open VComp.CPPLlb
open VComp.CPPLss
open VComp.CPPLc
open VComp.CPPL (* Also binds 'Return' *)
open VComp.WPPLi
open VComp.Common (* Also binds 'Bind' *)
open VComp.WPPL 
open VComp.List
open VComp.WPPLo

let cst = VComp.IdSet.IOT.str
let tsc = VComp.IdSet.IOT.rts

let rec nat2intr n i =
  match n with
    | VD.O -> i
    | VD.S n -> nat2intr n (i+1)

let nat2int n = nat2intr n 0

let string_new_id =
  let i = ref 0 in
    fun _ ->
      let new_s = Printf.sprintf "msg%d" (!i) in
	i := (!i) + 1 ;
	new_s

(* WPPL COMPILER INTERFACE *)
let compile_and_print_spec f s =
  let b = Filename.chop_extension f in
  let n r = Printf.sprintf "%s.%s.cppl" b r in
  let Spec (rss, _) = s in
    ignore 
      (VComp.List.map
	  (fun rs ->
	    let RSpec (r, _, _) = rs in
	    let r' = tsc r in
	      match compile_spec r r' s with
		| VD.Some p ->
		    let oc = open_out (n r') in
		      pr_proc (formatter_of_out_channel oc) p
		| VD.None ->
		    (Printf.eprintf "Cannot compile role %s\n" r' ;
		     exit 1))
	  rss)

(* WPPL BINDC *)
let bindc_new_id () = cst (string_new_id ())

(* WPPL EXPLICIT *)
let explicate_spec (s : spec) : spec =
  open_spec_all (bindc_spec bindc_new_id (gen_spec s))

(* WPPL WELL *)
let well_spec (s : spec) : bool =
  wellformed_spec_dec s

(* CPPL WELL *)
let well_proc (p : proc) : bool =
  type_proc_dec p

(* Dispatching *)
let sdispatchT_proc (p1:proc) : bool =
  not (strand_share_recvs_proc_type_self_dec p1)

let sdispatchD_proc (p1:proc) : bool =
  not (strand_share_recvs_proc_dist_self_dec p1)

let sdispatchJ_proc (p1:proc) : bool =
  not (strand_share_recvs_proc_jdist_self_dec p1)

let dispatchT_procs (p1:proc) (p2:proc) : bool =
  not (strand_share_recvs_proc_type_ext_dec p1 p2)

let dispatchD_procs (p1:proc) (p2:proc) : bool =
  not (strand_share_recvs_proc_dist_ext_dec p1 p2)

let dispatchJ_procs (p1:proc) (p2:proc) : bool =
  not (strand_share_recvs_proc_jdist_ext_dec p1 p2)

let do_sdispatch_opt p =
  let init = initial_set_selfp p in
  let ni = VComp.IdSet.IdSet.cardinal init in
  let f = strand_share_recvs_proc_dist_self_redact_dec in
    if not (f init p) then
      let vs = setopt (fun vs -> not (f vs p)) init in
      let es = VComp.IdSet.IdSet.elements vs in
      let n = VComp.List.length es in
	fprintf std_formatter "%a vs %a : %a\n"
	  pp_print_int (nat2int n)
	  pp_print_int (nat2int ni)
	  pr_ids es
    else
      Printf.printf "Cannot optimize\n"

let do_dispatch_opt p1 p2 =
  let init = initial_set_pair p1 p2 in
  let ni = VComp.IdSet.IdSet.cardinal init in
  let f = strand_share_recvs_proc_dist_redact_dec in
    if not (f init p1 p2) then
      let vs = setopt (fun vs -> not (f vs p1 p2)) init in
      let es = VComp.IdSet.IdSet.elements vs in
      let n = VComp.List.length es in
	fprintf std_formatter "%a vs %a : %a\n"
	  pp_print_int (nat2int n)
	  pp_print_int (nat2int ni)
	  pr_ids es
    else
      Printf.printf "Cannot optimize\n"

(* CPPL Backups *)
let print_backups p =  
  ignore
    (VComp.List.map
	(fun si ->
	  let (VD.Coq_pair (s,v)) = si in
	  let ns = strand_all_backups_sizes s in
	  let bs = strand_all_backups s in
	    pr_strand std_formatter bs ;
	    Printf.printf "\n|BS| = ";
	    ignore
	      (VComp.List.map 
		  (fun n -> Printf.printf "%n " (nat2int n))
		  (VComp.List.rev ns)) ;
	    Printf.printf "\n")
	(proc_suc_strands p))

let print_backup_dup p =
  ignore
    (VComp.List.map
	(fun si ->
	  let (VD.Coq_pair (s,v)) = si in
	  let bs = strand_all_backups s in
	    if backup_dup bs then
	      Printf.printf "Has duplicate backup\n"
	    else
	      Printf.printf "\n")
	(proc_suc_strands p))

(* INTERFACE *)
type mode = WPPL | CPPL
let moder = ref WPPL 
let input_file1 = ref "-"
let input_file2 = ref "-"
let explicit = ref true 
let print = ref true 
let well = ref true 
let compile = ref false
let prefix = ref false
let dispatchD = ref false
let dispatchT = ref false
let dispatchJ = ref false
let dispatchopt = ref false
let sdispatchD = ref false
let sdispatchT = ref false
let sdispatchJ = ref false
let sdispatchopt = ref false
let foldm = ref true
let strands = ref false
let backups = ref false
let do_backup_dup = ref false
let thread = ref false
let anon_fun s = 
  if (!input_file1 = "-") then
    input_file1 := s
  else
    input_file2 := s
let arg_spec =
  Arg.align
  [("-P", Arg.Set print, " enable printing (default)") ;
   ("-p", Arg.Clear print, " disable printing") ;
   ("-T", Arg.Set strands, " show strands") ;
   ("-t", Arg.Clear strands, " do not show strands") ;
   ("-W", Arg.Set well, " enable well-formedness check (default)") ;
   ("-w", Arg.Clear well, " disable well-formedness check") ;   
   ("-WPPL", Arg.Unit (fun () -> moder := WPPL), " WPPL mode (default)");
   ("-C", Arg.Set compile, " enable compilation") ;
   ("-c", Arg.Clear compile, " disable compilation (default)") ; 
   ("-E", Arg.Set explicit, " enable explict transformation (default)") ;
   ("-e", Arg.Clear explicit, " disable explict transformation") ;
   ("-Otd", Arg.Set thread, " run thread transformation");
   ("-CPPL", Arg.Unit (fun () -> moder := CPPL), " CPPL mode");
   ("-X", Arg.Set prefix, "  add prefix to identifiers");
   ("-B", Arg.Set backups, " show backup sizes") ;
   ("-BD", Arg.Set do_backup_dup, " perform duplicate backup check") ;
   ("-F", Arg.Set foldm, " enable match fold transformation (default)") ;
   ("-f", Arg.Clear foldm, " disable match fold transformation") ;      
   ("-Os", Arg.Set sdispatchopt, " enable self-dispatch optimizer") ;
   ("-Op", Arg.Set dispatchopt, " enable pair-dispatch optimizer") ;
   ("-Sd", Arg.Unit (fun () -> 
     sdispatchD := true;
     sdispatchT := false), " perform self dispatchD analysis");
   ("-sd", Arg.Clear sdispatchD, " do not perform self dispatchD analysis");
   ("-St", Arg.Unit (fun () -> 
     sdispatchT := true;
     sdispatchD := false), " perform self dispatchT analysis");
   ("-st", Arg.Clear dispatchT, " do not perform self dispatchT analysis");
   ("-Sj", Arg.Unit (fun () -> 
     sdispatchJ := true;
     sdispatchD := false), " perform self dispatchJ analysis");
   ("-sj", Arg.Clear dispatchJ, " do not perform self dispatchJ analysis");
   ("-Dd", Arg.Unit (fun () -> 
     dispatchD := true;
     dispatchT := false), " perform dispatchD analysis");
   ("-dd", Arg.Clear dispatchD, " do not perform dispatchD analysis");
   ("-Dt", Arg.Unit (fun () -> 
     dispatchT := true;
     dispatchD := false), " perform dispatchT analysis");
   ("-dt", Arg.Clear dispatchT, " do not perform dispatchT analysis");
   ("-Dj", Arg.Unit (fun () -> 
     dispatchJ := true;
     dispatchD := false), " perform dispatchJ analysis");
   ("-dj", Arg.Clear dispatchJ, " do not perform dispatchj analysis")]
let usage_msg = "Usage: wpplc [options] [input1] [input2]" ;;

let mkprefix file = 
  if !prefix then
    Printf.sprintf "x%s_" file
  else
    "" ;;

begin
  Arg.parse arg_spec anon_fun usage_msg ;
  match !moder with
    | CPPL ->
	begin
	  let ast1_ut = Reader.read_proc (!input_file1) in
	  let ast1_t = infer_proc (mkprefix (!input_file1)) ast1_ut in
	  let ast1_tf = if !foldm then
	      foldm_proc ast1_t
	    else
	      ast1_t
	  in
	    (if !print then
		print_proc ast1_tf) ;
	    (if !compile then
	       compile_proc ast1_tf) ;
	    (if !well then
		if well_proc ast1_tf then
		  Printf.eprintf "%s is Well-formed\n" (!input_file1)
		else
		  (Printf.eprintf "%s is Not well-formed\n" (!input_file1) ;
		   exit 1)) ;
	    (if !strands then
		print_strands 
		  (proc_suc_strands ast1_tf));
	    (if !backups then
		print_backups ast1_tf) ;
	    (if !do_backup_dup then
		print_backup_dup ast1_tf) ;
	    (if !sdispatchopt then
		do_sdispatch_opt ast1_tf) ;
	    (if !sdispatchT || !sdispatchT || !sdispatchJ then
		(if !sdispatchT then
		    (if sdispatchT_proc ast1_tf then
			Printf.eprintf "T-SDispatchable\n"
		      else
			Printf.eprintf "Not T-SDispatchable\n")) ;
	     (if !sdispatchD then
		 (if sdispatchD_proc ast1_tf then
		     Printf.eprintf "D-SDispatchable\n"
		   else
		     Printf.eprintf "Not D-SDispatchable\n")) ;
	     (if !sdispatchJ then
		 (if sdispatchJ_proc ast1_tf then
		     Printf.eprintf "J-SDispatchable\n"
		   else
		     Printf.eprintf "Not J-SDispatchable\n"))) ;
	    (if !dispatchT || !dispatchD || !dispatchJ || !dispatchopt then
		let ast2_ut = Reader.read_proc (!input_file2) in
		let ast2_t = infer_proc (mkprefix (!input_file2)) ast2_ut in
		let ast2_tf = if !foldm then
		    foldm_proc ast2_t
		  else
		    ast2_t
		in
		  (if !print then
		      print_proc ast2_tf) ;
		  (if !well then
		      if well_proc ast2_tf then
			Printf.eprintf "%s is Well-formed\n" (!input_file2)
		      else
			(Printf.eprintf "%s is Not well-formed\n"
			    (!input_file2) ;
			 exit 1)) ;
		  (if !strands then
		      print_strands 
			(proc_suc_strands ast2_tf));
		  (if !dispatchopt then
		      do_dispatch_opt ast1_tf ast2_tf) ;
		  (if !dispatchT then
		      (if dispatchT_procs ast1_tf ast2_tf then
			  Printf.eprintf "T-Dispatchable\n"
			else
			  Printf.eprintf "Not T-Dispatchable\n")) ;
		  (if !dispatchD then
		      (if dispatchD_procs ast1_tf ast2_tf then
			  Printf.eprintf "D-Dispatchable\n"
			else
			  Printf.eprintf "Not D-Dispatchable\n")) ;
		  (if !dispatchJ then
		      (if dispatchJ_procs ast1_tf ast2_tf then
			  Printf.eprintf "J-Dispatchable\n"
			else
			  Printf.eprintf "Not J-Dispatchable\n")) ;
		  ())
	end
    | WPPL ->    
	begin
	  let ast_ut = Reader.read_spec (!input_file1) in
	  let ast_t_imp = infer_spec ast_ut in
	  let ast_t_exp = if !explicit then
	      explicate_spec ast_t_imp 
	    else
	      ast_t_imp
	  in
	    (if !thread then
	       let t_ast = thread_spec ast_t_exp in
		 print_spec t_ast ;
		 if well_spec t_ast then
		  Printf.eprintf "Threaded: Well-formed\n"
		else
		  (Printf.eprintf "Threaded: Not well-formed\n" ;
		   exit 1)
	    ) ;
	    (* XXX Print strands *)
	    (if !print then
		print_spec ast_t_exp) ;
	    (* XXX Compute warnings (i.e., uninspected messages
	       components; if hash is on constants) *)
	    (if !well then
		if well_spec ast_t_exp then
		  Printf.eprintf "Well-formed\n"
		else
		  (Printf.eprintf "Not well-formed\n" ;
		   exit 1)) ;
	    (if !compile then
		compile_and_print_spec (!input_file1) ast_t_exp) ;
	    ()
	end
end
